/**  
 *  \file events/include/events.h					
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 *  \version 2.0
 *
 *  \brief Event Queue Manager 2.0 - Public Interface
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */

#ifndef _EVENT_QUEUE_H_
#define _EVENTS_QUEUE_H_

#include "../../kdm/include/kdm.h"


/** 
 * \defgroup EventType Event Type Constants 
 */

//@{
#define CREATION                 0x2
#define DELETION                 0x4
#define MODIFICATION             0x8
#define NEW_PREDICTION           0x10
#define MATCHED_PREDICTION       0x20
#define UNMATCHED_PREDICTION     0x40
#define ALL                      0x7d

#define CONSIDER_PERIODIC_AGENT  0x100
#define TIME_MESSAGE		 0x200

#define EXIT_APP                 0x1

#define EVENTS_OK                0

#define NO_EVENT                 0x0

#define NEVENTTYPES              9  ///< Number of different event types

/** 
 *  Event importance: 0..9 
 *
 *        - Lowest: 0
 *        - Biggest: 9
 */
#define MAX_IMPORTANCE_EV        10 

/** Indicates that the values in the event (old and new) doesn't contain time values (long long)
 *       but KDM_Value types
 */
#define NO_TIME_IN_EVENT         (long long) -1     

/**
 *   Value of the iDataType of an event record containing slack (RS-DS event)
 *            Event_T --> New_Value (or Old_Value) --> lTime (long long)
 */
#define TIME_VALUE               -1 

//@}


/// Function Interface

/********************** Event_Notify *************************/
/**
 * \fn Event_Notify
 *
 *  Inserts a new event with the given parameters and creates
 *  a new element in the list, placing it at the end.
 *    - Input Specials:
 *         - iDataType = TIME_VALUE --> indicating slack time in lTimeNew and lTimeOld
 *         - lTimeNew = lTimeOld = NO_TIME_IN_EVENT --> KDM event
 *    - Return Value:
 *         - (-1) : There even has not been generated
 *         -   0  : The event has been placed in the store
 */
extern int Event_Notify( int iType, void *pSlot, int iImp, int iDataType, 
                         long long lTimeNew, KDM_Value New_Value, 
                         long long lTimeOld, KDM_Value Old_Value );


/********************** Event_Get *************************/
/**
 * \fn Event_Get
 *
 * Extract the oldest node of the biggest level of importance 
 *   - Return Value: 
 *        - (-1) : There is no node
 *        -   0  : The args. have the event of the node
 *   - *piDataType = TIME_VALUE --> indicating slack time in lTimeNew and lTimeOld
 *   - *plTimeNew = *plTimeOld = NO_TIME_IN_EVENT --> KDM event
 */
extern int Event_Get( int *piType, void **ppSlot, int *piImp, int *piDataType, 
                      long long *plTimeNew, KDM_Value *pNew_Value, 
                      long long *plTimeOld, KDM_Value *pOld_Value );


/********************** Events_bIsEmpty *************************/
/**
 * \fn Events_bIsEmpty
 * 
 *  - Return Value: 
 *        - 1 : There is no event
 *        - 0 : There are events
 *
 */
extern int Events_bIsEmpty (void);


/********************** Events_Initialize *************************/
/**
 * \fn Events_Initialize
 *
 *  Public generic initializing function 
 *
 */
extern int Events_Initialize(void);


/********************** Events_Free *************************/
/**
 * \fn Events_Free
 *
 * Public generic freeing function 
 *
 ***/
extern void Events_Free(void);


#ifndef MODULE

/********************** Events_NameType *************************/
/**
 * \fn Events_NameType
 *
 * Returns the proper event name associated to the event type it 
 *     receives as a parameter (iType).
 *
 ***/
extern char * Events_NameType( int iType );

#endif


/********************** Events_ShowType *************************/
/**
 * \fn Events_ShowType
 *
 * Shows the proper event name associated to the event type it 
 *      receives as a parameter (iType).
 *
 ***/
extern void Events_ShowType ( int iType );


/********************** Events_ShowAll *************************/
/**
 * \fn Events_ShowAll
 *
 * Shows the information associated with all the events  
 *      which are currently in the Event Store.
 *
 ***/
extern void Events_ShowAll ( void );

/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */

#endif
