/**  
 *  \file events.c
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 *  \version 2.0
 *
 *  \brief Event Queue Manager 2.0
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */

#ifdef MODULE

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <mbuff.h>
#else
#include <stdio.h>
#include <stdlib.h>  /// malloc
#include <string.h>

extern int mbuff_alloc(const char *name, int size);
extern int mbuff_free(const char *name, void * mbuf);
extern void * mbuff_attach(const char *name, int size);
extern void mbuff_detach(const char *name, void * mbuf);

#endif

#ifdef MODULE
#include <rtl.h>   /// rtl_printf
#endif

#include "../../common/include/artis_conf.h"
#include "../../common/include/artis_event_definition.h"  /// Trace Event definitions
#include "../../trace/include/trace.h"

#include "../include/events/events_internal.h"


/// There is no meta-reasoning yet
///#include "../include/metareasoning.h"


///#define EVENTS_DEBUG


static __inline__ unsigned long long rdtsc(void)
{
        unsigned long long ret;
        __asm__ __volatile__("rdtsc"
                            : "=A" (ret)
                            : /* no inputs */);
        return ret;
}


/** Dynamic array of event queues. Each element of the array has the
 *     ptrs to the queue of events which importance is the same as 
 *     the index of the element in the array.
 *   The last element of this array will have the queue of empty nodes. 
 *     This is to mantain a constant number of nodes in the array.
 */
n_Imp * aEvImp;   

/** To control if there's anyone accessing the storage 
 *            Values:
 *               - 0 : There's no one accessing
 *               - 1 : There's someone here
 */
volatile int *piMutex;  

int *piEvents; ///< Current number of events


/********************** Events_InitList *************************/
/** 
 * Initializes a Event Queue, with iNodes empty nodes 
 */
int Events_InitList( n_Event * pBeginNodesAddr, 
           int * pFirst, int * pLast, int iNodes ) {           

  n_Event * new_node = NULL;
  int i;

  if ( iNodes <= 0 )
    return - 1;   /*** what the hell is that?? ***/
  else {
         
         * pFirst = * pLast = 0;

	 /*** Trying to struct & insert the iNode - 1 empty nodes ***/
         for ( i = 0; 
               ((i < iNodes) && 
                ((new_node = (n_Event *) (pBeginNodesAddr + i)) != NULL));
               /*** + ( sizeof( n_Event ) * i ))) != NULL )); ***/
               i ++ ) { 

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifndef MODULE                   
     fprintf(stderr, " (events.c:Events_InitList) Nodo %d (dir: %p) antes de cambiar los ptrs \n", 
                                i, new_node);  
#else
     rtl_printf(" (events.c:Events_InitList) Nodo %d (dir: %x) antes de cambiar los ptrs \n",
                                i, new_node);
#endif
#endif

                   new_node->ev.iType = -1;  /*** Empty Event ***/
                   ( * ( pBeginNodesAddr + * pFirst )).prev = i; 
                   new_node->sig = * pFirst;

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifndef MODULE                   
     fprintf(stderr, " (events.c:Events_InitList) Nodo %d (dir: %p) cambiando los ptrs \n", 
                                i, new_node); 
#else
     rtl_printf(" (events.c:Events_InitList) Nodo %d (dir: %p) cambiando los ptrs \n", 
                                i, new_node);
#endif
#endif

                   new_node->prev = -1; ///* pLast;

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifndef MODULE                   
     fprintf(stderr, " (events.c:Events_InitList) Nodo %d (dir: %p) cambiando los ptrs \n", 
                                i, new_node);  
#else
     rtl_printf(" (events.c:Events_InitList) Nodo %d (dir: %p) cambiando los ptrs \n", 
                                i, new_node);
#endif
#endif

                   * pFirst = i;  /***new_node;***/

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifndef MODULE                   
     fprintf(stderr, " (events.c:Events_InitList) Nodo %d (dir: %p) cambiando los ptrs \n", 
                                i, new_node); 
#else
     rtl_printf(" (events.c:Events_InitList) Nodo %d (dir: %p) cambiando los ptrs \n", 
                                i, new_node);
#endif
#endif

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifndef MODULE                   
                  fprintf(stderr, " (events.c:Events_InitList) Insertando nodo de evento vacio: Num: %d, Sig: %d, Prev: %d\n", 
                                i, new_node->sig, new_node->prev); 
#else
                  rtl_printf(" (events.c:Events_InitList) Insertando nodo de evento vacio: Num: %d, Sig: %d, Prev: %d\n",  
                                i, new_node->sig, new_node->prev); 
#endif
#endif

         }
         
         (*(pBeginNodesAddr + * pLast)).sig = -1; ///*pFirst; /***new_node;***/

         if ( new_node == NULL ) 
           /*** There is not enough memory ***/
           return - 1;
         else return EVENTS_OK;
  }
}


/********************** Events_CreateQueueArray *************************/
/**
 * Createses the Event queue array, shaping all its empty nodes 
 *    This is used in the DS level, to allocate the memory that will
 *    be used by the DS to manage events
 */
int Events_CreateQueueArray( void ) {

  n_Event * pBeginNodesAddr;


  if (( piMutex = ( volatile int * ) mbuff_attach ( EVENT_STORE, 
                                 ( 2 * ((int) sizeof(int)) + 
                                 ( MAX_IMPORTANCE_EV + 1 ) * ((int) sizeof( n_Imp ))) +
                                 ( ((int) sizeof( n_Event )) * MAX_EV ) ) ) == NULL )
     /**< Attach to the memory reserved to store:
      *      - The number of events in the queues (*piEvents --> int)
      *      - The array of event queues (aEvImp --> 
      *                 (MAX_IMPORTANCE_EV + 1) n_Imp elements
      *      - The event nodes ( MAX_EV nodes of type n_Event )
      */
    return - 1;  
  else {
         /**< piEvents points to an integer, and the next address is the array
          *    of event queues.
          */

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifndef MODULE
  fprintf(stderr, "(events.c:Events_CreateQueueArray) He allocado la memoria bien? Dir. Inicial: %x Dir. Final: %x\n",
                        (unsigned int) piMutex, 
                        (unsigned int) (((int) piMutex) + ( sizeof (int) + 
                            ( MAX_IMPORTANCE_EV + 1 ) * sizeof( n_Imp ) +
                             sizeof( n_Event ) * MAX_EV )));
#else
  rtl_printf( "(events.c:Events_CreateQueueArray) He allocado la memoria bien? Dir. Inicial: %x Dir. Final: %x\n",
                        piMutex, ((int) piMutex) + ( sizeof( int) + 
                            ( MAX_IMPORTANCE_EV + 1 ) * sizeof( n_Imp ) +
                             sizeof( n_Event ) * MAX_EV ));
#endif
#endif

         *piMutex = 0; /*** Initially there's no one inside ***/

         piEvents = (int *) (piMutex + 1); 

         aEvImp = ( n_Imp * ) ( piEvents + 1); /*** + sizeof(int); ***/  

         /// pBeginNodesAddr is a pointer to the beginning of the space for the nodes
         pBeginNodesAddr = ( n_Event * ) (aEvImp + /***( sizeof( n_Imp ) * ***/
                                                           ( MAX_IMPORTANCE_EV + 1 ));

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifndef MODULE
  fprintf( stderr, "(events.c:Events_CreateQueueArray) Antes del Events_InitList\n");
#else
  rtl_printf("(events.c:Events_CreateQueueArray) Antes del Events_InitList\n");
#endif
#endif

         return EVENTS_OK;

  }
}


/********************** Events_InitQueueArray *************************/
/**
 * Initializes the Event queue array, shaping all its empty nodes 
 *   This is used in the RTOS level, to contact with the memory 
 *   allocated by the RTOS in the initialization phase 
 */
int Events_InitQueueArray( void ) {
  int i;

  n_Event * pBeginNodesAddr;

  if (( piMutex = ( volatile int * ) mbuff_attach ( EVENT_STORE, 
                 ( 2 * ((int) sizeof( int)) + 
                   ( MAX_IMPORTANCE_EV + 1 ) * ((int) sizeof( n_Imp ))) +
                   ( ((int) sizeof( n_Event )) * MAX_EV ) ) ) == NULL )
     /**< Attach to the memory reserved to store:
      *      - The number of events in the queues (*piEvents --> int)
      *      - The array of event queues (aEvImp --> (MAX_IMPORTANCE_EV + 1)
      *           n_Imp elements
      *      - The event nodes ( MAX_EV nodes of type n_Event )
      */
    return - 1;  
  else {
         
    *piMutex = 0; ///< Initially there's no one inside

         piEvents = (int *) (piMutex + 1); 

         *piEvents = 0;      /*** Initial number of events ***/

         /** piEvents points to an integer, and the next address is the array
          *    of event queues.
          */
         aEvImp = ( n_Imp * ) ( piEvents + 1 ); /*** + sizeof(int) ); ***/

         /// All the queues corresponding to events are empty 
         for ( i = 0; i < MAX_IMPORTANCE_EV; i++ )
	   aEvImp[i].pFirst = aEvImp[i].pLast = -1; 
  
         /// pBeginNodesAddr is a pointer to the beginning of the space for the nodes
         pBeginNodesAddr = ( n_Event * ) aEvImp + 
                                                (MAX_IMPORTANCE_EV + 1);

         /// The last queue has all the nodes (empty ones)
          return( Events_InitList( pBeginNodesAddr, 
                                   &aEvImp[MAX_IMPORTANCE_EV].pFirst,
                                   &aEvImp[MAX_IMPORTANCE_EV].pLast,
                                   MAX_EV ) );

  }
}


/********************** Events_Initialize *************************/
/**
 * Public generic initializing function 
 */
int Events_Initialize() {
  /// RT-SIDE does an mbuff_attach
#ifdef MODULE
  return  Events_InitQueueArray();
  ///while OPTIONAL SIDE does first a mbuff_alloc
#else
  return  Events_CreateQueueArray();
#endif
}


/********************** Events_Free *************************/
/**
 * Public generic freeing function 
 */
void Events_Free() {
#ifdef MODULE
  mbuff_detach(EVENT_STORE,(void *)piMutex);
#else
  mbuff_detach(EVENT_STORE,(void *)piMutex);
#endif
}


/********************** Events_GetEmptyNode *************************/
/**
 * Extracts one empty node to assign it to pEmpty, 
 *    or the oldest event node of the lowest 
 *    event importance lesser than iImp.
 *  - Return Value: 
 *        - (- 1) : There is no node of importance lower than iImp.
 *        -    0  : pEmpty has the pointer to the node
 *
 */
int Events_GetEmptyNode ( int * pEmpty, int iImp ) { 

  int pAux;

  /// pBeginNodesAddr is a ptr to the beginning of the nodes space
  n_Event * pBeginNodesAddr = ( n_Event * ) aEvImp + 
                                                ( MAX_IMPORTANCE_EV + 1 );

  if ((pAux = aEvImp[MAX_IMPORTANCE_EV].pLast) == -1) {
    /**< There are not empty nodes 
     *       -> Extract the oldest node of the lowest level of importance 
     *              that is lower than iImp. 
     */
    int i;

    for ( i = 0; 
          (( i < iImp ) && (( pAux = aEvImp[i].pLast ) == -1 )); 
          i++ );
   
    if ( pAux == -1 ) {
      ///< There is no node of importance lower than iImp
      return - 1;
    }
    else {
      /// Extract the node
      if ( (*(pBeginNodesAddr + pAux)).prev != -1 ) {
        (*(pBeginNodesAddr + (*(pBeginNodesAddr + pAux)).prev)).sig = ///-1;
                (*(pBeginNodesAddr + pAux)).sig;
		
        aEvImp[i].pLast = (*(pBeginNodesAddr + pAux)).prev;
      }
      else aEvImp[i].pFirst = aEvImp[i].pLast = -1;
      
      (* piEvents) --;
    }
    
  }
  else {
    ///< There are empty nodes

    if (aEvImp[MAX_IMPORTANCE_EV].pFirst ==
        aEvImp[MAX_IMPORTANCE_EV].pLast) {
      ///< There are only one empty node
      aEvImp[MAX_IMPORTANCE_EV].pFirst =
        aEvImp[MAX_IMPORTANCE_EV].pLast = -1; 
    }
    else {
      ///< There are more than one empty node
      aEvImp[MAX_IMPORTANCE_EV].pLast = (*(pBeginNodesAddr + pAux)).prev;
      (*(pBeginNodesAddr + aEvImp[MAX_IMPORTANCE_EV].pLast)).sig = -1;
    }
  }
  
  * pEmpty = pAux;
  return 0;
  
}


/********************** Events_Insert_Event *************************/
/**
 * Creates a new event with the given parameters and creates
 *  a new element in the list, placing it at the end 
 *   - Return Value:
 *        - (- 1) : There even has not been generated
 *        -    0  : The event has been placed in the store
 */
int Events_Insert_Event( Event_T regEvent ) {

  int pAux;

  n_Event *pBeginNodesAddr = (n_Event *) aEvImp +
                                                   (MAX_IMPORTANCE_EV + 1);

  if (*piMutex) {  
    ///< There is someone inside!!!!!!!!!!!
#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifdef MODULE
    rtl_printf("\t(events.c)::Events_Insert_Event ERROR ==> INSERTANDO CUANDO HAY YA ALGUIEN EN CASA \n");
#endif   
#endif
    return -2;
  }
  else *piMutex = 1; ///< And now, it is mine

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifdef MODULE
    rtl_printf("\t(events.c)::Events_Insert_Event: Before rdtsc \n");
#endif   
#endif

  regEvent.lTimeStamp = rdtsc();

#ifdef ARTIS_CONF_EVENTS_DEBUG_0
#ifdef MODULE
  trace_event( INSERT_EVENT, 0, (long long) regEvent.iType, 0 ); 
#else
  trace_event( INSERT_EVENT, 1, (long long) regEvent.iType, 0 ); 
#endif
#endif

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifdef MODULE
  rtl_printf("-_-_-_-(RTOS)_(events.c)::Events_Insert_Event = %d, Importance = %d\n",
             regEvent.iType,
             (int) regEvent.iImp );      
#else
  fprintf(stderr,"-_-_-_-_\t(CONTROL)(events.c)::Events_Insert_Event = %d, Importance = %d\n",
         regEvent.iType,
         (int) regEvent.iImp );      
#endif
#endif

  if ( !Events_GetEmptyNode ( & pAux, (int) regEvent.iImp ) ) {
    ///< pAux has an empty node

#ifdef ARTIS_CONF_EVENTS_DEBUG_2
#ifdef MODULE
    rtl_printf("\t(events.c)::Events_Insert_Event ==> Empty Node found\n" );  
#endif
#endif

    ( * ( pBeginNodesAddr + pAux )).ev = regEvent;
    ( * ( pBeginNodesAddr + pAux )).prev = -1;

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifdef MODULE
    rtl_printf("\t|\t(events.c)::Events_Insert_Event:Tipo:%d, Imp.:%d, lTimeNew:%lld, lTimeOld: %lld, lTimeStamp: %lld\n",
             ( * ( pBeginNodesAddr + pAux )).ev.iType,
             ( * ( pBeginNodesAddr + pAux )).ev.iImp,
	     ( * ( pBeginNodesAddr + pAux )).ev.New_Value.lTime,
	     ( * ( pBeginNodesAddr + pAux )).ev.Old_Value.lTime,
             ( * ( pBeginNodesAddr + pAux )).ev.lTimeStamp);      
#endif
#endif
    
    if ( aEvImp[(int) regEvent.iImp].pFirst == -1 ) {
      ///< This event queue is empty
      aEvImp[(int) regEvent.iImp].pFirst = 
                               aEvImp[(int) regEvent.iImp].pLast = pAux;
      ( * ( pBeginNodesAddr + pAux )).sig = -1;
    }
    else {
      ///< This is the new first one
      (*(pBeginNodesAddr + pAux)).sig = aEvImp[(int) regEvent.iImp].pFirst;
      (*(pBeginNodesAddr + aEvImp[(int) regEvent.iImp].pFirst)).prev = pAux;
      aEvImp[(int) regEvent.iImp].pFirst = pAux;
    }
    
    (*piEvents) ++;

#ifdef ARTIS_CONF_EVENTS_DEBUG_2
#ifndef MODULE
    fprintf( stderr, "(events.c:Events_Insert_Event) Num. de Eventos: %d\n",
             *piEvents);
#else
    rtl_printf("(events.c:Events_Insert_Event) Num. de Eventos: %d\n",
               *piEvents);
#endif
#endif

    *piMutex = 0; ///< If U want it, U got it

    return 0;
  }
  else {
    /** 
     *   We must sustitute the last node of this event importance level 
     *     if there is any 
     */
    if ( aEvImp[(int) regEvent.iImp].pLast == -1 ){
      ///< This event queue is empty

#ifdef ARTIS_CONF_EVENTS_DEBUG_2      
#ifdef MODULE
      rtl_printf("(events.c:Events_Insert_Event) %d Eventos, pero no puedo\n",
                 *piEvents);
      //ShowEvents();
#endif
#endif

      *piMutex = 0; ///< If U want it, U got it 

      return -1;
    }
    else {
      /// Substitute the last node of the given importance
      (*(pBeginNodesAddr + aEvImp[(int) regEvent.iImp].pLast)).ev = 
        regEvent;  
      
      if (aEvImp[(int) regEvent.iImp].pFirst != 
          aEvImp[(int) regEvent.iImp].pLast) {
        ///< The last will be the 1st
        (*(pBeginNodesAddr + aEvImp[(int)regEvent.iImp].pLast)).sig = 
          aEvImp[(int)regEvent.iImp].pFirst;
        (*(pBeginNodesAddr + aEvImp[(int)regEvent.iImp].pFirst)).prev =
          aEvImp[(int) regEvent.iImp].pLast;

        aEvImp[(int) regEvent.iImp].pFirst = 
          aEvImp[(int) regEvent.iImp].pLast;
        aEvImp[(int) regEvent.iImp].pLast = 
          (*(pBeginNodesAddr + aEvImp[(int)regEvent.iImp].pLast)).prev;

	/***
        if ((*(pBeginNodesAddr + aEvImp[(int)regEvent.iImp].pLast)).prev 
            != aEvImp[(int)regEvent.iImp].pFirst)
          ///< To avoid that the prev of the last will be itself
          (*(pBeginNodesAddr + aEvImp[(int)regEvent.iImp].pLast)).prev =
            (*(pBeginNodesAddr + aEvImp[(int)regEvent.iImp].pFirst)).prev;
***/

        (*(pBeginNodesAddr + aEvImp[(int)regEvent.iImp].pFirst)).prev = -1;
        (*(pBeginNodesAddr + aEvImp[(int) regEvent.iImp].pLast)).sig = -1;
        
#ifdef ARTIS_CONF_EVENTS_DEBUG_2                     
#ifndef MODULE
        fprintf( stderr, "(events.c:Events_Insert_Event) Queue FULL Num. de Eventos: %d\n", 
                 *piEvents);
#else
        rtl_printf("(events.c:CreateEventQueueArray) Queue FULL Num. de Eventos: %d\n", 
                   *piEvents);
#endif
#endif
      }
      *piMutex = 0; ///< If U want it, U got it
      
      return 0;
    }
  }
}


/********************** Events_Extract_Event *************************/
/**
 * Extract the oldest node of the biggest level of importance 
 *  - Return Value: 
 *        - (- 1) : There is no node
 *        -    0  : regEvent has the event of the node
 */
int Events_Extract_Event( Event_T * regEvent ) {

   int i;
   int pAux;

   n_Event * pBeginNodesAddr = ( n_Event * ) aEvImp + 
                                           ( MAX_IMPORTANCE_EV + 1 );

   if (*piMutex) {  
     ///< There is someone inside!!!!!!!!!!!

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifndef MODULE    
     fprintf( stderr, "(events.c::Events_Extract_Event) ERROR ==> NO HE PODIDO EXTRAER EVENTO POR EL MUTEX\n");
#endif
#endif

     return -2;
   }
   
   *piMutex = 1; ///< And now, it is mine

   for ( i = MAX_IMPORTANCE_EV - 1; 
         (( i >= 0  ) && (( pAux = aEvImp[i].pLast ) == -1 )); i-- );
   
   if ( pAux == -1 ) {
     ///< There is no node
#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifndef MODULE
     fprintf(stderr, "(events.c) Can't find any node\n");
#endif
#endif
     *piMutex = 0; ///< If U want it, U got it
     return - 1;
   }
   else {
     ///< Extract the node
     if ( aEvImp[i].pFirst == aEvImp[i].pLast )
       aEvImp[i].pFirst = aEvImp[i].pLast = -1;
     else {
       /// Extract the node
       (*(pBeginNodesAddr + (*(pBeginNodesAddr + pAux )).prev )).sig = ///-1;
                (*(pBeginNodesAddr + pAux)).sig;
       aEvImp[i].pLast = ( * ( pBeginNodesAddr + pAux )).prev;
     }
     
     * regEvent = ( * ( pBeginNodesAddr + pAux )).ev;
     
     ( * ( pBeginNodesAddr + pAux )).ev.iType = -1; ///< Empty Event
     
     (*piEvents) --;
     
#ifdef ARTIS_CONF_EVENTS_DEBUG_0
#ifdef MODULE
     trace_event( EXTRACT_EVENT, 0, (long long) regEvent->iType, 0 ); 
#else
     trace_event( EXTRACT_EVENT, 1, (long long) regEvent->iType, 0 ); 
#endif
#endif
     
#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifndef MODULE
     fprintf(stderr, "(events.c) Extraigo un evento: Numero %d, importancia %d, tipo %d, quedan %d\n", 
             pAux, (int) regEvent->iImp, regEvent->iType, (*piEvents));
#else 
     rtl_printf("(rtl_events) Extraigo un evento: Numero %d, importancia %d, tipo %d, quedan %d\n",
                pAux, (int) regEvent->iImp, regEvent->iType, (*piEvents));
#endif
#endif  

     /// Insert the extracted event node in the Empty Queue
     (*(pBeginNodesAddr + pAux)).sig = aEvImp[MAX_IMPORTANCE_EV].pFirst;
     if (aEvImp[MAX_IMPORTANCE_EV].pFirst != -1)
       (*(pBeginNodesAddr + aEvImp[MAX_IMPORTANCE_EV].pFirst)).prev = pAux;
     else aEvImp[MAX_IMPORTANCE_EV].pLast = pAux;
     aEvImp[MAX_IMPORTANCE_EV].pFirst = pAux;
     (*(pBeginNodesAddr + aEvImp[MAX_IMPORTANCE_EV].pFirst)).prev = -1;
   }
   
   *piMutex = 0; ///< If U want it, U got it
   
   return 0;
}


/********************** Event_Notify *************************/
/**
 * Inserts a new event with the given parameters and creates
 *  a new element in the list, placing it at the end.
 *   - Return Value:
 *        - (- 1) : There even has not been generated
 *        -    0  : The event has been placed in the store
 */
int Event_Notify( int iType, void *pSlot, int iImp, int iDataType, 
                  long long lTimeNew, KDM_Value New_Value, 
                  long long lTimeOld, KDM_Value Old_Value ){
  Event_T regEvent;

  regEvent.iType     = iType;
  regEvent.pSlot     = pSlot;
  regEvent.iImp      = iImp;

  if (( lTimeNew == NO_TIME_IN_EVENT ) && ( iDataType != TIME_VALUE )) {
    regEvent.iDataType = iDataType;
    regEvent.New_Value.KDM_v = New_Value;
    regEvent.Old_Value.KDM_v = Old_Value;
  }
  else {
    regEvent.iDataType       = TIME_VALUE;
    regEvent.New_Value.lTime = lTimeNew;
    regEvent.Old_Value.lTime = lTimeOld;
  }

#ifdef ARTIS_CONF_EVENTS_DEBUG_1
#ifdef MODULE
  rtl_printf("(events.c) Event_Notify, before Events_Insert_Event\n");
#endif
#endif

  return ( Events_Insert_Event( regEvent ) );
}


/********************** Event_Get *************************/
/**
 * Extract the oldest node of the biggest level of importance 
 *   - Return Value: 
 *        - (- 1) : There is no node
 *        -    0  : The args. have the event of the node
 */
int Event_Get( int *piType, void **ppSlot, int *piImp, int *piDataType, 
               long long *plTimeNew, KDM_Value *pNew_Value, 
               long long *plTimeOld, KDM_Value *pOld_Value ){
  Event_T regEvent;

  if (Events_Extract_Event( &regEvent ) < 0)
    return -1;
  else {
    *piType     = regEvent.iType;
    *ppSlot     = regEvent.pSlot;
    *piImp      = regEvent.iImp;
    *piDataType = regEvent.iDataType;

    if ( *piDataType != TIME_VALUE ) {
      *plTimeNew = *plTimeOld = NO_TIME_IN_EVENT;
      *pNew_Value = regEvent.New_Value.KDM_v;
      *pOld_Value = regEvent.Old_Value.KDM_v;
    }
    else {
      pNew_Value = pOld_Value = (KDM_Value *) NULL;
      *plTimeNew = regEvent.New_Value.lTime;
      *plTimeOld = regEvent.Old_Value.lTime;
    }
    
    return 0;
  }
}


/********************** Events_bIsEmpty *************************/
/**
 *   - Return Value: 
 *        - 1 : There is no event
 *        - 0 : There are events
 *
 */
int Events_bIsEmpty ( ) {

   return ( (*piEvents) == 0 );

}


#ifndef MODULE
/********************** Events_NameType *************************/
/**
 * Returns the proper event name associated to the event type it 
 *     receives as a parameter (iType).
 */
char * Events_NameType ( int iType ) {
   char *sAux;
   
   sAux = (char *) malloc (100 * sizeof(char));

   switch ( iType ) {
        case CONSIDER_PERIODIC_AGENT: sprintf(sAux, "CONSIDER_PERIODIC_AGENT");
                                      break;
        case EXIT_APP: sprintf(sAux, "EXIT");
                       break;
	case CREATION: sprintf(sAux, "CREATION");
	               break;
	case DELETION: sprintf(sAux, "DELETION");
	               break;    
	case MODIFICATION: sprintf(sAux, "MODIFICATION");
		           break;    
        case NEW_PREDICTION: sprintf(sAux, "NEW_PREDICTION");
	                     break;    
        case MATCHED_PREDICTION: sprintf(sAux, "MATCHED_PREDICTION");
	  	                 break;   
        case UNMATCHED_PREDICTION: sprintf(sAux, "UNMATCHED_PREDICTION");
		                   break;    
        case -1: fprintf(stderr, "\tEMPTY NODE\n");
   }   
   return sAux;
}
#endif


/********************** Events_ShowType *************************/
/**
 * Shows the proper event name associated to the event type it 
 *      receives as a parameter (iType).
 */
#ifdef MODULE
void Events_ShowType ( int iType ) {
   switch ( iType ) {
        case CONSIDER_PERIODIC_AGENT: 
          rtl_printf("\tCONSIDER_PERIODIC_AGENT\n");
          break;
	case CREATION: rtl_printf("\tCREATION\n");
	               break;
	case DELETION: rtl_printf("\tDELETION\n");
	               break;    
	case MODIFICATION: rtl_printf("\tMODIFICATION\n");
		           break;    
        case NEW_PREDICTION: rtl_printf("\tNEW_PREDICTION\n");
	                     break;    
        case MATCHED_PREDICTION: rtl_printf("\tMATCHED_PREDICTION\n");
	  	                 break;   
        case UNMATCHED_PREDICTION: rtl_printf("\tUNMATCHED_PREDICTION\n");
		                   break;    
        case -1: rtl_printf("\tEMPTY NODE\n");
   }   
}
#else
void Events_ShowType ( int iType ) {
   switch ( iType ) {
        case CONSIDER_PERIODIC_AGENT: fprintf( stderr, "\tCONSIDER_PERIODIC_AGENT\n" );
                                      break;
	case CREATION: fprintf(stderr, "\tCREATION\n");
	               break;
	case DELETION: fprintf(stderr, "\tDELETION\n");
	               break;    
	case MODIFICATION: fprintf(stderr, "\tMODIFICATION\n");
		           break;    
        case NEW_PREDICTION: fprintf(stderr, "\tNEW_PREDICTION\n");
	                     break;    
        case MATCHED_PREDICTION: fprintf(stderr, "\tMATCHED_PREDICTION\n");
	  	                 break;   
        case UNMATCHED_PREDICTION: fprintf(stderr, "\tUNMATCHED_PREDICTION\n");
		                   break;    
        case -1: fprintf(stderr, "\tEMPTY NODE\n");
   }   
}
#endif


/********************** Events_ShowAll *************************/
/**
 * Shows the information associated with all the events  
 *      which are currently in the Event Store.
 */
void Events_ShowAll ( void ) {
  int i, pAux, nEvents;
  n_Event * pBeginNodesAddr = ( n_Event * ) aEvImp +
                                                    ( MAX_IMPORTANCE_EV + 1 );

#ifdef MODULE
  rtl_printf("\t(events.c::ShowEvents) Events Store:\n");
#else
  fprintf(stderr, "\t(events.c::ShowEvents) Events Store: %d Events:\n",
		  *piEvents);
#endif

  for ( i = 0,   nEvents = *piEvents; 
        (i < MAX_IMPORTANCE_EV) && (nEvents); 
        i++ ) {
      pAux = aEvImp[i].pFirst;
      if ( pAux != -1 ) 
         do {
#ifdef MODULE
           rtl_printf(" \t   Importance: %d TimeStamp: %lld ", 
#else
           fprintf(stderr, " \t   Importance: %d TimeStamp: %lld ", 
#endif
                   (int) ( * (pBeginNodesAddr + pAux)).ev.iImp,
                   (long long) ( * (pBeginNodesAddr + pAux)).ev.lTimeStamp);
           Events_ShowType (( * (pBeginNodesAddr + pAux)).ev.iType );
#ifdef MODULE
           rtl_printf("\n");
#else
           fprintf(stderr, "\n");
#endif
           
           nEvents--;

           pAux =  ( * (pBeginNodesAddr + pAux)).sig;
         } while ((nEvents) && (pAux != aEvImp[i].pFirst));
  }
}


#ifdef MODULE

/* ====================================================================	*/
int Events_init_module(void) 
/* ====================================================================	*/
{
  rtl_printf("KDM EVENTS QUEUER STARTING\n");
  return Events_Initialize();
}


/* ====================================================================	*/
void Events_cleanup_module(void)
/* ====================================================================	*/
{
  mbuff_detach(EVENT_STORE,(void *)piMutex);
}

#endif

/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */

