/*!
  \file artis/include/kdm.h
  \date  Time-stamp: <16-03-2005 15:28>
  \author M. Rebollo <mrebollo@dsic.upv.es>
  \version 0.2

  \brief user access functions to KDM information

  The KDM (Knowledge Data Manager) is the module which manages
  the set of beliefs os the agent. It provides the user with
  several functions to access to the data that the user has defined.

  Mainly, these functions are classified in two types, depending on
  the temporal characterisation of the data: static or temporal. 
  Static information does not maintain any kind of information about
  the instant in which values are updated. Temporal slots accessing
  functions label each data with their instant of occurrence.

  Static slots' accessing functions get or set the value for the slot.
  An static slot have to be passed to these functions. The function 
  has to be rewritten to use a pointer to the correspondent structure.
*/

#ifndef _KDMUSER_H_
#define _KDMUSER_H_

#include <stdarg.h>
#include "../kdm/include/kdm.h"


////////////////////////////////////////////////////////////////////
//static data management
////////////////////////////////////////////////////////////////////


/*!
  \fn    void* KDM_Get_Value(s)
  \param static_slot s The slot whose value we want to recover.
  \return A pointer to the value
  \brief Returns the address of thevalue stored in slot \s, 
         as a void pointer.

         This function always returns a void* value, so the
	 user have to make a casting before assign the value
	 to a variable.

	 Furthermore, the format used for the slot is instance->slot
	 The KDM automatically takes the address of this slot and
	 uses a relative reference to access to the required slot.

	 For instance, if we have defined a instance called robot, with
	 a integer slot xpos, we have write the following code.

	 Example:
	 \code
	 int x_pos;
	 ...
	 x_pos = *(int*)KDM_Get_Value( robot->xpos );
	 \endcode
*/
#define KDM_Get_Value(s) kdm_get_slot_value((&s))


#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
/*!
  \fn    int KDM_Set_Value(s,val)
  \param static_slot s The slot whose value we want to change
  \param val the new value for the slot
  \return 0: value changed; -1: otherwise
  \brief Sets the value of 's' to 'val'

         This function allows the user to write values on the KDM 
	 data structure. 

	 The function has to indicate the name of the slot (using a
	 instance->slot structure) and the value to be stored. The 
	 types must be compatibles. The system does not perform
	 any type checking.

	 Example:
	 \code
	 int x_pos;
	 ...
	 KDM_Set_Value( robot->xpos, x_pos + 1 );
	 \endcode

	 This function is implemented as a define, which rewrites the 
	 call to the function. We have defined two possible conversions
	 depending on the compiler version.
*/
# define KDM_Set_Value(s,...)  kdm_set_slot_value( (&s), _VA_ARGS__ )
#else
/*!
  \fn    int KDM_Set_Value(s,val)
  \param static_slot s The slot whose value we want to change
  \param val the new value for the slot
  \return 0: value changed; -1: otherwise
  \brief Sets the value of 's' to 'val'

         This function allows the user to write values on the KDM 
	 data structure. 

	 The function has to indicate the name of the slot (using a
	 instance->slot structure) and the value to be stored. The 
	 types must be compatibles. The system does not perform
	 any type checking.

	 Example:
	 \code
	 int x_pos;
	 ...
	 KDM_Set_Value( robot->xpos, x_pos + 1 );
	 \endcode

	 This function is implemented as a define, which rewrites the 
	 call to the function. We have defined two possible conversions
	 depending on the compiler version.
*/
# define KDM_Set_Value(s,args...)  kdm_set_slot_value((&s),##args )
#endif


////////////////////////////////////////////////////////////////////
//event management
////////////////////////////////////////////////////////////////////

/*!
  \fn    int KDM_Add_Event(s,e,i)
  \param static slot s slot whose event list we want to change
  \param event e new event to add to the event list
  \param int i iportance associated to all the events for the slot
  \return 0: event added; -1: otherwise
  \brief Adds the 'e' event to the events of interest list
         for the slot 's'

	 This funcion adds a new event to the list of events associated
	 to the slot 's'. In so doing, the system get informed when the
	 corresponding change, which is represented by the event,
	 is produced in the agent belief.

	 The importance is an integer between 0 (maximum importance) and
	 9 (minimum importance).

	 Example:
	 \code
	 KDM_Add_Event( robot->xpos, MODIFICATION, 0 );
	 \endcode
	 Adds a MODIFICATION event to the current list of events for
	 the slot xpos in the instance robot at maximum importance.
*/
#define KDM_Add_Event(s,e,i) kdm_event_add( (&s),e,i )

/*!
  \fn    KDM_Release_Event(s,e)
  \param static slot s slot whose event list we want to change
  \param event e event to revome from the event list
  \return 0: event deleted; -1: otherwise
  \brief Release the 'e' event from the list of slot 's'
 
         Allws the user to delete a concrete event from the list
	 of events associated to the slot 's'. The system won't be
	 informed of changes in this slot any more.

	 Example:
	 \code
	 KDM_Release_Event( robot->xpos, MODIFICATION );
	 \endcode
	 It stops paying attention to modification event (changes in the
	 values) of the slot xpos.
*/
#define KDM_Release_Event(s,e) kdm_event_release( (&s),e )

/*!
  \fn   KDM_Set_Event(s,e,i) 
  \param static slot s slot whose event list we want to change
  \param event e final event value for the event (logical-or of 
         the desired events)
  \param int i iportance associated to all the events for the slot
  \return  0: event list set; -1: otherwise
  \brief Set the events of interest list exactly to 'e'

         In order to avoid declaring all the events one by one,
	 this function has been defined. It allows us to define
	 the set of events shot by the KDM. The user must express
	 the events using an 'or' operator to add all the events
	 together.

	 Example:
	 \code
	 KDM_Set_Event( robot->xpos, MODIFICATION | DELETION, 3 );
	 \endcode
	 It sets the list of events associated to the slot xpos
	 of the instance robot.

	 IMPORTANT:
	 if you include he NO_EVENT flag in the list, no event
	 it will be defined at all.
*/
#define KDM_Set_Event(s,e,i) kdm_event_set( (&s),e,i )


#endif //_KDMUSER_H_
