/**  
 *  \file include/meta_rule.h
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 *  \version 2.0
 *
 *  \brief Meta-Rule Creator 2.0 - Public Interface
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */

#ifndef META_RULE_H
#define META_RULE_H

#include "../meta_rule/include/meta_rule_actions.h"

///#include "meta-rules_quantities.h"
///#include "meta-rules_structs.h"
///#include "../events/events.h"   
    /*** It is basically for the MAX_EV_TYPES constant ***/

///#define META_RULE_STORE "MetaRuleStore"         /*** Id for the mbuff ***/


///#define NMAXMETARULES 10   
    /*** Max. number of meta-rules: It indicates not only the max. number of 
     *  meta-rules, but also the number of regEventTypes and regpSlot
     ***/
 

#define METARULE_OK 1


///////////////////////////////////////////////////////////////////////////////
/// Private Functions 
///////////////////////////////////////////////////////////////////////////////
/** 
 * \defgroup Private Functions 
 */

//@{

/**
 * \fn Meta_Rule_Store_Initialize
 * 
 * \brief Generic initializing function -> DS and RTOS levels
 */
extern int Meta_Rule_Store_Initialize(void); 

/**
 * \fn Meta_Rule_Get_Empty_Att_Focus_Node
 *
 * \brief Returns two possible values:
 *      -1, if there is no empty regActiveAttFocusNode
 *      n, where aregEmptyAttFocusNode[n] is the empty 
 *         regActiveAttFocusNode
 */
extern int Meta_Rule_Get_Empty_Att_Focus_Node(void);
 
/**
 * \fn Meta_Rule_Get_Empty_pMetaRuleNode
 *
 * \brief Returns two possible values:
 *      -1, if there is no empty regpMetaRuleNode
 *      n, where aregEmptypMetaRuleNode[n] is the empty regpMetaRuleNode
 *         and aregEmptyMetaRule[n] is the empty regMetaRule
 */
extern int Meta_Rule_Get_Empty_pMetaRuleNode(void);

/**
 * \fn Meta_Rule_Get_Empty_Event_Type
 *
 * \brief Returns two possible values:
 *      -1, if there is no empty regEventType
 *      n, where aregEmptyEventType[n] is the empty regEventType
 */
extern int Meta_Rule_Get_Empty_Event_Type(void);

/**
 * \fn Meta_Rule_Get_Empty_pSlot
 *
 * \brief Returns two possible values:
 *      -1, if there is no empty regpSlot
 *      n, where aregEmptypSlot[n] is the empty regpSlot
 */
extern int Meta_Rule_Get_Empty_pSlot(void);
     
/**
 * \fn Meta_Rule_Get_Empty_Cond_Node
 *
 * \brief Returns two possible values:
 *      -1, if there is no empty regExprNode
 *      n, where aregEmptyCondNode[n] is the empty regExprNode
 */
extern int Meta_Rule_Get_Empty_Cond_Node(void);
   
/**
 * \fn Meta_Rule_Get_Empty_Act_Node
 *
 * \brief Returns two possible values:
 *      -1, if there is no empty regActNode
 *      n, where aregEmptyActNode[n] is the empty regActNode
 */
extern  int Meta_Rule_Get_Empty_Act_Node(void);

//@}


///////////////////////////////////////////////////////////////////////////////
/// Public Interface Functions 
///////////////////////////////////////////////////////////////////////////////
/** 
 * \defgroup Public Interface Functions
 */

//@{

/**
 * \fn  Meta_Rule_AF_Store_Creation
 *
 * \brief Attention Focus Store (or Meta-Rule Store) initializing function
 *
 *
 *  Creation of the Attention Focus Store (or Meta-Rule Store). 
 *  This function needs as arguments:
 *    - nEventTypes: Number of different event types used in the Meta-Rules 
 *                   defined by the designer.
 *    - nExprUsed: Number of different expression nodes used in the 
 *                   Meta-Rules defined by the designer.
 *    - nActUsed: Number of different action nodes used in the Meta-Rules
 *                   defined by the designer.
 *    - nMetaRules: Number of different Meta-Rules defined by the designer.
 *    - nAttFocus: Number of different Attention Focuses defined by the designer.
 *    - nActiveFocus: Number of different Active Attention Focuses defined by the designer.
 *  
 */
extern int Meta_Rule_AF_Store_Creation ( int nEventTypes, int nExprUsed, int nActUsed,
                                         int nMetaRules, int nAttFocus, int nActiveFocus );

/**
 * \fn Meta_Rule_Store_Free
 *
 * It Releases the memory allocated or attached to the Meta-Rule Store.
 * It must be used both in the RT and non-RT Linux worlds
 */
extern void Meta_Rule_Store_Free(void);


extern int Meta_Rule_init_module(void);

extern int Meta_Rule_cleanup_module(void);

/**
 * \fn Meta_Rule_Create_Att_Focus
 * 
 * \brief Reserves an attention focus node, and return its id or (-1) if error.
 */
extern int Meta_Rule_Create_Att_Focus(void);

/**
 * \fn Meta_Rule_Create_Expr
 * 
 * \brief Reserves an expr (cond) node, and returns its id or (-1) if error.
 *
 * It receives 5 arguments:
 *    - nOp: Operation code.
 *    - nType1: Type of the first factor.
 *    - nType2: Type of the second factor.
 *    - First factor.
 *    - Second factor.
 */
extern int Meta_Rule_Create_Expr(int nOp, int nType1, int nType2, ...);

/**
 * \fn Meta_Rule_Create_Act
 * 
 * \brief Reserves an act node, and returns its id or (-1) if error.
 */
extern int Meta_Rule_Create_Act(int nAct, ...);

/**
 * \fn Meta_Rule_Activate_Attention_Focus
 *
 * Creates a new Active Attention Focus with the given parameters and creates
 *  a new element in the list, placing it at the head of the list 
 *  Returns three possible values:
 *      -2, if the mutex is occupied 
 *      -1, if there is no empty regActiveAttFocusNode
 *      0, The new active attention focus has been inserted
 * The arguments are:
 *      - nAF: Index of the regAttentionFocus
 *      - fAttentDegree: Attention Degree of the Attention Focus to activate 
 */ 
extern int Meta_Rule_Activate_Attention_Focus( int nAF, float fAttentDegree );

/**
 * \fn Meta_Rule_Deactivate_Attention_Focus
 *
 * Destroys the Active Attention Focus corresponding to the attention 
 *  focus indicated by nAF 
 *  Returns three possible values:
 *      -2, if the mutex is occupied 
 *      -1, if there is no Active Attention Focus corresponding to nAF
 *      0, The active attention focus has been destroyed
 * The argument is nAF, the Index of the regAttentionFocus
 */
extern int Meta_Rule_Deactivate_Attention_Focus( int nAF );

/**
 * \fn Meta_Rule_Create
 * 
 * \brief Creates a new Meta-Rule node with the given parameters.
 *
 *  Returns three possible values:
 *        -2, if the mutex is occupied
 *        -1, if there is no empty regpMetaRuleNode
 *         n, where aregEmptypMetaRuleNode[n] has the Meta-Rule pointer, and
 *                   aregMetaRule[n] has the corresponding Meta-Rule 
 * The arguments are:
 *      - nET: Event Type (usually MODIFICATION, but who knows...
 *      - pSlot: Pointer to the slot
 *      - nCond: Index of the regExprNode
 *      - nAct: Index of the regActNode 
 */
extern int Meta_Rule_Create ( int nET, void * pSlot, int nCond, int nAct );

/**
 * \fn Meta_Rule_Insert_Cond
 *
 * \brief Inserts the expr. identified by nCond as part of the condition of 
 *        the meta-rule identified by nMetarule (in the last position).
 */
extern int Meta_Rule_Insert_Cond ( int nCond, int nMetaRule );

/** 
 * \fn Meta_Rule_Insert_Act
 *
 * \brief Inserts the action identified by nAct as part of the actions of
 *        the meta-rule identified by nMetaRule (in the last position).
 */
extern int Meta_Rule_Insert_Act ( int nAct, int nMetaRule );

/**
 * \fn Meta_Rule_Insert_In_Att_Focus
 *
 * Inserts a Meta-Rule node within the Attention Focus indicated.
 *  Returns three possible values:
 *      -2, if the mutex is occupied 
 *      -1, if the mutex is occupied 
 *      0, The Meta-Rule node has been inserted
 * The arguments are:
 *      - nMetarule, where aregEmptypMetaRuleNode[nMetaRule] has the Meta-Rule 
 *                   pointer, and aregMetaRule[nMetaRule] has the 
 *                   corresponding Meta-Rule  
 *      - nAF: Index of the regAttentionFocus
 */
extern int Meta_Rule_Insert_In_Att_Focus( int nMetaRule, int nAF );

/**
 * \fn Meta_Rule_Print_Active_Att_Focus
 *
 * \brief This function allows to print the contents of the active attention 
 *        focuses in the console (RT-LINUX) or in the error standard output (LINUX).
 */
extern void Meta_Rule_Print_Active_Att_Focus(void);


//@}

/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */

#endif
