/*

   FILE:
	init.c

   Copyright (C) 1998-2004
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-16

  CHANGELOG:


  DESCRIPTION:

  The INIT module contains the initialization/cleanup of the different
  ARTIS modules, in either RT-Linux or Linux, which are *independent*
  of the application (that is, they contain the same actions for every
  possible application). 

  In the case of RT-Linux, such functions will be the actual
  "init_module" and "cleanup_module" of the RT-Linux ARTIS main module
  (rtl_artis.o), and thus these functions are not *exported* to any
  other module (in fact, it's the other way around). In the case of
  Linux, these functions are included in the "linux_side" and then
  called at the initialization / ending phases of the process.

  It is important to note the *order* at which all the initialization
  and cleanup functions of each ARTIS module are called, since this
  order denotes the dependencies between such modules.

  This file is compiled twice, one as a kernel module and another one
  as a regular Linux object file

*/


// #include (ALL the modules whose init/cleanup is done here!!)



#ifdef MODULE


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>

#include "../../common/include/artis_conf.h"

#include "../../events/include/events.h"
#include "../../rserver/include/rserver.h"
#include "../../mutex/include/mutex.h"
#include "../../trace/include/trace.h"
#include "../../storage/include/storage.h"
#include "../../include/meta_rule.h"

MODULE_LICENSE("GPL");

int init_module (void) {
  int retvalue;
  /* 
     Calls to the "init_module" functions of all other ARTIS RT-Linux
     modules (IN ORDER)
  */

  if ((retvalue = Events_Initialize()) < 0) {
    rtl_printf("(rtl-init) Init. of module 'Events' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Init. of module 'Events' OK\n");
#endif
	    
  if ((retvalue = rserver_init_module()) < 0) {
    rtl_printf("(rtl-init) Init. of module 'rserver' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Init. of module 'rserver' OK\n");
#endif
	    
  if ((retvalue = mutex_init_module()) < 0) {
    rtl_printf("(rtl-init) Init. of module 'mutex' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Init. of module 'mutex' OK\n");
#endif

  if ((retvalue = storage_init_module()) < 0) {
    rtl_printf("(rtl-init) Init. of module 'storage' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Init. of module 'storage' OK\n");
#endif

  if ((retvalue = kdm_init_module()) < 0) {
    rtl_printf("(rtl-init) Init. of module 'KDM' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Init. of module 'KDM' OK\n");
#endif

  /***
  if ((retvalue = metaclass_init_module()) < 0) {
    rtl_printf("(rtl-init) Init. of module 'Meta-KDM' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Init. of module 'Meta-KDM' OK\n");
#endif
  ***/

  if ((retvalue = trace_init_module()) < 0) {
    rtl_printf("(rtl-init) Init. of module 'trace' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Init. of module 'trace' OK\n");
#endif

  if ((retvalue = Meta_Rule_init_module()) < 0) {
    rtl_printf("(rtl-init) Init. of module 'Meta-Rule' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Init. of module 'Meta-Rule' OK\n");
#endif

  if ((retvalue = MR_Action_init_module()) < 0) {
    rtl_printf("(rtl-init) Init. of module 'Meta-Rule Action' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Init. of module 'Meta-Rule Action' OK\n");
#endif

  // etc.

  return 0;
  
}


void cleanup_module (void) {
  /* 
     Calls to the "cleanup_module" functions of all other ARTIS
     RT-Linux modules (IN ORDER)
  */

#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Beginning of Init: cleanup module\n");
#endif

  trace_cleanup_module();
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Cleanup of module 'trace' OK\n");
#endif

  /***
  metaclass_cleanup_module();
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Cleanup of module 'Meta-KDM' OK\n");
#endif
  ***/

  kdm_cleanup_module();
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Cleanup of module 'KDM' OK\n");
#endif

  storage_cleanup_module();
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Cleanup of module 'storage' OK\n");
#endif

  mutex_cleanup_module();
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Cleanup of module 'mutex' OK\n");
#endif

  rserver_cleanup_module();
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Cleanup of module 'rserver' OK\n");
#endif

  Events_Free();
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Cleanup of module 'Events' OK\n");
#endif

  Meta_Rule_cleanup_module();
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Cleanup of module 'Meta-Rule' OK\n");
#endif

  MR_Action_cleanup_module();
#ifdef ARTIS_INIT_DEBUG
  rtl_printf("(rtl-init) Cleanup of module 'Meta-Rule Actions' OK\n");
#endif

  /// etc.
  
}


#else // MODULE

#define ARTIS_INIT_DEBUG

#include <stdio.h>

#include "../../common/include/artis_conf.h"

#include "../../trace/include/trace.h"
#include "../../linux_core/include/linux_core.h"
#include "../../events/include/events.h"
#include "../../mutex/include/mutex.h"

int linux_artis_init (void) {
  /* 
     Calls to the "init" functions of all other ARTIS Linux modules
     (IN ORDER)
  */

  int retvalue;
  
  if ((retvalue = Events_Initialize()) < 0) {
    fprintf(stderr, "(linux-init) Init. of 'Events' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
  
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Init. of 'Events' OK\n");
#endif 

  if ((retvalue = Linux_Core_Init()) < 0) {
    fprintf(stderr, "(linux-init) Init. of 'Linux Core' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
  
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Init. of 'Linux Core' OK\n");
#endif 

  if ((retvalue = mutex_init_module()) < 0) {
    fprintf(stderr, "(linux-init) Init. of 'Mutex' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
  
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Init. of 'Mutex' OK\n");
#endif 

  if ((retvalue = kdm_init()) < 0) {
    fprintf(stderr, "(linux-init) Init. of 'KDM' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
  
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Init. of 'KDM' OK\n");
#endif 


  /***
  if ((retvalue = meta_kdm_create()) < 0) {
    fprintf(stderr, "(linux-init) Init. of 'Meta-KDM' failed, error %d\n",
	       retvalue);
    return retvalue;
  }
  
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Init. of 'Meta-KDM' OK\n");
#endif 
  ***/

  
/// Here we initialize CoMo if implemented
#ifdef COMO
 #ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Before Init. of 'COMO'\n");
 #endif 
   como_init();
 #ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Init. of 'COMO' OK\n");
 #endif 
#endif
  

  if ((retvalue = trace_init("linuxapp")) < 0) {
    fprintf(stderr, "(linux-init) Init. of 'trace' failed, error %d\n",
	       retvalue);
    return retvalue;
  }

  trace_start();

#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Init. of 'Trace' OK\n");
#endif

  /// etc.

}


int linux_artis_cleanup (void) {

  /* 
     Calls to the "cleanup" functions of all other ARTIS Linux modules
     (IN ORDER)
  */

  trace_cleanup();
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Cleanup of 'Trace' OK\n");
#endif

  /***
  meta_kdm_destroy();
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Cleanup of 'Meta-KDM' OK\n");
#endif
  ***/

  kdm_clean();
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Cleanup of 'KDM' OK\n");
#endif

  mutex_cleanup_module();
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Cleanup of 'Mutex' OK\n");
#endif

  Events_Free();
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Cleanup of 'Events' OK\n");
#endif

  trace_stop();
  trace_cleanup();
#ifdef ARTIS_INIT_DEBUG
  fprintf(stderr, "(linux-init) Cleanup of 'trace' OK\n");
#endif

  /// etc.

}


#endif // MODULE
