/*!
  \file static.c
  \date Time-stamp: <16-03-2005 00:54>
  \author M. Rebollo <mrebollo@dsic.upv.es>
  \version 0.3

  \brief internal access funtions to static information
*/

#include <assert.h>
#include <stdlib.h>
#include <stdarg.h>
#include "../include/kdm/globals.h"
#include "../include/kdm/static.h"

#include "../../common/include/artis_conf.h"
#include "../../events/include/events.h"
#include "../../storage/include/storage.h"
#include "../../include/mutex.h"


kdmtype (*static_get[KDM_PRIMITIVE_TYPES])(const static_slot *slot );


//---------------------------------------------------
// inline internal functions
//--------------------------------------------------- 

/*!
  \fn kdmtype static_get_char( const static_slot *slot )
  \param static_slot *slot slot to recover
  \brief access function to get character values
*/
inline kdmtype 
static_get_char( const static_slot *slot ){ return (kdmtype)&slot->value.c; }
/*!
  \fn kdmtype static_get_int( const static_slot *slot )
  \param static_slot *slot slot to recover
  \brief access function to get integer values
*/
inline kdmtype 
static_get_int( const static_slot *slot ){ return (kdmtype)&slot->value.i; }
/*!
  \fn kdmtype static_get_uint( const static_slot *slot )
  \param static_slot *slot slot to recover
  \brief access function to get unsigned integer values
*/
inline kdmtype 
static_get_uint( const static_slot *slot ){ return (kdmtype)&slot->value.ui; }
/*!
  \fn kdmtype static_get_long( const static_slot *slot )
  \param static_slot *slot slot to recover
  \brief access function to get long integer values
*/
inline kdmtype 
static_get_long( const static_slot *slot ){ return (kdmtype)&slot->value.l; }
/*!
  \fn kdmtype static_get_ulong( const static_slot *slot )
  \param static_slot *slot slot to recover
  \brief access function to get unsigned long integer values
*/
inline kdmtype 
static_get_ulong( const static_slot *slot ){ return (kdmtype)&slot->value.ul; }

/*
    MRP: Till it works
  
inline kdmtype 
static_get_int64( const static_slot *slot ){ return (kdmtype)&slot->value.i64; }
inline kdmtype 
static_get_uint64( const static_slot *slot ){ return (kdmtype)&slot->value.i64; } 

*/

/*!
  \fn kdmtype static_get_float( const static_slot *slot )
  \param static_slot *slot slot to recover
  \brief access function to get float values
*/
inline kdmtype 
static_get_float( const static_slot *slot ){ return (kdmtype)&slot->value.f; }
/*!
  \fn kdmtype static_get_double( const static_slot *slot )
  \param static_slot *slot slot to recover
  \brief access function to get double values
*/
inline kdmtype 
static_get_double( const static_slot *slot ){ return (kdmtype)&slot->value.d; }

/*!
  \fn kdmtype static_get_pointer( const static_slot *slot )
  \param static_slot *slot slot to recover
  \brief access function to get pointer values

  The Storage only defines one memory space, but with to 
  different accessing points: one from the rt-linux side (critical
  tasks) an the other from the linux side (optional tasks). And each
  side address this space using a different address range.

  If we use pointers, they must be relative addresses to the initial
  address of the whole memory block, and they must be converted to 
  a rt-linux or a linux address, depending on the part of the code 
  in which this function is called.
*/
inline kdmtype 
static_get_pointer( const static_slot *slot ){
	/*
#ifdef MODULE
  rtl_printf("dir: %p %p\n", tbn.stdata, slot->value.pt); 
#else
  printf("dir: %p %p\n", tbn.stdata, slot->value.pt); 
#endif
*/
 return (kdmtype)Storage_Get_Address(tbn.stdata, slot->value.pt);
//return (kdmtype) ((unsigned int)tbn.stdata + (unsigned int)slot->value.pt);
}

/*
inline void 
static_set_char( static_slot *slot, kdmchar value );
inline void 
static_set_int( static_slot *slot, kdmint value );
inline void 
static_set_uint( static_slot *slot, kdmuint value );
inline void 
static_set_long( static_slot *slot, kdmlong value );
inline void 
static_set_ulong( static_slot *slot, kdmulong value );
inline void 
static_set_float( static_slot *slot, kdmfloat value );
inline void 
static_set_double( static_slot *slot, kdmdouble value );
inline void 
static_set_array( static_slot *slot, kdmpointer pvalue );
*/



int
//---------------------------------------------------
static_set_slot_value( static_slot *slot,... )
//---------------------------------------------------
{
  va_list ap;

  if ( slot == NULL ) return -1;
 
#ifndef MODULE
  ///mutex_lock( MUTEX_CONTROL );
#else 
///  mutex_system_lock( MUTEX_CONTROL, WCET_KDM_ACCESS );
#endif

  va_start( ap, slot );
  switch( slot->type ){
  case KDM_INT:
    if (slot->event) {
      KDM_Value ValueOld = slot->value;
      slot->value.i = va_arg( ap, int );
      Event_Notify( slot->event, Storage_Get_Relative( tbn.stdata, slot ),  
		    slot->imp, slot->type, 
		    (long long) 0, slot->value, 
		    (long long) 0, ValueOld );
    }
    else slot->value.i = va_arg( ap, int );
    break;
  case KDM_FLOAT:
    if (slot->event) {
      KDM_Value ValueOld = slot->value;
      slot->value.f = va_arg( ap, double );
      Event_Notify( slot->event, Storage_Get_Relative( tbn.stdata, slot ),  
		    slot->imp, slot->type, 
		    (long long) 0, slot->value, 
		    (long long) 0, ValueOld );
    }
    else slot->value.f = va_arg( ap, double );
    break;
  case KDM_CHAR:
    if (slot->event) {
      KDM_Value ValueOld = slot->value;
      slot->value.c = (char) va_arg( ap, int );
      Event_Notify( slot->event, Storage_Get_Relative( tbn.stdata, slot ), 
		    slot->imp, slot->type, 
		    (long long) 0, slot->value, 
		    (long long) 0, ValueOld );
    }
    else slot->value.c = (char)va_arg( ap, int ); //char is promoted to int
    break;
  default:
    va_end( ap );
    ///mutex_unlock( MUTEX_CONTROL );
    return -1;
  } //switch 

  va_end( ap );
  ///mutex_unlock( MUTEX_CONTROL );
  return 0;

} //_kdm_set_slot_value()



int 
//---------------------------------------------------
static_register_access( void )
//---------------------------------------------------
{
  /* Links the access function with each datatype */
  if( (static_get[KDM_CHAR] = &static_get_char ) == NULL ) return -1;
  if( (static_get[KDM_INT] = &static_get_int ) == NULL ) return -1;
  if( (static_get[KDM_UINT] = &static_get_uint ) == NULL ) return -1;
  if( (static_get[KDM_LONG] = &static_get_long ) == NULL ) return -1;
  if( (static_get[KDM_ULONG] = &static_get_ulong ) == NULL ) return -1;
  /*
    MRP: Till it works
  if( (static_get[KDM_INT64] = &static_get_int64 ) == NULL ) return -1;
  if( (static_get[KDM_UINT64] = &static_get_uint64 ) == NULL ) return -1;
  */
  if( (static_get[KDM_FLOAT] = &static_get_float ) == NULL ) return -1;
  if( (static_get[KDM_DOUBLE] = &static_get_double ) == NULL ) return -1;
  if( (static_get[KDM_PTR] = &static_get_pointer ) == NULL ) return -1;
  if( (static_get[KDM_ARRAY] = &static_get_pointer ) == NULL ) return -1;

  return 0;
} //static_register_access()
