
/*

   FILE:
	linux_core.h						

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*


  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-06

  CHANGELOG:


  DESCRIPTION:

    The LINUX_CORE module provides a set of support services related
    to managing time and threads, in order to be used by other modules
    in the linux_side process (e.g., the DSERVER and MUTEX
    modules).These facilities provides a higher-level layer at the top
    of the Linux system calls (some POSIX, some not), specifically
    designed to the needs of ARTIS.
    

    In particular, this module exports a set of functions related to
    the management of::


    (1) Levels. A "level" is the minimum unit of execution in ARTIS
    and, in this implementation, each level is run by its own
    thread. The level abstraction is provided by a "level" (level_t)
    type and set of functions in order to create, execute, stop and
    kill a level.


*/


#ifndef __LINUX_CORE_H
#define __LINUX_CORE_H

#include "../../time/include/time.h"




////////////////////////////////////////////////////////
// Visible Constants
////////////////////////////////////////////////////////

#define LEVEL_ENDED      0
#define LEVEL_TIMEOUT    1
#define LEVEL_SUSPENDED  2

#define LINUX_CORE_SUCCESS  0
#define LINUX_CORE_FAILURE -1




////////////////////////////////////////////////////////
// Visible types
////////////////////////////////////////////////////////

typedef int  level_t;



////////////////////////////////////////////////////////
// Visible functions:
////////////////////////////////////////////////////////


// All these functions returns an error code indicating
// LINUX_CORE_SUCCESS or LINUX_CORE_FAILURE


int
Level_Create (level_t   Level_Id, 
	      void    (*Func)(void *),
	      void *    Args);

//  Function "Level_Create" Creates a new level



int  
Level_Execute_Until (level_t    Level_Id,
		     systime_t  Delay,
		     int        quivi_entity);

//  Function "Level_Execute_Until" is used by the DSERVER module to
//  execute a level during some time. This function returns three
//  alternative values: 

//    + LEVEL_ENDED,     if the level has ended before the timeout.
//    + LEVEL_TIMEOUT,   if the timeout has expired before the 
//                       level has ended.
//    + LEVEL_SUSPENDED, if the level has suspended itself when
//                       trying to acquire a resource 
//                       (e.g., inside a 'mutex_lock' operation.)



int  
Level_Regenerate (level_t Level_Id, int quivi_entity);

//  Function "Level_Regenerate" is used by the DSERVER module to
//  kill a level and to create it again.



int
Level_Suspend_Self (void);

//  Function "Level_Suspend_Self" is called by the execution level in
//  order to suspend its execution. This action causes that funcion
//  Level_Execute_Until returns LEVEL_SUSPENDED.


int
Level_Destroy (level_t Level_Id);

// Function used to destroy a level at the end of application


int
Linux_Core_Init (void);

//  Function Linux_Core_Init initializes this module




////////////////////////////////////////////////////////
// Visible (global) variables:
////////////////////////////////////////////////////////


//    When the DSERVER executes a level (by calling "execute_until"), it
//    always puts a timeout on the level execution. If the level calls to
//    lock a mutex, the lenght of the critical section cannot exceed this timeout.
//    For this reason, the LINUX_CORE module has to export either a global
//    variable containing (or a function returning) this value, which has to
//    be updated inside "execute_until".
//    This variable is the absolute time of the timeout.

systime_t current_level_abs_timeout;


#endif // __LINUX_CORE_H
