#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>

#include "../../common/include/artis_conf.h"
#include "../include/linux_core.h"
#include "../../time/include/time.h"
#include "../../common/include/artis_event_definition.h"

// #include "../../include/rt_trace.h"   ///


////////////////////////////////////////////////////////
// CONSTANTS
////////////////////////////////////////////////////////

#define MAX_LEVELS  100

#define LEVEL_IS_WAITING      0
#define LEVEL_IS_STOPPED      1
#define LEVEL_IS_RUNNING      2
#define LEVEL_IS_NEW          3
#define LEVEL_IS_ABORTED      4
#define LEVEL_IS_SUSPENDED    5

#define NO_LEVEL          -1

#define DEFAULT_PRIORITY 99 ///10
#define MAIN_THREAD_PRIORITY 99 ///12
 
#define SIGNAL_EXECUTE    SIGUSR1
#define SIGNAL_STOP       SIGUSR2
#define SIGNAL_ABORT      SIGTERM
#define SIGNAL_LEVEL_END  SIGHUP
#define SIGNAL_TIMEOUT    SIGALRM


#define TRUE  1
#define FALSE 0

#ifdef ARTIS_TRACE_EVENTS
#define LINUX_CORE_TRACE
#endif
#define LINUX_CORE_DEBUG

////////////////////////////////////////////////////////
// GLOBALS
////////////////////////////////////////////////////////


int Linux_Core_Initialized = FALSE;

pthread_attr_t       Default_Thread_Attributes;
struct sched_param   Default_Sched_Parameters;

pthread_t  Main_Thread;
sigset_t   Main_Thread_Mask_When_Waiting;
sigset_t   Level_Mask_When_Waiting;
sigset_t   Level_Normal_Mask;

pthread_mutex_t    Global_Mutex = PTHREAD_MUTEX_INITIALIZER;

struct Level_Descriptor {
  void         (*Func)(void *);
  void *         Args;
  int            Status;
  int            Id;
  pthread_t      Thread;
  sigjmp_buf     Context;
} Level[MAX_LEVELS]; 

int Current_Level = NO_LEVEL;


/// System-time related variables
/***
systime_t         auxiliar; ///as we can not use many local vbles.
systime_t         current_level_kill_time;
systime_t         now;
***/

////////////////////////////////////////////////////////
// INTERNAL FUNCTIONS
////////////////////////////////////////////////////////


static struct itimerval stop_timer;


int timer_start(systime_t delay) {

	struct itimerval timer_delay;

	/* set up the variable to program the timer */
	timer_delay.it_value = systime_time_to_timeval( delay );
	timer_delay.it_interval.tv_sec  = 0L;
	timer_delay.it_interval.tv_usec = 0L;

	/* OK. Program it */
	return setitimer(ITIMER_REAL,&timer_delay, NULL);
}





int timer_stop(void) {
        /* as any static variable, stop_timer is set to 0s */
	return setitimer(ITIMER_REAL,&stop_timer, NULL);
}


////////////////////////////////////////////////////////
// Level_Kill
////////////////////////////////////////////////////////


static int  
Level_Kill (level_t Level_Id, int quivi_entity) {

  int res;

#ifdef LINUX_CORE_DEBUG
  systime_t time_aux, now;

  fprintf(stderr,"\t(linux_core::Level_Kill) gonna kill level %d\n",
          Level_Id);
#endif

#ifdef LINUX_CORE_TRACE
  trace_event( LEVEL_RESET, quivi_entity, 0, 0);
#endif

  if (Level [Level_Id].Id == NO_LEVEL) {
#ifdef LINUX_CORE_DEBUG
    fprintf(stderr,"\t(linux_core::Level_Kill) Level %d doesn't exist\n",
          Level_Id);
#endif
    return LINUX_CORE_FAILURE;
  }

  if (Level [Level_Id].Status != LEVEL_IS_STOPPED &&
      Level [Level_Id].Status != LEVEL_IS_SUSPENDED) {
    // Reset is not necessary
#ifdef LINUX_CORE_DEBUG
    fprintf(stderr,"\t(linux_core::Level_Kill) Level %d isn't paused\n",
          Level_Id);
#endif
    return LINUX_CORE_SUCCESS;
  }

  Level [Level_Id].Status = LEVEL_IS_ABORTED;

#ifdef LINUX_CORE_DEBUG
  time_aux = systime_gettime();
#endif

  res = pthread_cancel (Level [Level_Id].Thread);

#ifdef LINUX_CORE_DEBUG
  now = systime_gettime();
  fprintf(stderr,"\t(linux_core::Level_Kill) Before the end %lld\n",
	  now - time_aux);
#endif

  return LINUX_CORE_SUCCESS;
}





////////////////////////////////////////////////////////
// SIGNAL_TIMEOUT_Handler
////////////////////////////////////////////////////////

static void SIGNAL_TIMEOUT_Handler (int signal) {
  // Nothing to do
}




/////////////////////////////////////////////////////////
// SIGNAL_EXECUTE_Handler
////////////////////////////////////////////////////////

static void SIGNAL_EXECUTE_Handler (int signal) {
  // Nothing to do 
}




////////////////////////////////////////////////////////
// SIGNAL_STOP_Handler
////////////////////////////////////////////////////////

static void SIGNAL_STOP_Handler (int signal) {

  int Self_Id = Current_Level;
  fprintf (stderr, "\n(linux_core:SIGNAL_STOP_Handler) Me paran %d\n", 
	Self_Id);

  pthread_kill (Main_Thread, SIGNAL_LEVEL_END);
  sigsuspend (&Level_Mask_When_Waiting);

  if ( Level [Self_Id].Status == LEVEL_IS_ABORTED ) {
#ifdef LINUX_CORE_DEBUG
    fprintf(stderr, "(linux_core:SIGNAL_STOP_HANDLER) Antes del siglongjmp\n");
#endif
    siglongjmp ( Level [Self_Id].Context, 1 );
  }
  
}




/////////////////////////////////////////////////////////
// SIGNAL_ABORT_Handler
////////////////////////////////////////////////////////

static void SIGNAL_ABORT_Handler (int signal) {
  // Nothing to do
}




//////////////////////////////////////////////////////
// SIGNAL_LEVEL_END_Handler
////////////////////////////////////////////////////////

static void SIGNAL_LEVEL_END_Handler (int signal) {
  // Nothing to do
}





////////////////////////////////////////////////////////
// Level_Runner
////////////////////////////////////////////////////////

static void * 
Level_Runner (void * Level_Id_Ptr) 
{
  
  int Level_Id = *((int *) Level_Id_Ptr);
  

  while (1) {
    if (sigsetjmp (Level [Level_Id].Context, 1) != 0) {
      
      // Thread is restarted. Jump again to the sentence above
  
    } else {
      
      pthread_sigmask (SIG_SETMASK, &Level_Normal_Mask, NULL);
      
      Level [Level_Id].Status = LEVEL_IS_WAITING;
#ifdef LINUX_CORE_DEBUG
      fprintf( stderr, "(linux_core:Level_Runner) Antes del pthread_kill\n");
#endif
      pthread_kill (Main_Thread, SIGNAL_LEVEL_END);
      
      while (1) {
		
	sigsuspend (&Level_Mask_When_Waiting);
	
	(Level [Level_Id].Func) (Level [Level_Id].Args);
	
	pthread_mutex_lock (&Global_Mutex);
	{
	  Level [Level_Id].Status = LEVEL_IS_WAITING;
	  pthread_kill (Main_Thread, SIGNAL_LEVEL_END);
	}
	pthread_mutex_unlock (&Global_Mutex);
      }
    }
    
  }
  return NULL; 

}


////////////////////////////////////////////////////////
// INTERFACES
////////////////////////////////////////////////////////




////////////////////////////////////////////////////////
// Level_Create
////////////////////////////////////////////////////////

int
Level_Create (level_t      Level_Id, 
	      void       (*Func)(void *),
	      void *       Args)
{
  
  int res;

#ifdef LINUX_CORE_DEBUG
  fprintf(stderr,"                 -----------------------             \n");
  fprintf(stderr,"              -- LEVEL %d INITIALIZATION --             \n",
          Level_Id);
  fprintf(stderr,"                 -----------------------         \n");
#endif
  
  if (Linux_Core_Initialized == FALSE ) {

#ifdef LINUX_CORE_DEBUG
    fprintf(stderr,"(linux_core)ERROR inicializando una tarea con id=0\n");
#endif

    return LINUX_CORE_FAILURE;
  }

  if (Level_Id < 0 || Level_Id > MAX_LEVELS) {

#ifdef LINUX_CORE_DEBUG
    fprintf(stderr,"(linux_core)ERROR inicializando una tarea con id=0\n");
#endif

    return LINUX_CORE_FAILURE;
  }
  
  if (Level [Level_Id].Id != NO_LEVEL) {

#ifdef LINUX_CORE_DEBUG
    fprintf(stderr,"(linux_core)ERROR inicializando una tarea con id=0\n");
#endif

    return LINUX_CORE_FAILURE;
  }
    
  if (Func == NULL) {

#ifdef LINUX_CORE_DEBUG
    fprintf(stderr,"(linux_core)ERROR inicializando una tarea con id=0\n");
#endif

    return LINUX_CORE_FAILURE;
  }
  
  Level [Level_Id].Id      = Level_Id;
  Level [Level_Id].Func    = Func;
  Level [Level_Id].Args    = Args;
  Level [Level_Id].Status  = LEVEL_IS_NEW;
 
  res = pthread_create (&(Level [Level_Id].Thread), 
			&Default_Thread_Attributes, 
			Level_Runner, &(Level [Level_Id].Id));
  
  if ( res != 0) {
    return LINUX_CORE_FAILURE;
  }

  pthread_setschedparam (Level [Level_Id].Thread, SCHED_FIFO, &Default_Sched_Parameters);

  // Main thread waits until new level is LEVEL_IS_WAITING
  
  sigsuspend (&Main_Thread_Mask_When_Waiting);

#ifdef LINUX_CORE_DEBUG
  fprintf(stderr,"\\------------------------------------------------------------------------------/\n\n");
#endif  
  
  return LINUX_CORE_SUCCESS;
}




////////////////////////////////////////////////////////
// Level_Execute_Until
////////////////////////////////////////////////////////

int
Level_Execute_Until (level_t    Level_Id, 
		     systime_t  Delay, 
		     int        quivi_entity)
{
  int Result;
  
#ifdef LINUX_CORE_DEBUG
  systime_t time_aux, now;
#endif

  if (Level [Level_Id].Id == NO_LEVEL) {
#ifdef LINUX_CORE_DEBUG
    fprintf(stderr, "ERR (linux_core.execute_until) level %d does not exist\n",
	    Level_Id);
#endif
    return LINUX_CORE_FAILURE;
  }

  timer_start (Delay);
  current_level_abs_timeout = systime_gettime () + Delay;

#ifdef LINUX_CORE_DEBUG
  fprintf(stderr, "\t(linux_core.execute_until) level: %d, tarea: %d \n", 
	  Level_Id, quivi_entity);
#endif
#ifdef LINUX_CORE_TRACE
  trace_event( LEVEL_START, quivi_entity, 0, 0);
#endif

  Level [Level_Id].Status = LEVEL_IS_RUNNING;
  Current_Level = Level_Id;

  pthread_kill (Level [Level_Id].Thread, SIGNAL_EXECUTE);

#ifdef LINUX_CORE_DEBUG
  time_aux = systime_gettime();
  fprintf(stderr, "(linux_core::execute_until) Before suspending at %lld\n",
	time_aux);
#endif

  sigsuspend (&Main_Thread_Mask_When_Waiting);

#ifdef LINUX_CORE_DEBUG
  now = systime_gettime();
  fprintf(stderr, "(linux_core::execute_until) Time suspended: %lld\n", 
	  now - time_aux);
#endif

  timer_stop ();
  ///time_get_system_time(&now);

  pthread_mutex_lock (&Global_Mutex);
  {
    switch (Level [Level_Id].Status) {

    case LEVEL_IS_RUNNING : {
      pthread_kill (Level [Level_Id].Thread, SIGNAL_STOP);
      Level [Level_Id].Status = LEVEL_IS_STOPPED;

      pthread_mutex_unlock (&Global_Mutex);

      sigsuspend (&Main_Thread_Mask_When_Waiting);
     
      pthread_mutex_lock (&Global_Mutex);
      
#ifdef LINUX_CORE_TRACE
      trace_event( LEVEL_SUSPEND, quivi_entity, 0, 0);
#endif
      Result = LEVEL_TIMEOUT;
      break;
    }
      
    case LEVEL_IS_WAITING : {       
#ifdef LINUX_CORE_TRACE
      trace_event( LEVEL_STOP, quivi_entity, 0, 0);
#endif
      Result = LEVEL_ENDED;
      break;
    }

    case LEVEL_IS_SUSPENDED : {       
#ifdef LINUX_CORE_TRACE
      trace_event( LEVEL_SUSPEND, quivi_entity, 0, 0);
#endif
      Result = LEVEL_SUSPENDED;
      break;
    }
    }
  }
  pthread_mutex_unlock (&Global_Mutex);      
  
  return Result;
}




////////////////////////////////////////////////////////
// Level_Reset
////////////////////////////////////////////////////////

/* int   */
/* Level_Reset (level_t Level_Id, int quivi_entity) { */

/*   int res; */

/* #ifdef LINUX_CORE_DEBUG */
/*   systime_t time_aux, now; */

/*   fprintf(stderr,"\t(linux_core::Level_Reset) gonna reset level %d\n", */
/*           Level_Id); */
/* #endif */

/* #ifdef LINUX_CORE_TRACE */
/*   trace_event( LEVEL_RESET, quivi_entity, 0, 0); */
/* #endif */

/*   if (Level [Level_Id].Id == NO_LEVEL) { */
/* #ifdef LINUX_CORE_DEBUG */
/*     fprintf(stderr,"\t(linux_core::Level_Reset) Level %d doesn't exist\n", */
/*           Level_Id); */
/* #endif */
/*     return LINUX_CORE_FAILURE; */
/*   } */

/*   if (Level [Level_Id].Status != LEVEL_IS_STOPPED && */
/*       Level [Level_Id].Status != LEVEL_IS_SUSPENDED) { */
/*     // Reset is not necessary */
/* #ifdef LINUX_CORE_DEBUG */
/*     fprintf(stderr,"\t(linux_core::Level_Reset) Level %d isn't paused\n", */
/*           Level_Id); */
/* #endif */
/*     return LINUX_CORE_SUCCESS; */
/*   } */

/*   Level [Level_Id].Status = LEVEL_IS_ABORTED; */
/*   res = pthread_kill (Level [Level_Id].Thread, SIGNAL_ABORT); */

/* #ifdef LINUX_CORE_DEBUG */
/*   fprintf(stderr,"\t(linux_core::Level_Reset) Before being suspended\n"); */
/*   time_get_system_time( &time_aux ); */
/* #endif */

/*   sigsuspend (&Main_Thread_Mask_When_Waiting); */
 
/* #ifdef LINUX_CORE_DEBUG */
/*   time_get_system_time( &now ); */
/*   fprintf(stderr,"\t(linux_core::Level_Reset) Before the end %lld\n", */
/* 	  now - time_aux); */
/* #endif */

/*   return LINUX_CORE_SUCCESS; */
/* } */




////////////////////////////////////////////////////////
// Level_Regenerate
////////////////////////////////////////////////////////

int  
Level_Regenerate (level_t Level_Id, int quivi_entity) {

  int res;

#ifdef LINUX_CORE_DEBUG
  systime_t time_aux, now;
#endif

  if ((res = Level_Kill( Level_Id, quivi_entity )) == LINUX_CORE_FAILURE)
    return LINUX_CORE_FAILURE;

  Level [Level_Id].Status  = LEVEL_IS_NEW;
 
  res = pthread_create (&(Level [Level_Id].Thread), 
			&Default_Thread_Attributes, 
			Level_Runner, &(Level [Level_Id].Id));
  
  if ( res != 0) {
    return LINUX_CORE_FAILURE;
  }

  pthread_setschedparam (Level [Level_Id].Thread, SCHED_FIFO, &Default_Sched_Parameters);

  // Main thread waits until new level is LEVEL_IS_WAITING
  
  sigsuspend (&Main_Thread_Mask_When_Waiting);

#ifdef LINUX_CORE_DEBUG
  fprintf(stderr,"\\------------------------------------------------------------------------------/\n\n");
#endif  
  
  return LINUX_CORE_SUCCESS;
}


////////////////////////////////////////////////////////
// Level_Suspend_Self
////////////////////////////////////////////////////////

int
Level_Suspend_Self (void) {

  level_t Self_Id = Current_Level;
  
  pthread_mutex_lock (&Global_Mutex);
  {
    Level [Self_Id].Status = LEVEL_IS_SUSPENDED;
    pthread_kill (Main_Thread, SIGNAL_LEVEL_END);
  }
  pthread_mutex_unlock (&Global_Mutex);

  sigsuspend (&Level_Mask_When_Waiting);

  if ( Level [Self_Id].Status == LEVEL_IS_ABORTED ) {
    siglongjmp ( Level [Self_Id].Context, 1 );
  }
  
  return LINUX_CORE_SUCCESS;
}



////////////////////////////////////////////////////////
// Level_Destroy
////////////////////////////////////////////////////////


int  
Level_Destroy (level_t Level_Id) {

  int res;

#ifdef LINUX_CORE_DEBUG
  systime_t time_aux, now;

  fprintf(stderr,"\t(linux_core::Level_Destroy) gonna destroy level %d\n",
          Level_Id);
#endif

  if (Level [Level_Id].Id == NO_LEVEL) {
#ifdef LINUX_CORE_DEBUG
    fprintf(stderr,"\t(linux_core::Level_Destroy) Level %d doesn't exist\n",
          Level_Id);
#endif
    return LINUX_CORE_FAILURE;
  }

  if (Level [Level_Id].Status != LEVEL_IS_STOPPED &&
      Level [Level_Id].Status != LEVEL_IS_SUSPENDED) {
    // Reset is not necessary
#ifdef LINUX_CORE_DEBUG
    fprintf(stderr,"\t(linux_core::Level_Destroy) Level %d isn't paused\n",
          Level_Id);
#endif
    return LINUX_CORE_SUCCESS;
  }

  Level [Level_Id].Status = LEVEL_IS_ABORTED;

#ifdef LINUX_CORE_DEBUG
  time_aux = systime_gettime();
#endif

  res = pthread_cancel (Level [Level_Id].Thread);
  res = pthread_join (Level [Level_Id].Thread, NULL);

#ifdef LINUX_CORE_DEBUG
  now = systime_gettime();
  fprintf(stderr,"\t(linux_core::Level_Destroy) Before the end %lld\n",
	  now - time_aux);
#endif

  return LINUX_CORE_SUCCESS;
}






////////////////////////////////////////////////////////
// Linux_Core_Init
////////////////////////////////////////////////////////

int
Linux_Core_Init () {

  struct sched_param Main_Thread_Sched_Parameters;
  sigset_t           Main_Thread_Normal_Mask;
  struct sigaction   Signal_Action;
  int                i;
  int                Res;

  // Main_Thread stuff

  Main_Thread = pthread_self ();
  Main_Thread_Sched_Parameters.sched_priority = MAIN_THREAD_PRIORITY;
  Res = pthread_setschedparam (Main_Thread, SCHED_FIFO, &Main_Thread_Sched_Parameters);

  if (Res != 0) {
    // This process is not owned by root
    // The SCHED_FIFO policy is required !!
    return LINUX_CORE_FAILURE;
  }

  //  Signals which are managed synchronously
  //  (SIGNAL_LEVEL_END and SIGNAL_TIMEOUT)
  //  must be masked except when the main thread is waiting
  //  for them in sigsuspend

  pthread_sigmask (SIG_BLOCK, NULL, &Main_Thread_Mask_When_Waiting);
  sigdelset (&Main_Thread_Mask_When_Waiting, SIGNAL_LEVEL_END);
  sigdelset (&Main_Thread_Mask_When_Waiting, SIGNAL_TIMEOUT);

  pthread_sigmask (SIG_BLOCK, NULL, &Main_Thread_Normal_Mask);
  sigaddset (&Main_Thread_Normal_Mask, SIGNAL_LEVEL_END);
  sigaddset (&Main_Thread_Normal_Mask, SIGNAL_TIMEOUT);
  pthread_sigmask (SIG_SETMASK, &Main_Thread_Normal_Mask, NULL);

  // All signals managed by the main thread are handled.

  Signal_Action.sa_handler = SIGNAL_LEVEL_END_Handler;
  Signal_Action.sa_flags = 0;
  sigemptyset (&(Signal_Action.sa_mask));
  sigaction (SIGNAL_LEVEL_END, &Signal_Action, NULL);

  Signal_Action.sa_handler = SIGNAL_TIMEOUT_Handler;
  Signal_Action.sa_flags = 0;
  sigemptyset (&(Signal_Action.sa_mask));
  sigaction (SIGNAL_TIMEOUT, &Signal_Action, NULL);


  // Levels stuff


  for (i = 0; i < MAX_LEVELS; i++) {
    Level [i].Id = NO_LEVEL;
  }

  pthread_attr_init      (&Default_Thread_Attributes);
  
  Default_Sched_Parameters.sched_priority = DEFAULT_PRIORITY;
  

  //  Signals which are managed synchronously
  //  (SIGNAL_EXECUTE and SIGNAL_ABORT)
  //  must be masked except when the level is waiting
  //  for them in sigsuspend

  pthread_sigmask (SIG_BLOCK, NULL, &Level_Normal_Mask);
  sigdelset (&Level_Normal_Mask, SIGNAL_STOP);
  sigaddset (&Level_Normal_Mask, SIGNAL_EXECUTE);
  sigaddset (&Level_Normal_Mask, SIGNAL_ABORT);

  Level_Mask_When_Waiting = Level_Normal_Mask;
  sigdelset (&Level_Mask_When_Waiting, SIGNAL_EXECUTE);
  sigdelset (&Level_Mask_When_Waiting, SIGNAL_ABORT);

  // All signals managed by a level are handled.

  Signal_Action.sa_handler = SIGNAL_ABORT_Handler;
  Signal_Action.sa_flags = 0;
  sigemptyset (&(Signal_Action.sa_mask));
  sigaction (SIGNAL_ABORT, &Signal_Action, NULL);

  Signal_Action.sa_handler = SIGNAL_STOP_Handler;
  Signal_Action.sa_flags = SA_RESTART;
  sigemptyset (&(Signal_Action.sa_mask));
  sigaction (SIGNAL_STOP, &Signal_Action, NULL);

  Signal_Action.sa_handler = SIGNAL_EXECUTE_Handler;
  Signal_Action.sa_flags = 0;
  sigemptyset (&(Signal_Action.sa_mask));
  sigaction (SIGNAL_EXECUTE, &Signal_Action, NULL);

  // Ready to play

  Linux_Core_Initialized = TRUE;

  return LINUX_CORE_SUCCESS;
}
 
