/**  
 *  \file meta_rule_struct.h					
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 *  \version 2.0
 *
 *  \brief Meta-Rule Store Manager 2.0 - Private Structures
 *
 *       This module incorporates all the functions to on-line interpret the 
 *     meta-rules. 
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */


#ifndef _META_RULE_STRUCT_
#define _META_RULE_STRUCT_


/** NOTE FOR ME & MY SHADOW: 
 *  \todo I'm trying to change all the pointers to integers.
 */

/// event type list 
typedef struct aregEventType {
  int nEventType;  
  int pFirst;  ///< 1st regMetaRule associated to this event type
} aregEventType;

typedef struct regEventType {
  int nEventType;
  int nEventTypeIndex;    ///< order within aregEventType
  /** Pointer to the next event type (in the empty queue) or to the next
   *     Meta-Rule with the same event type (to implement the rule's 
   *     event slot indexation)
   */ 
  int pNext;   
} regEventType;

typedef struct regpSlot {
  void *pSlot; ///< It should be an int, instead of KDM_Handle * 
  int npSlotIndex;    ///< order within paregpSlot 
  /** Pointer to the next event slot (in the empty queue) or to the next
   *     Meta-Rule with the same pointer to slot (to implement the rule's 
   *     event slot indexation)
   */    
  int pNext;
} regpSlot;

/// event slot list
typedef struct aregpSlot {
  int nSlots; ///< number of slots with events associated
  int pFirst; ///< Indicates the 1st Meta-Rule sharing the same pSlot
} aregpSlot;


/// Union of all the posible types of the factor of an expr.
typedef union uniTypeFactor {
  int pExpr;   ///< ptr to expr. It is declared before regExprNode
  void *pSlot; ///< It should be an int, instead of KDM_Handle * 
  int nFunctId;
  int nFactor;
  float fFactor;
} uniTypeFactor;

typedef struct regTypeFactor {
  /** Code indicating the type in the other struct member:
   *    - code
   *       - 0   --> pExpr
   *       - 1   --> pSlot
   *       - 2   --> nFunctId
   *       - 3   --> nFactor
   *       - 4   --> fFactor
   */
    int nType; 
    uniTypeFactor uniType;
} regTypeFactor;

typedef struct regExprNode {
    regTypeFactor regF1;
   /**   - op
    *      - 0   --> no operation
    *      - 1   --> <>
    *      - 2   --> <
    *      - 3   --> >
    *      - 4   --> >=
    *      - 5   --> <=
    *      - 6   --> = (compare)
    *      - 7   --> +  
    *      - 8   --> -
    *      - 9   --> *
    *      - 10  --> /
    */
  int nOp; 
  regTypeFactor regF2;
  int pNext; ///< Indicates the next regExprNode
} regExprNode;



/// Union of all the posible types of the 1st arg. of an action
typedef union uniTypeArg1 {
  void *pSlot; ///< It should be an int, instead of KDM_Handle *
  int nFunctId;
  int nLevelId;
  int nReactDegree;
  int nSlsPolicy;
  int nAttentionFocus;    ///< id 
  float fAttentionDegree;
  float fGain;
  int nBehav;             ///< id 
} uniTypeArg1;

/// Union of all the posible types of the 2nd arg. of an action
typedef union uniTypeArg2 {
  int pExpr; ///< Indicates the regExprNode associated
  int iType;   ///< Indicates the type of the event to add or to release
  float fFocusAttentDegree;
} uniTypeArg2;


/** Valid data combinations for the regActNode:
 *     - nAct (int -> action code)      regArg1                 regArg2
 *     - 0 -> Assign                    pSlot (relative ptr)    pExpr
 *     - 1 -> Function Call             nFunctId                ---
 *     - 2 -> TriggerSporadicLevel      nLevelId                ---
 *     - 3 -> SetReactivityDegree       nReactDegree            ---
 *     - 4 -> SetPolicy                 nSlsPolicy              ---
 *     - 5 -> SetFocus                  nAttentionFocus         fFocusAttentDegree
 *     - 6 -> SetAttentionDegree        fAttentionDegree        ---
 *     - 7 -> SetBehaviour              nBehav                  ---
 *     - 8 -> KDM_Add_Event             pSlot                   iType       iImp (nAct from the next regActNode)
 *     - 9 -> KDM_Release_Event         pSlot                   iType 
 */
typedef struct regActNode {
  int nAct; ///< code of the action to execute
  uniTypeArg1 regArg1;
  uniTypeArg2 regArg2;
  int pNext;  ///< Indicates the next regActNode
} regActNode;


/// Internal representation of a Meta-Rule
typedef struct regMetaRule {
  int pregEvent;     ///< Indicates the regEventType associated
  int pregEventSlot; ///< Indicates the regEventSlot associated
  int pregCondition; ///< Indicates the regCondition associated
  int pregAction;    ///< Indicates the regAction associated
} regMetaRule;

/// Meta-Rule of an attention focus
typedef struct regpMetaRuleNode {
  int pMetaRule;        ///< Points to the regMetaRule associated
  int pNext; ///< Pointer to the next Meta-Rule Node: It implements the AttentionFocus set of rules.
} regpMetaRuleNode;

/// Attention Focus
typedef struct regAttentionFocus {
  int nMetaRules; ///< number of meta-rules composing the focus
  int pFirst; ///< Indicates the 1st regpMetaRuleNode of the Attention Focus
} regAttentionFocus;

/// Active Attention Focus Node
typedef struct regActiveAttFocusNode {
  float fAttentionDegree; 
  int pAttFocus; ///< Points to the regAttentionFocus
  /** Pointer to the next Active Attention Focus Node:
   *      It implements the Active Attention Focus indexation
   */
  int pNext; ///< Indicates the next regActiveAttFocusNode
} regActiveAttFocusNode;

/// Active Attention Focus list
typedef struct aregActiveAttFocus {
  int nActive; ///< Number of active attention focuses
  int pFirst;  ///< Points to the 1st regActiveAttFocusNode
} aregActiveAttFocus;





#endif
