/**  
 *  \file meta_rule_actions.c
 *  \date 2004
 *  \author Carlos Carrascosa Casamayor (carrasco@dsic.upv.es)
 *  \version 2.0
 *
 *  \brief Meta-Rule Actions 2.0 - Functions implementing the actions in the Meta-Rules
 *
 *  There are also actions for Meta-Rules:
 *    - To activate an Attention Focus: 
 *       Through the function defined in meta_rule_store.c
 *        void Meta_Rule_Activate_Attention_Focus(int nAF, float fAttentDegree)
 *    - To deactivate an Attention Focus: 
 *       Through the function defined in meta-rule_store.c
 *        void Meta_Rule_Deactivate_Attention_Focus(int nAF, float fAttentDegree)
 *    - To change the value of an slot (that will be a Meta-Reasoning slot):
 *        Through the function defined in kdm.c
 *             int KDM_Set_Value( ... ) 
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software 
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */


#include "../include/meta_rule_actions.h"  
    /**
     *   For the Meta-Reasoning "slots":
     *     - int heuristic;
     *         This should be called SLS - Second Level Scheduler current policy
     *     - float system_quality ;   
     *         IT WOULDN'T HAVE TO BE A FLOAT 
     *     - int nExecIS;     
     *         Number of times the IS has been executed 
     *     - float MaxQuality; 
     *     - int ReactivityDegree;   
     *         [0..100] 
     *          --> scheduler.c (cuando se puedan adelantar las partes finales)
     *     - int EventRate; 
     *          Max. number of events treated in a control cycle 
     *     - int LevelsCompleted; 
     *           Number of levels completed with the current scheduling policy 
     *     - int LevelsInterrupted;
     *           Number of levels interrupted with the current scheduling policy
     *   More Meta - Slots needed: CurrentAttentionFocus 
     */

#include "../../common/include/artis_conf.h"  ///< For the ARTIS_TRACE_EVENTS

#include "../../include/kdm.h"

#include "../../include/metaclass.h"

#include "../../trace/include/trace.h"

#include "../../events/include/events/events_internal.h"  ///< MAX_EV, ...

#ifndef MODULE
#include <stdio.h>
#endif

#include "../../events/include/events.h"  ///< For the MAX_EV constant

#ifndef MODULE
extern void Trigger_Optional( float fGain, Event_T regAux ); ///< Defined in em.c
#else
void Trigger_Optional( float fGain, Event_T regAux ) {
}
#endif


/**
 * \fn MR_Action_Trigger_Sporadic_Level
 *
 * \brief This function inserts in the tsa the sporadic level indicated by nLevelId.
 */
void MR_Action_Trigger_Sporadic_Level( int nLevelId ) {

#ifndef MODULE
  fprintf(stderr, "\t(meta-rule_actions.c::TriggerSporadicLevel) The new Sporadic Level is %d\n", 
	  nLevelId );
#endif
  
}


/**
 * \fn MR_Action_Trigger_Optional
 *
 * \brief This function activates / deactivates the optional parts of an in-agent.
 */
void MR_Action_Trigger_Optional( float fGain, Event_T regAux ) {
  Trigger_Optional( fGain, regAux );
}


/**
 * \fn MR_Action_Set_Reactivity_Degree
 *
 * \brief Sets nReactDegree as the Reactivity Degree of the AA 
 */
void MR_Action_Set_Reactivity_Degree( int nReactDegree ) {
#ifdef ARTIS_CONF_METAREASONING
  KDM_Set_Value( pMetaClass->reactivityDegree, nReactDegree );
#endif
}


/**
 * \fn MR_Action_Set_Policy
 *
 * \brief Sets the Second Level Scheduling Policy.
 */
void MR_Action_Set_Policy( int nSlsPolicy ) {
#ifdef ARTIS_CONF_METAREASONING
  ///  if (nSlsPolicy == DBS)
  ///  free_l_agent(&tsa);

  KDM_Set_Value( pMetaClass->slsPolicy, nSlsPolicy );
#ifndef MODULE
  fprintf(stderr, "\t(meta-rule_actions.c::SetPolicy) The new SLS Policy is %d\n", 
	  nSlsPolicy );
#endif

#ifdef ARTIS_TRACE_EVENTS
#ifdef MODULE
  trace_event( SLS_POLICY, 0, nSlsPolicy, 0 ); 
#else
  trace_event( SLS_POLICY, 1, nSlsPolicy, 0 );  
#endif
#endif
#endif

  //heuristic = nSlsPolicy;
}


/**
 * \fn MR_Action_Set_Attention_Degree
 * 
 * \brief Sets the percentage of the DS time that is used by the EM
 *        It comes in a [0, 1] format, and is transformed in [0, MAX_EV]
 *            representing the maximal number of events treated by one EM exec.
 */
void MR_Action_Set_Attention_Degree ( float fAttentionDegree ) {
  int nEventRate;

  nEventRate = (int) ( ((double) fAttentionDegree) * MAX_EV);
#ifdef ARTIS_CONF_METAREASONING
#ifdef MODULE
#ifdef DEBUG
  rtl_printf("(RTL) meta-rule_actions.c ==> SetAttentionDegree(): Before\n");
#endif
#endif
  KDM_Set_Value( pMetaClass->eventRate, nEventRate );

#ifdef ARTIS_TRACE_EVENTS
#ifdef MODULE
  trace_event( EVENT_RATE, 0, nEventRate, 0 ); 
#else
  trace_event( EVENT_RATE, 1, nEventRate, 0 );  
#endif
#endif
#endif
}


/**
 * \fn MR_Action_Get_Attention_Degree
 *
 *  Gets the percentage of the DS time that is used by the EM.
 *        It comes in a [0, MAX_EVENTS] format, representing the maximal 
 *        number of events treated by one EM exec, and is transformed to [0, 1]
 */
float MR_Action_Get_Attention_Degree (void) {
  int nEventRate;

#ifdef ARTIS_CONF_METAREASONING
  nEventRate = * (int *) KDM_Get_Value( pMetaClass->eventRate );

#ifdef MODULE
#ifdef DEBUG
  rtl_printf("(RTL) meta-rule_actions.c ==> nEventRate: %d\n", nEventRate);
#endif
#endif
#endif

  return ((float) ((double) nEventRate) / MAX_EV);
}


/**
 * \fn MR_Action_Set_Behaviour 
 *
 * \brief Establishes nBehav as the current behaviour.
 */
void MR_Action_Set_Behaviour( int nBehav ) {

#ifdef ARTIS_CONF_METAREASONING
  ///int nOldBeh;

  ///KDM_Get_Slot_Value( MSlot_int.nCurrentBehavior.nValue, &nOldBeh );
  KDM_Set_Value( pMetaClass->nCurrentBehavior, nBehav);

#ifdef MODULE
#ifdef DEBUG
  rtl_printf("(RTL) meta-rule_actions.c ==> nBehav: %d\n", nBehav);
#endif
#endif
#endif
}


/**
 * \fn MR_Action_Initialize
 *
 * This function has to be executed after the Meta-KDM is initialized 
 *      in the optional part. This is the reason it is called in the 
 *      init_module() of rt_tasks.c
 */
void MR_Action_Initialize(void) {
  ///#ifdef MODULE
  ///  Handle_Register_Meta_Class();
  ///#endif
}


/* ==================================================================== */


#ifdef MODULE
int MR_Action_init_module(void) {
  return 0;
}

void MR_Action_cleanup_module(void) {
}
#endif
