/*

   FILE:
	mutex.c

   Copyright (C) 1998-2003
        GTI-IA, DSIC, Universidad Politecnica de Valencia


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2004-05-10

  CHANGELOG:


  DESCRIPTION:

    The MUTEX module provides access to CSP-like mutexes to the
    application at both sides (RT-Linux tasks and Linux levels).  

    The CSP protocol sets a priority ceiling to each mutex, being
    computed as the priority of the highest-priority task accessing
    the mutex.  When a task locks a mutex, the task's priority is
    immediately raised to the mutex ceiling. When the task unlocks the
    mutex, it retrieves the priority it had when it locked that
    mutex. The current implementation meets this behaviour only at the
    RT-Linux level.
    
    At the Linux level this functionality is different, since we do
    not have a general-purpose mutual exclusion mechanism between
    RT-Linux task and Linux levels. Instead, the lock operation at the
    Linux level has two parameters: the mutex and the interval that
    the level is going to lock the mutex. Only if the mutex is free
    and the critical section can be completely executed within the
    current slack interval can the level successfully lock the mutex.

    In general, all these functions return 0 if everything went OK and
    some negative number if something went wrong.

*/




#include "../../common/include/artis_conf.h"
#include "../../trace/include/trace.h"
#include "../../time/include/time.h"
#include "../../dserver/include/dserver.h"

#ifdef MODULE
#include <mbuff.h>
#include "../../rtl_core/include/rtl_core.h"
#include "../../rserver/include/rserver.h"
#else
#include "../../linux_core/include/linux_core.h"

extern systime_t end_time;

#endif




// DEFINES:
#define PARANOIC_MODE // Extra tests in lock/unlock (comment it if not needed)

#define MAX_MUTEXES       ARTIS_LAST_MUTEX+1

#define MBUFF_MUTEX_NAME  "artis_mutex_mbuff"
#define MBUFF_MUTEX_SIZE  (sizeof(struct mutex_struct)*MAX_MUTEXES)

#define MAGIC_MUTEX_FREE  0xebf1
#define MAGIC_MUTEX_BUSY  0xebf2

#define NO_TASK           (rt_task_t *) 0
#define NO_MUTEX          (rt_mutex_struct *) -1

#define MUTEX_FIND(index) \
                          (rt_mutexes[index].magic == MAGIC_MUTEX_BUSY? \
	                   &rt_mutexes[index] : \
			   NO_MUTEX)
     


// NEW TYPES:
typedef enum {				
                MUTEX_OPEN,   
	        MUTEX_CLOSED 
} rt_mutex_status;


typedef struct mutex_struct {			
        int             magic;            /* magic: "free" or "busy" */
        int	        ceiling;          /* mutex's priority ceiling */ 
        rt_mutex_status status;           /* "open" or "closed" */
        int 	        entry_priority;   /* owner's priority at lock */
} rt_mutex_struct;

typedef rt_mutex_struct *rt_mutex_t;



// GLOBAL VARIABLES:
static rt_mutex_struct *rt_mutexes;




// FUNCTION IMPLEMENTATION:

#ifdef MODULE
//
// RT-Linux implementation of mutexes:
//

int  
mutex_init_module (void) 
{

  rt_mutexes = (rt_mutex_struct *) mbuff_alloc(MBUFF_MUTEX_NAME,
					       MBUFF_MUTEX_SIZE);

  if (rt_mutexes == NULL) {
    rtl_printf("(rt-mutex) error: mbuff alloc failed\n");
    return -1;
  }
#ifdef RTLINUX_ONLY // in "../../common/include/artis_conf.h"
  { 
    int index;
    for(index=0; index < MAX_MUTEXES; index++) {
      rt_mutexes[index].magic = MAGIC_MUTEX_FREE;
    }
    rtl_printf("(rt-mutex) RTLINUX ONLY: Initializing mutex in RTL\n");
  }
#endif
  if (rt_mutexes[0].magic != MAGIC_MUTEX_FREE 
      && rt_mutexes[0].magic != MAGIC_MUTEX_BUSY) {
    rtl_printf("(rt-mutex) error: array of mutexes not initialized\n");
    return -1;
  }

  return 0;
}



int  
mutex_cleanup_module (void) 
{
  mbuff_free(MBUFF_MUTEX_NAME, (void*)rt_mutexes);
  return 0;
}



int  
mutex_lock (int mutex_index) 
{
  rt_mutex_t mutex;

  if ((mutex = MUTEX_FIND(mutex_index)) == NO_MUTEX) {
    rtl_printf("(rt_mutex) error: mutex %d not found in lock operation\n",
	       mutex_index);
    return -1;
  }

#ifdef PARANOIC_MODE
  if (mutex->status == MUTEX_CLOSED 
      || mutex->ceiling > TASK_PRIO(rtl_running)) {
    rtl_printf("(rt-mutex) error: Illegal locking of mutex %d\n", mutex_index);
    rtl_printf("Reason: mutexes not properly nested or wrong ceilings\n");
    return -1;
  }
#endif

  rtl_update(); // In "rserver" module

  mutex->status          = MUTEX_CLOSED;
  mutex->entry_priority  = TASK_PRIO(rtl_running);
  TASK_PRIO(rtl_running) = mutex->ceiling;

#ifdef ARTIS_TRACE_EVENTS
  trace_event(MUTEX_LOCK, rtl_running->identifier, (int) mutex_index, 0);
#endif

  rtl_resume(); // In "rtl_core" module
  return 0;
}



int  
mutex_unlock (int mutex_index) 
{
  rt_mutex_t mutex;

  if ((mutex = MUTEX_FIND(mutex_index)) == NO_MUTEX) {
    rtl_printf("(rt_mutex) error: mutex %d not found in unlock operation\n",
	       mutex_index);
    return -1;
  }
  rtl_update(); // In "rserver" module

  mutex->status = MUTEX_OPEN;

#ifdef ARTIS_TRACE_EVENTS
  trace_event(MUTEX_UNLOCK, rtl_running->identifier, (int) mutex_index, 0);
#endif

  if (mutex->entry_priority > TASK_PRIO(rtl_running)) {
    TASK_PRIO(rtl_running) = mutex->entry_priority;
    rtl_schedule();
  } else {
    rtl_resume();
  }

  return 0;
}



int  
mutex_destroy (int mutex_index) 
{
  rt_mutex_t mutex;

  if ((mutex = MUTEX_FIND(mutex_index)) == NO_MUTEX) {
    rtl_printf("(rt_mutex) error: mutex %d not found in destroy operation\n",
	       mutex_index);
    return -1;
  }
  mutex->magic = MAGIC_MUTEX_FREE;
  return 0;
}




#ifdef RTLINUX_ONLY
int  
mutex_create (int mutex_index, int priority_ceiling) 
{
  if (rt_mutexes[mutex_index].magic != MAGIC_MUTEX_FREE) {
    rtl_printf("(mutex) error: cannot create mutex %d\n",
	       mutex_index);
    return -1;
  }
  rt_mutexes[mutex_index].magic   = MAGIC_MUTEX_BUSY;
  rt_mutexes[mutex_index].status  = MUTEX_OPEN;
  rt_mutexes[mutex_index].ceiling = priority_ceiling;

  return 0;
}
#endif


//
//
//
#else // ------------------------------------------------------------
//
// Linux implementation of mutexes:
//

#include <stdio.h>

extern int mbuff_alloc(const char *name, int size);
extern int mbuff_free(const char *name, void * mbuf);
extern void * mbuff_attach(const char *name, int size);
extern void mbuff_detach(const char *name, void * mbuf);



int  
mutex_init_module (void) 
{
  int index;

  rt_mutexes = (rt_mutex_struct *) mbuff_attach(MBUFF_MUTEX_NAME,
						MBUFF_MUTEX_SIZE);

  if (rt_mutexes == NULL) {
    fprintf(stderr, "(mutex) error: mbuff alloc failed\n");
    return -1;
  }

  for(index=0; index < MAX_MUTEXES; index++) {
    rt_mutexes[index].magic = MAGIC_MUTEX_FREE;
  }

  return 0;
}



int  
mutex_cleanup_module (void) 
{
  mbuff_detach(MBUFF_MUTEX_NAME, (void*)rt_mutexes);
  return 0;
}



int  
mutex_create (int mutex_index, int priority_ceiling) 
{
  if (rt_mutexes[mutex_index].magic != MAGIC_MUTEX_FREE) {
    fprintf(stderr, "(mutex) error: cannot create mutex %d\n",
	    mutex_index);
    return -1;
  }
  rt_mutexes[mutex_index].magic   = MAGIC_MUTEX_BUSY;
  rt_mutexes[mutex_index].status  = MUTEX_OPEN;
  rt_mutexes[mutex_index].ceiling = priority_ceiling;

  return 0;
}


/***
 *  This function is used by the DS to access to the KDM
 *  It returns 0 on success, -1 if the mutex referenced
 *     doesn't exist, and -2 if the blocking_duration is not enough
 */
int  
mutex_system_lock (int mutex_index, systime_t blocking_duration) 
{
  rt_mutex_t mutex;
  systime_t  now;

  if ((mutex = MUTEX_FIND(mutex_index)) == NO_MUTEX) {
    fprintf(stderr,"(mutex) error: mutex %d not found in lock operation\n",
	       mutex_index);
    return -1;
  } 

  now = systime_gettime();

  if (mutex->status != MUTEX_OPEN 
      || now + blocking_duration >= end_time) {
      return -2;
  }
#ifdef ARTIS_TRACE_EVENTS
  trace_event(OPT_MUTEX_LOCK, running_level, (int) mutex_index, 0);
#endif

  return 0;
}



int  
mutex_lock (int mutex_index, systime_t blocking_duration) 
{
  rt_mutex_t mutex;
  systime_t  now;

  if ((mutex = MUTEX_FIND(mutex_index)) == NO_MUTEX) {
    fprintf(stderr,"(mutex) error: mutex %d not found in lock operation\n",
	       mutex_index);
    return -1;
  } 

  now = systime_gettime();

  if (mutex->status != MUTEX_OPEN 
      || now + blocking_duration >= current_level_abs_timeout) {
    // The task (level) cannot lock the mutex in these circumstances
    // so we need to suspend it
    Level_Suspend_Self();
  }
#ifdef ARTIS_TRACE_EVENTS
  trace_event(OPT_MUTEX_LOCK, running_level, (int) mutex_index, 0);
#endif

  return 0;
}



int  mutex_unlock (int mutex_index) 
{
  // In Linux, this operation is really void!
#ifdef ARTIS_TRACE_EVENTS
  trace_event(OPT_MUTEX_UNLOCK, running_level, (int) mutex_index, 0);
#endif
  return 0;
}



int  
mutex_destroy (int mutex_index) 
{
  rt_mutex_t mutex;

  if ((mutex = MUTEX_FIND(mutex_index)) == NO_MUTEX) {
    fprintf(stderr,"(mutex) error: mutex %d not found in destroy operation\n",
	    mutex_index);
    return -1;
  }
  mutex->magic = MAGIC_MUTEX_FREE;
  return 0;
}


#endif




