/**
 * \file rserver.c
 * \date 1998-2004
 * \author A. Terrasa, I. Pachs, C. Carrascosa
 * \version 3.0
 *
 * \brief Reactive Server 3.0 
 *            Based on original RT-Linux v1 scheduler (c) FSMLabs
 * 
 *   The RSERVER module implements the Reactive Server (also known as
 *   the "first-level scheduler") of the ARTIS architecture. It is
 *   basically a scheduler of RT-Linux real-time threads, adapted to 
 *   match the requirements of ARTIS.
 *
 *   This module offers two interfaces: 
 *
 *      a) To the application real-time module. It exports the task
 *      type and functions to create and configure the application
 *      tasks and to start and stop the application.
 *      
 *      b) To other fixed parts of the ARTIS architecture, such as the
 *      RT-Linux MUTEX module
 *
 *   In particular, this second interface exports a pointer to the
 *   current running task and functions to suspend and wake up a
 *   particular task to the MUTEX module.
 *
 *
 * This is free software;  you can  redistribute it  and/or modify it under
 * terms of the  GNU General Public License as published  by the Free Soft-
 * ware  Foundation;  either version 2,  or (at your option) any later ver-
 * sion.  It is distributed in the hope that it will be useful, but WITH-
 * OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for  more details.  You should have  received  a copy of the GNU General
 * Public License;  see file COPYING.  If not, write to  the Free Software
 * Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.
 *
 */

/** 
 * ######################################################################
 * Public functions:
 *  (1) To other rt (system) modules:
 *     - rtl_update         Updates the task' dynamic timing attributes
 *     - rtl_schedule       Selects a new running task
 *     - rt_timer_handler   Catches the timer interrupts
 *
 *  (2) To the rt application (i.e., these are "system calls"):
 *     - rt_task_init          Initializes a task
 *     - rt_make_periodic      Makes a task periodic
 *     - rt_start_application  Starts running the rt application
 *     - rt_task_wait          Ends the current part of a task
 *                             (initial or final)
 *     - rt_task_delete        Deletes a task
 *
 *     - rt_end_scheduler      Sends a finishing message to the 2nd 
 *                             level and stops scheduling
 * ######################################################################
 *
 */


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <rtl_fifo.h>

#include "../../common/include/artis_conf.h"
#include "../../time/include/time.h"
#include "../../common/include/artis_fifo_definition.h"

#include "../../include/rserver.h"

#include "../../rtl_core/include/rtl_core.h"
#include "../../trace/include/trace.h"
#include "../../rserver/include/rserver.h"

#include "../../events/include/events.h" 




#ifdef ARTIS_CONF_METAREASONING
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <rtl_posixio.h>
#include <stdarg.h>

#include "../../include/kdm.h"

#include "../../include/metaclass.h"
#include "../../meta_rule/include/meta_rule_actions.h"

#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
/// It stores the old Reactivity Degree, to calculate the original slack
/// Initially there is no slack variation
static int nOldReactDegree = 100;
#endif

/// It stores the last slack (not graded by the react. degree) sended to the DS
///static systime_t OldSlack;
#endif
/***/

rt_task_t *preemp_task = NULL; ///< Next task that will wake up the FLS

#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_WAKEUP
static int fdWakeUpRTFIFO;     ///< File descriptor for the wake up rt-fifo
#endif

/** 
 * NOTE about bWakeUpbyFIFO:
 *
 * If the DS has sent a message through the WakeUp FIFO, it means that it
 * has nothing else to run by now. This affects the scheduling:
 * if bWakeUpbyFIFO is "1", then do not send a message to wake up the DS.
 *
 * This prevents the DS and RTLinux to wake up one another constantly.
 *
 */
static int bWakeUpbyFIFO = 0;  

/**
 * nFinalsExecuted is used to compute the number of tasks that have executed 
 *    (in this hyperperiod) its final part. When nFinalsExecuted == n_tasks,
 *    then the Event Rate is changed to its maximum to try to execute the 
 *    maximun events possible (to deactivate the task's optinal parts).
 *
 */
static int nFinalsExecuted = 0;

#endif


/// Misc definitions (ATB: Maybe remove them sometime)

#define FALSE                 0
#define TRUE                  1
#define RTL_LINUX_ID         -1
#define MAXLONG              (systime_t) 0xffffffff 
#define EVER                 ;1;
#define OVERHEAD_FACTOR      0


#define SUBTRACT( a, b )     a > b ? a - b : 0;
#define MIN( a, b )          a > b ? b : a
#define MAX( a, b )          a > b ? a : b


/**
 *  The following value must be set in a conservative way. Don't want
 *  the 2nd-level scheduler to be started and just preempted without
 *  giving it the opportunity of doing something useful (at least, the
 *  time to find out that it does not have time enough and suspend
 *  itself :)
 *
 */
#define MIN_SLACK_TO_RUN_OPTIONAL   (systime_t) 1000000


/// Task-related global variables

rt_task_t *rtl_running;		///< a pointer to the running task

/**
 * Header of the (priority-ordered) list of RT tasks. Each task has
 * its own rt_task_t struct. 
 *
 */
static rt_task_t *rt_tasks;	

static rt_task_t rtl_linux_task;	///< struct of the linux "task"






///#define ARTIS_CONF_METAREASONING_BEHAVS




#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_BEHAVS

/// It will point to the beginning of the Behavior list. It is 
///  initiallized by rtl_tasks.c
static rt_behav_t *rt_behav = (rt_behav_t *) NULL;

/// It will be -1 except when there is a transition between behaviors 
///  that will keep the previous behav. id
static int nLastBehavior;  

static int nCurrentBehavior;  /// < It will have the current behavior

/// It will keep the number of previous beh. tasks that are still
///  running in a behavior transition.
/// So, when there is no transition, its value will be 0. The
///  unchanged tasks are not considered as previous
static int nOldBehTasks;

static int nBehTasks;  /// < Number of tasks of the current behavior

#endif

/// Old Attention Degree, previous to the transition between behaviors
static float fOldAttentionDegree; 
#endif


/**
 *  Time-related variables.
 *
 *  'now', 'last_update' 'last_kernel_time' in rtl_core.h
 *
 *  Note: With these three time variables, we know:
 *
 *   - The interval between events (now-last_update), which is used to
 *     update the tasks' dynamic attributes, like their slack time.
 *
 *   - The interval used by the kernel to service an interrupt or a
 *     system call (now-last_kernel_time), which is used to measure the
 *     kernel overhead.
 *
 */

static systime_t preempt_time;
static systime_t ini_time;


/**
 * This global variable indicates whether or not the application is
 *  currently running. If not, the timer handler does not have to 
 *  update/schedule.
 *
 */
static int app_started = FALSE;


/**
 * Message for communication between the rtl scheduler and the 2nd.-level
 *  scheduler.
 *
 */
static struct sched_mssg_t sched_mssg;


///#ifdef ARTIS_CONF_METAREASONING
/** 
 *   To indicate the moment it is going to end the application, so the FLS
 *     is not going to schedule any critical task anymore.
 *
 */
///static int bExitApp = 0;
///#endif


static int rt_task_wait (void);
static int rt_slack_available (int prio_level, systime_t * available_slack);
static int rt_calculate_slack (rt_task_t * task, systime_t * result);
int rt_task_delete (rt_task_t * task);
void rt_end_scheduler (void);


#ifdef ARTIS_CONF_RS_DEBUG_1
void _debug_print_tasks (void) {
  rt_task_t *task;
  rtl_printf("(RSERVER) [print_tasks]. Current list of tasks:\n");
  for (task = rt_tasks; task; task = task->next) {
      rtl_printf("    task (%x), id=%d, prio=%d (base %d), next=%x\n",
                 task,
                 task->identifier,
                 task->priority,
                 task->basepriority,
                 task->next);
    }
}
#endif


/* ========================================================================= */
int
rtl_startup (void (*fn) (int), int data)
/* ========================================================================= */
{

  rtl_allow_interrupts ();
  (*fn) (data);
  rt_task_delete (rtl_running);
  /* should never reach this line */
  return 0;
}


#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_BEHAVS
/// ThreeC: Behavior
/* ========================================================================= */
void ChangeBehavior( int nBehav )
/* ========================================================================= */
{
  rt_task_t *task, *prev;
  ///Event_T regEvent;	///< The record with the event information

/***
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  int nReactDegree;
#endif  
***/

  rt_behav_t * rt_auxBeh;
  rt_behav_task_t *beh_task;
  ///int nTasks, nPrio, i;
  int i;

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
  trace_event (MCR_BEGIN, 0, 0, 0);
#endif
#endif

  nLastBehavior = nCurrentBehavior;

#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RTL) Change Behavior: previous beh.: %d\n", 
             nLastBehavior);
#endif
  
  nCurrentBehavior = nBehav;
  
  ///Indicates the transition between behaviors
  KDM_Set_Value( pMetaClass->nLastBehavior, nLastBehavior ); 

  /// The Attention Degree must be maximum, to elliminate all the previous behavior levels
  fOldAttentionDegree = MR_Action_Get_Attention_Degree();

#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RTL) Change Behavior, Old Attention Degree.: %d\n", 
             (int)(((double) fOldAttentionDegree ) * 100));
#endif

  MR_Action_Set_Attention_Degree( 1.0 );

  /// rt_auxBeh must point to the new behavior
  for (i = 1, rt_auxBeh = rt_behav; i <= nBehav; i++)
    rt_auxBeh = rt_auxBeh->pNext;
  
#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RTL) Change Behavior: Before inserting tasks\n");
#endif

  nOldBehTasks = nBehTasks;
  nBehTasks = rt_auxBeh->nTasks;

#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RTL) Change Behavior to %d, and the old beh. have %d tasks\n", 
             nBehav, nOldBehTasks);

  for ( beh_task = rt_auxBeh->task_list, i = 0;
        i < nBehTasks; i++, beh_task = beh_task->pNext)
    rtl_printf("(RTL) Change Behavior:: New Beh. Task: %d\n", 
               beh_task->task->identifier );
#endif


  /*** Asynchronous MCProtocol:
   *        Insert the new behavior tasks in the rt_tasks list, 
   *           seting task->nBehav to nBehav with the corresponding offsets. 
   *           The previous tasks will die when they execute their final part
   *
   *   Synchronous MCProtocol:
   *        Change the task->nBehav member of the unchanged tasks to nBehav.
   *        When all the previous behavior tasks die, the rest of the new 
   *           behavior tasks will be inserted
   ***/
  /// Synchronous MCProtocol:
  for (i = 0, beh_task = rt_auxBeh->task_list, prev = task = rt_tasks;
       (task != &rtl_linux_task) && (nOldBehTasks) && (i < nBehTasks);
       ) {

#ifdef ARTIS_CONF_RS_DEBUG_2
    rtl_printf("(RTL) ChangeBehavior:: task: %d, beh_task: %d\n", 
               task->identifier, task->priority, 
               beh_task->task->identifier, beh_task->task->priority);
#endif

    if (task == beh_task->task) {
      /// The unchanged task is not of the previous beh. from now on
      task->nBehav = nBehav;
      beh_task = beh_task->pNext;
      i++;
      nOldBehTasks--;  
      prev = task; 
      task = task->next;
#ifdef ARTIS_CONF_RS_DEBUG_2
      rtl_printf("(RTL) ChangeBehavior:: I have changed the task->nBehav to %d\n", nBehav);
#endif
    }
    else if ((task != rtl_running) &&
             (task->priority < beh_task->task->priority)) {
      /// This task doesn't belong to the new beh.
#ifdef ARTIS_CONF_RS_DEBUG_2
      rtl_printf("(RTL) ChangeBehavior:: Task %d does'nt belong to the new beh.\n", task->identifier);
#endif

      if (task->state == RT_TASK_DELAYED) {
        /// Old beh. task which execution has yet ended
        if (prev == task) {
          prev = task = rt_tasks = task->next;
        }
        else {
          /// Task is not the 1st task in rt_tasks
          prev->next = task->next;
          task = task->next;

#ifdef ARTIS_CONF_RS_DEBUG_2
          rtl_printf("(RTL) ChangeBehavior:: task: id: %d, state: %d\n",
                     task->identifier, task->state);
#endif

        }

        nOldBehTasks--;
        
#ifdef ARTIS_CONF_RS_DEBUG_2
        rtl_printf("(RTL) ChangeBehavior: %d of the old beh. tasks are still active\n",
                   nOldBehTasks);
#endif
        
      }
      else {
        /// This task will finish its execution
        task = task->next;
      }
    }
    else {
      /// the current beh_task was not also of the previous beh.
      beh_task = beh_task->pNext;
      i++;
    }
  }


  /// Eliminate all the old tasks that are delayed (its initial part has not
  /// been yet executed
  for ( ; (task != &rtl_linux_task) && (nOldBehTasks); )
      if (task->state == RT_TASK_DELAYED) {
        /// Old beh. task which execution has yet ended
        if (prev == task) {
          prev = task = rt_tasks = task->next;
        }
        else {
          /// Task is not the 1st task in rt_tasks
          prev->next = task->next;
          task = task->next;

#ifdef ARTIS_CONF_RS_DEBUG_2
          rtl_printf("(RTL) ChangeBehavior:: task: id: %d, state: %d\n",
                     task->identifier, task->state);
#endif

        }

        nOldBehTasks--;
        
#ifdef ARTIS_CONF_RS_DEBUG_2
        rtl_printf("(RTL) ChangeBehavior: %d of the old beh. tasks are still active\n",
                   nOldBehTasks);
#endif
        
      }
      else {
        /// This task will finish its execution
        task = task->next;
      }  




  /// Asynchronous MCProtocol:
  /***
  nTasks = rt_auxBeh->nTasks;
  for (beh_task = rt_auxBeh->task_list, task = rt_tasks, i = 0;
       i < nTasks;
       beh_task = beh_task->pNext) {
 
    for (prev = post = task, nPrio = beh_task->task->priority;
         post->priority <= nPrio; prev = post, post = post->next);

    beh_task->task->next = post;

    if (post == rt_tasks) {
      rt_tasks = beh_task->task;
    }
    else {
      prev->next = beh_task->task;
    }
    
    beh_task->task->nBehav = nBehav;
  }
  ***/


  /***
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    /// This 1st approach aborts all the tasks of the previous behavior
    ***  if ((task->nBehav != nBehav) &&
        (task->priority != NOT_IN_BEHAVIOR)) {
      task->priority = NOT_IN_BEHAVIOR;

      if (task->state == RT_TASK_OPTIONAL)
      {
        regEvent.iType = CONSIDER_PERIODIC_AGENT;
        regEvent.iImp = (char) MAX_IMPORTANCE_EV - 2;
        regEvent.Old_Value.i = task->optional_func_ids[0];
        regEvent.New_Value.li = (long long) -1; 
        * Convention: a '-1' means to DEactivate levels */
  ///        KDM_Insert_Event (regEvent);
  ///    }

  ///  }
  ///  else {  /// task belongs to the new behavior
  /***    task->priority = task->basepriority;
      task->initial_computed_time = (RTIME) 0;
      task->final_computed_time = (RTIME) 0;
      task->resume_time = rt_get_time () + task->offset;
      task->absolute_deadline = task->resume_time + task->deadline;
    }

    task->state = RT_TASK_DELAYED;
  }

 
  ///  Calculate the initial slack time for each task
  ***
  KDM_Get_Slot_Value (MSlot_int.ReactivityDegree.nValue, &nReactDegree);

  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    if (task->priority != NOT_IN_BEHAVIOR) {
      rt_calculate_slack (task, &(task->slack));
      task->slack =
        (RTIME) (((double) task->slack * nReactDegree) /
                 (double) 100);
    }
  }

  nOldReactDegree = nReactDegree;


  KDM_Get_Slot_Value( MSlot_int.nCurrentBehavior.nValue, &i );

  ***/




#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RTL) End of Change Behavior, there are still %d tasks of the old beh. actives\n", 
             nOldBehTasks);
#endif

  if (!nOldBehTasks) {

#ifdef ARTIS_CONF_RS_DEBUG_2  
    rtl_printf("(RTL) Change Beh.:: There is no previous task left\n");
#endif


/***    
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
    nReactDegree = * (int *) KDM_Get_Value( pMetaClass->reactivityDegree );
#endif
***/
    
    /// Insert the new beh tasks
    /// rt_auxBeh must point to the new behavior
    for (beh_task = rt_auxBeh->task_list, 
           prev = task = rt_tasks, i = 0;
         i < nBehTasks;
         i++, beh_task = beh_task->pNext) {
      /// Look for the proper place for beh_task
      for ( ; 
            (task != &rtl_linux_task) && 
              (task->priority < beh_task->task->priority);
            prev = task, task = task->next );
      
      if (task != beh_task->task) {
        beh_task->task->next = task;
        
        if (task == rt_tasks) {
          rt_tasks = beh_task->task;
        }
        else {
          prev->next = beh_task->task;
        }
        
        prev = beh_task->task;
        
        beh_task->task->nBehav = nBehav;
        beh_task->task->priority = beh_task->task->basepriority;
        beh_task->task->resume_time = 
          systime_gettime () + beh_task->task->offset;
        beh_task->task->absolute_deadline = 
          beh_task->task->resume_time + beh_task->task->deadline;
        beh_task->task->state = RT_TASK_DELAYED;
        rt_calculate_slack (beh_task->task, &(beh_task->task->slack));

/***	
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
	if (!nReactDegree)
		beh_task->task->slack = (systime_t) 0;
	else beh_task->task->slack =
                (systime_t) (((double) beh_task->task->slack / (double) 100 
                    * nReactDegree ));
#endif
***/

      }
    }
    /// End of Beh. Transition
    nLastBehavior = -1;
    KDM_Set_Value( pMetaClass->nLastBehavior, nLastBehavior );

    MR_Action_Set_Attention_Degree( fOldAttentionDegree );

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
    trace_event (MCR_END, 0, 0, 0);
#endif
#endif
  }

  rtl_schedule();
}
#endif
#endif


/* ========================================================================= */
void
rt_task_runner (int taskptr)
/* ========================================================================= */
{
  rt_task_t *task; 
  
  task = (rt_task_t *) taskptr;
  
  for (EVER) {
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_1
    trace_event (INI_START, task->identifier, 0, 0);
#endif
#endif
    (task->initial_function) (NULL);
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_1
    trace_event (INI_END, task->identifier, 0, 0);
#endif
#endif
    rt_task_wait ();
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_1
    trace_event (FIN_START, task->identifier, 0, 0);
#endif
#endif
    (task->final_function) (NULL);
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_1
    trace_event (FIN_END, task->identifier, 0, 0);
#endif
#endif
    rt_task_wait ();
  }
}


#ifdef ARTIS_CONF_METAREASONING
/* ==================================================================== */
void
rtl_update (void)
/* ==================================================================== */
{
  systime_t interval;
  rt_task_t *task;

/***
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  int nReactDegree;
#endif
***/

#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  int nBehav;
#endif

  /* Update system time. Calculate the interval since last event */
  rtl_update_system ();	/* (in rtl_core.c) update 'last_update' and 'now' */

  interval = now - last_update;


/***
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  nReactDegree = * (int *) KDM_Get_Value( pMetaClass->reactivityDegree );
#endif
***/


  /* The current task has consumed part of its execution time */
  if (rtl_running != &rtl_linux_task) {
    if (rtl_running->state == RT_TASK_INITIAL 
        || rtl_running->state == RT_TASK_FINAL) {
      rtl_running->execution_left =
	SUBTRACT (rtl_running->execution_left, interval);
      /* WCET error handling should be placed here */
    }
  }


#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  /// ThreeC: Behavior
  if (nLastBehavior == -1) {  /// To indicate the system is not in a transition
    nBehav = * (int *) KDM_Get_Value( pMetaClass->nCurrentBehavior );

    if (nBehav != nCurrentBehavior){

#ifdef ARTIS_CONF_RS_DEBUG_2
      rtl_printf("(RTL) Before changing current behav. %d to new behav. %d\n",
                 nCurrentBehavior, nBehav);
#endif

      ChangeBehavior(nBehav);
      
    }
  }

  /***
  if (nLastBehavior == -1) 
    if (nOldReactDegree != nReactDegree) 
      if (nCurrentBehavior != 1) {
#ifdef DEBUG
        rtl_printf("(RTL) (sched_artis) Before calling change behavior\n"); 
#endif
        ChangeBehavior(1);
      }
      else ChangeBehavior(0);
  ***/
#endif
#endif

  /*
     For each rt task, update its slack time, and see if anyone
     has to be released.
   */
#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RSERVER) [rtl_update] Before SLACK updating\n");
#endif


#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  /// Recuperate the original slack used in last execution
  interval = (systime_t) ((double) interval) / ((double) nOldReactDegree) * (double)100;
#endif

  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    // Slack update
    if (rtl_running == &rtl_linux_task
	|| rtl_running->basepriority > task->basepriority) {
      // 20040513 ATB: In above condition, remove rtl_running == rtl_linux_task





      if (task->state != RT_TASK_DELAYED) {







#ifdef ARTIS_CONF_RS_DEBUG_1
      rtl_printf("(RSERVER) [rtl_update] Before SLACK updating: %lld, interval %lld - Task %d\n",
                 task->slack, interval, task->identifier);
#endif

      task->slack = SUBTRACT (task->slack, interval);






      }






    }

#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RSERVER) [rtl_update] After SLACK updating\n");
#endif

/***
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
    if (!nReactDegree) {
      task->slack = (systime_t) 0;
    }
    else {
	    if (nOldReactDegree != nReactDegree) {
      		/// If the React. Degree is new, update the slack
      		if (!nOldReactDegree) {
        		task->slack = (systime_t) ((double) task->slack / (double) 100) * nReactDegree;
		}
      		else {
			task->slack =
             			(systime_t) ((double) task->slack / (double) nOldReactDegree
                      				* nReactDegree );
		}
#ifdef ARTIS_CONF_RS_DEBUG_0
                rtl_printf("(RSERVER) [rtl_update] SLACK: %lld React. Degree: %d\n", 
                           task->slack, nReactDegree);
#endif
	    }
    }
#endif
***/

    //  Task release 
    if (task->state == RT_TASK_DELAYED && task->resume_time < now + ADJUSTUP) {
      task->state          = RT_TASK_INITIAL;
      task->execution_left = task->initial_wcet + task->final_wcet;
      task->priority       = task->basepriority;
      task->resume_time   += task->period;
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
      trace_event (ACTIV_START, task->identifier, task->deadline, 
                   task->resume_time-systime_gettime());
#endif
#endif
    }
  }


  /***
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  nOldReactDegree = nReactDegree;
#endif
***/


}
#else
/* ==================================================================== */
void
rtl_update (void)
/* ==================================================================== */
{
  systime_t interval;
  rt_task_t *task;

  /* Update system time. Calculate the interval since last event */
  rtl_update_system ();	/* (in rtl_core.c) update 'last_update' and 'now' */

  interval = now - last_update;

  /* The current task has consumed part of its execution time */
  if (rtl_running != &rtl_linux_task) {
    if (rtl_running->state == RT_TASK_INITIAL 
        || rtl_running->state == RT_TASK_FINAL) {
      rtl_running->execution_left =
	SUBTRACT (rtl_running->execution_left, interval);
      /* WCET error handling should be placed here */
    }
  }

  /*
     For each rt task, update its slack time, and see if anyone
     has to be released.
   */
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    // Slack update
    if (rtl_running == &rtl_linux_task
	|| rtl_running->basepriority > task->basepriority) {
      // 20040513 ATB: In above condition, remove rtl_running == rtl_linux_task
      task->slack = SUBTRACT (task->slack, interval);
    }
   
    //  Task release 
    if (task->state == RT_TASK_DELAYED && task->resume_time < now + ADJUSTUP) {
      task->state          = RT_TASK_INITIAL;
      task->execution_left = task->initial_wcet + task->final_wcet;
      task->priority       = task->basepriority;
      task->resume_time   += task->period;
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
      trace_event (ACTIV_START, task->identifier, task->deadline, 
                   task->resume_time-systime_gettime());
#endif
#endif
    }
  }
}
#endif /// #ifdef ARTIS_CONF_METAREASONING


#ifdef ARTIS_CONF_METAREASONING
/* ==================================================================== */
void
rtl_schedule (void)
/* ==================================================================== */
{
  systime_t time_until_preemption;
  systime_t slack_interval = 0;
  systime_t slack_time = 0;
  rt_task_t *task;
  rt_task_t *new_task;

  
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  int nReactDegree = -1;
#endif


#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RSERVER) [rtl_schedule] Beginning\n");
#endif

  // *** Step 1) Selection of new running task
  new_task = &rtl_linux_task;
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next)
    if (task->state == RT_TASK_INITIAL ||
	task->state == RT_TASK_OPTIONAL || task->state == RT_TASK_FINAL)
      if (task->priority <= new_task->priority) {
	new_task = task;
      }


#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  if (new_task == &rtl_linux_task) { 
    nFinalsExecuted = nBehTasks;
  }
#endif

  preemp_task = NULL; /// Initialize preemp_task


  // *** Step 2) Look for the nearest preemption time
  preempt_time = RT_TIME_END;
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    if (task->state == RT_TASK_DELAYED && task->priority < new_task->priority
	&& task->resume_time < preempt_time) {
      preempt_time = task->resume_time;
    }
  }


  slack_interval = preempt_time - now; /// Initialize slack_interval


  // *** Step 3) If there is an optional part to run, check slack available
  if (new_task != &rtl_linux_task && new_task->state == RT_TASK_OPTIONAL) {


#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
    nReactDegree = * (int *) KDM_Get_Value( pMetaClass->reactivityDegree );
    if (nReactDegree) {  /// If it is 0, there is no need to execute optional parts
#endif

      rt_slack_available (new_task->priority, &slack_time);

      /***/
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
#ifdef ARTIS_CONF_RS_DEBUG_1
      rtl_printf("(RSERVER) [rtl_schedule] Calculating real slack time for react. degree %d y slack %lld\n", nReactDegree, slack_time);
#endif

      /***
      if ((nOldReactDegree != nReactDegree) && (nOldReactDegree)) {
        /// If the React. Degree is new, update the slack
        ///OldSlack = 
        slack_time =
          (systime_t) ((double) slack_time / (double) nOldReactDegree
                       * nReactDegree );
      }      
      ***/
      

      ///OldSlack = 
      slack_time = (systime_t) ((double) slack_time)/ (double) 100 * nReactDegree;

      /// slack_time = (systime_t) ((double) slack_time)/ (double) 100 * nReactDegree;
#endif
      /***/
      
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
#ifdef ARTIS_CONF_RS_DEBUG_1
      rtl_printf("(RSERVER) [rtl_schedule] After Calculating real slack time for react. degree %d y slack %lld\n", nReactDegree, slack_time);
#endif
#endif

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
      trace_event (SLACK_AVAILABLE, new_task->identifier, slack_time, 0);
#endif
#endif

#ifdef ARTIS_CONF_RS_DEBUG_2
      rtl_printf("(RSERVER) [rtl_schedule] {%d} Task %d, Optional %d ready. Slack=%d\n",
                 (int) (systime_gettime()-ini_time), 
		 new_task->identifier,
		 new_task->basepriority, 
                 slack_time);
#endif
    
      if (slack_time < MIN_SLACK_TO_RUN_OPTIONAL) {
      // Not enough slack --> run the task's final part
        new_task->state = RT_TASK_FINAL;
      }
      else {
        /* 
           OK, there is slack available, so we put Linux to run.  

           However, the next preemption time may be due to a task
           activation, so the slack *interval* may be lower than this
           minimum. If so, Linux will run but the DSERVER is not waken up
           (by sending the message through the SCHED_RTFIFO).
        */

        preemp_task = new_task;

        time_until_preemption = SUBTRACT (preempt_time, now);
        slack_interval        = MIN (slack_time, time_until_preemption);
        new_task              = &rtl_linux_task;
        preempt_time          = now + slack_interval;

      }
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
    }
#endif
  }  /// if (new_task != &rtl_linux_task && new_task->state == RT_TASK_OPTIONAL) 
  /// Ending there the above mentioned if gets that the DS can be executed also after all the
  ///   tasks have ended its execution

  if ((new_task == &rtl_linux_task) && (!bWakeUpbyFIFO)) { 
    
    if (slack_interval >  MIN_SLACK_TO_RUN_OPTIONAL) {
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
      if (nReactDegree == -1)
        nReactDegree = * (int *) KDM_Get_Value( pMetaClass->reactivityDegree );
      trace_event (SEND_SCHED_FIFO,0, 0, (long long) nReactDegree);
#endif
#endif
#endif

#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
#ifdef ARTIS_CONF_RS_DEBUG_1
      rtl_printf("(RSERVER) [rtl_schedule] Task %d - React. degree %d y slack %lld\n", 
                 new_task->identifier, nReactDegree, slack_time);
#endif
#endif

      sched_mssg.slack_interval_time = slack_interval;
      sched_mssg.TimeStamp           = now - ini_time;
      sched_mssg.magic               = SCHED_MSSG_MAGIC;
      rtf_put (SCHED_RTFIFO, (char *) &sched_mssg, sizeof (sched_mssg_t));





#ifdef ARTIS_CONF_METAREASONING_BEHAVS
      if (nFinalsExecuted == nBehTasks) {
        /**
         *  Due to the FLS scheduling algorithm, all the final components
         *     have ended its execution, and we are going to call again the DS,
         *     before activating a new initial part, so to make the DS know 
         *     all of this the best thing to do is to try to execute the max.
         *     number of events possible
         *
         */
        MR_Action_Set_Attention_Degree( 1.0 );
      }
#endif





    }
  }
  
  
  bWakeUpbyFIFO = 0;

#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  if ( nFinalsExecuted == nBehTasks )
    nFinalsExecuted = 0;
#endif

  // *** Step 4) Put the 'new_task' to run until the 'preempt_time'
  rtl_running = new_task;

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
  trace_event (CONTEXT_SW, new_task->identifier, 0, 0);
#endif
#endif


#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(RSERVER) [rtl_schedule] Before dispatching task %d, state %d, slack %lld...\n", 
             new_task->identifier, task->state, new_task->slack);
#endif

  /***/
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  nOldReactDegree = nReactDegree;
#endif
/***/


  rtl_dispatch (new_task->context, preempt_time);
}
#else
/* ==================================================================== */
void
rtl_schedule (void)
/* ==================================================================== */
{
  systime_t time_until_preemption;
  systime_t slack_interval = 0;
  systime_t slack_time = 0;
  rt_task_t *task;
  rt_task_t *new_task;


  // *** Step 1) Selection of new running task
  new_task = &rtl_linux_task;
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next)
    if (task->state == RT_TASK_INITIAL ||
	task->state == RT_TASK_OPTIONAL || task->state == RT_TASK_FINAL)
      if (task->priority <= new_task->priority) {
	new_task = task;
      }


  // *** Step 2) Look for the nearest preemption time
  preempt_time = RT_TIME_END;
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    if (task->state == RT_TASK_DELAYED && task->priority < new_task->priority
	&& task->resume_time < preempt_time) {
      preempt_time = task->resume_time;
    }
  }


  // *** Step 3) If there is an optional part to run, check slack available
  if (new_task != &rtl_linux_task && new_task->state == RT_TASK_OPTIONAL) {
    rt_slack_available (new_task->priority, &slack_time);
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_1
    trace_event (SLACK_AVAILABLE, new_task->identifier, slack_time, 0);
#endif
#endif

#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RSERVER) [rtl_schedule] {%d} Optional %d ready. Slack=%d\n",
             (int) (systime_gettime()-ini_time), new_task->basepriority, 
             slack_time);
#endif
    
    if (slack_time < MIN_SLACK_TO_RUN_OPTIONAL) {
      // Not enough slack --> run the task's final part
      new_task->state = RT_TASK_FINAL;
    }
    else {
      /* 
         OK, there is slack available, so we put Linux to run.  

         However, the next preemption time may be due to a task
         activation, so the slack *interval* may be lower than this
         minimum. If so, Linux will run but the DSERVER is not waken up
         (by sending the message through the SCHED_RTFIFO).
       */
      time_until_preemption = SUBTRACT (preempt_time, now);
      slack_interval        = MIN (slack_time, time_until_preemption);
      new_task              = &rtl_linux_task;
      preempt_time          = now + slack_interval;

      if (slack_interval >  MIN_SLACK_TO_RUN_OPTIONAL) {
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
        trace_event (SEND_SCHED_FIFO,0, 0, 0);
#endif
#endif
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
        trace_event (SLACK_AVAILABLE, new_task->identifier, slack_time, 0);
#endif
#endif
	sched_mssg.slack_interval_time = slack_interval;
        sched_mssg.TimeStamp           = now - ini_time;
        sched_mssg.magic               = SCHED_MSSG_MAGIC;
        rtf_put (SCHED_RTFIFO, (char *) &sched_mssg, sizeof (sched_mssg_t));
      }
    }
  }

  // *** Step 4) Put the 'new_task' to run until the 'preempt_time'
  rtl_running = new_task;

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
  trace_event (CONTEXT_SW, new_task->identifier, 0, 0);
#endif
#endif

  rtl_dispatch (new_task->context, preempt_time);
}
#endif /// #ifdef ARTIS_CONF_METAREASONING


/* ========================================================================= */
static void
rtl_sched_timer_interrupt (struct pt_regs *regs)
/* ========================================================================= */
{
  unsigned long interrupt_state;

  if (app_started == FALSE)   	  
  {	  
    return;
  }

  /*
  if (systime_gettime () < preempt_time) {
#ifdef ARTIS_TRACE_EVENTS
  trace_event (ADVANCED_INT, 0, 0, 0);
#endif
    return;
  }
  */
  if (preempt_time > systime_gettime() + ADJUSTUP) {
    rtl_no_interrupts(interrupt_state);
    rtl_program_timer(preempt_time);  // in rtl_core.c
    rtl_restore_interrupts(interrupt_state);
    return;
  }

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_1
  trace_event (INTERRUPT, 0, 0, 0);
#endif
#endif

  rtl_update ();
  rtl_schedule ();
}


#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_WAKEUP
/* ========================================================================= */
int
fifo_handler (unsigned int fifo)
{
/* ========================================================================= */
  char c[3]; /// To read from the wake up buffer

#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  int nReactDegree;
#endif



#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RTL) fifo_handler():: Beginning\n");
#endif

  rtl_update ();



#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  nReactDegree = * (int *) KDM_Get_Value( pMetaClass->reactivityDegree );
#endif



  if (read( fdWakeUpRTFIFO, c, sizeof("3C") ) != sizeof("3C")) {
    rtl_printf("(RTL) ERROR: fifo_handler() ==> was not able to read\n");
  }
  else if ((c[0] == '3') ///< DSleeping beauty
           || (nReactDegree <= nOldReactDegree)) { ///< DS wants more time, but we are not giving any
         bWakeUpbyFIFO = 1;   
         /// To indicate the rtl_schedule() that it has been waked up this way

         /// To advance the final parts
         if (preemp_task != NULL) {
             preemp_task->state = RT_TASK_FINAL;

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_1
             trace_event (ADVANCE_FINAL, preemp_task->identifier, 0, 0);
#endif
#endif
         }
#ifdef ARTIS_CONF_RS_DEBUG_2
         else rtl_printf("(RTL) fifo_handler():: Preemp_Task == NULL!!!!!!\n");
#endif



#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(RTL) fifo_handler():: Wake up, but not executing again: Old RD: %d, current RD: %d\n"
             , nOldReactDegree, nReactDegree);
#endif




#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
         trace_event (WAKEUP_FIFO, 0, 0, 0);
#endif
#endif
  }

#ifdef ARTIS_CONF_RS_DEBUG_2
  else rtl_printf("(RTL) fifo_handler():: DS needs more time\n");
#endif

  rtl_schedule ();
  
  return 0;
}
#endif /// #ifdef ARTIS_CONF_METAREASONING_WAKEUP
#endif /// #ifdef ARTIS_CONF_METAREASONING


/* ==================================================================== */
int
rtl_task_create (rt_task_t * task,
                 struct task_create_info *info)
/* ===================================================================== */
{
  int        error;
  rt_task_t *prev, *post;
  long       interrupt_state;
  int        stack_size = 8000;

  if (task->magic == RT_TASK_MAGIC) {
    rtl_printf("(rserver) error: Task already created for *task!!\n");
    return -EINVAL;
  }

  if (info->deadline > info->period) {
    rtl_printf ("(rserver) error: deadline > period for task %d\n",
                info->identifier); 
    return -EINVAL;
  }

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(RSERVER) [task_create] No errors. Goto fill fields\n");
#endif
 
  task->magic            = RT_TASK_MAGIC;
  task->priority         = info->priority;
  task->basepriority     = info->priority;
  task->identifier       = info->identifier;
  task->initial_function = info->initial_function;
  task->initial_data     = info->initial_parameters;
  task->final_function   = info->final_function;
  task->final_data       = info->final_parameters;
  task->num_optionals    = (info->has_optionals ? 1 : 0);
  // Static temporal attributes
  task->offset            = info->offset;
  task->period            = info->period;
  task->deadline          = info->deadline;
  task->initial_wcet      = info->initial_wcet;
  task->final_wcet        = info->final_wcet;
  task->blocking_time     = info->blocking_time;
  // Dynamic temporal attributes (they're reset in rt_start_app below)
  task->resume_time       = RT_TIME_END;
  task->absolute_deadline = RT_TIME_END;
  task->execution_left    = (systime_t) 0;
  // ATB Study if execution left = 0 is correct for initial slack calculation
  task->slack             = (systime_t) 0;
  // Initial state: DORMANT (cannot be awakened until rtl_start_application)
  task->state             = RT_TASK_DORMANT;
#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(RSERVER) [task_create] Fields filled. Goto create context\n");
#endif

  // Create the context for the task
  if ((error = rtl_context_create (rt_task_runner,
				   (int) task, stack_size, &(task->context))))
    return error;

  // By default, consider all tasks using  FP operations
  rtl_context_use_fp (task->context, 1);
#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(RSERVER) [task_create] Context created. Goto insert task\n");
#endif

  // Insert task into priority-ordered task list
  rtl_no_interrupts (interrupt_state);
  for (prev = post = rt_tasks;
       post->basepriority <= task->basepriority 
         && post->next != (rt_task_t *) 0; 
       prev = post, post = post->next)
    /* void */;

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(RSERVER) [task_create] Located position. Now inserting\n");
#endif
  task->next = post;
  if (post == rt_tasks) {
    rt_tasks = task;
  }
  else {
    prev->next = task;
  }
  rtl_restore_interrupts (interrupt_state);

#ifdef ARTIS_CONF_RS_DEBUG_1
  _debug_print_tasks();
  rtl_printf("(RSERVER) [task_create] All done. Exiting\n");
#endif

  return 0;
}


#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_BEHAVS
int rtl_behav_create( rt_behav_t *pBeh, rt_behav_task_t *aTaskBeh,
                      int nTasks, ... ) {
  va_list ap;
  long interrupt_state;
  int i, j;
  rt_behav_task_t *prev, *post, *pBehTask;
  rt_behav_t *pPrev, *pPost;
    
  rtl_no_interrupts (interrupt_state);

  va_start( ap, nTasks );

  if (rt_behav == (rt_behav_t *) NULL)
    rt_behav = pBeh;  /// < 1st behaviour of the AA
  else {
	for ( pPrev = rt_behav, pPost = rt_behav->pNext;
		(pPost != (rt_behav_t *) NULL);
	      pPrev = pPost, pPost = pPost->pNext );
	pPrev->pNext = pBeh;
  }
    
  pBeh->task_list = aTaskBeh;
  pBeh->nTasks = nTasks;
  pBeh->pNext = (rt_behav_t *) NULL;

  pBeh->task_list[0].task = va_arg( ap, rt_task_t * );
  pBeh->task_list[0].pNext = NULL;

#ifdef ARTIS_CONF_RS_DEBUG_2
    rtl_printf("(RTL) rserver.c:: rtl_behav_create: 1st Task: %d, prio: %d\n",
               pBeh->task_list[0].task->identifier, 
               pBeh->task_list[0].task->priority );
#endif

  /// insert tasks into priority ordered task list
  for ( i = 1; i < nTasks; i++ ) {
    pBeh->task_list[i].task = va_arg( ap, rt_task_t * );
    pBehTask = &(pBeh->task_list[i]);
  
#ifdef ARTIS_CONF_RS_DEBUG_2
    rtl_printf("(RTL) rserver.c:: rtl_behav_create: Before inserting Task: %d, pos: %d, priority: %d\n",
               pBehTask->task->identifier, i, pBehTask->task->priority);
#endif

    for (prev = post = pBeh->task_list, j = 0;
         ( j < i ) && ( post->task->priority <= pBehTask->task->priority ); 
         prev = post, post = post->pNext, j++);
    
#ifdef ARTIS_CONF_RS_DEBUG_2
    rtl_printf("(RTL) rtl_sched_artis.c:: rt_task_in_beh Task: %d, pos: %d\n",
               pBehTask->task->identifier, i);
#endif
    
    pBehTask->pNext = post;
    
    if (post == pBeh->task_list) {
      pBeh->task_list = pBehTask;
    }
    else {
      prev->pNext = pBehTask;
    }
  }

#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RTL) rt_task_in_beh, the beh. has now %d tasks\n",
             pBeh->nTasks);

  for ( i = 0, post = pBeh->task_list; 
        i < pBeh->nTasks;
        i++, post = post->pNext )
    rtl_printf("(RTL) rt_task_in_beh: Task: id: %d, priority: %d\n", 
               post->task->identifier, post->task->priority );
#endif

  va_end( ap );
  rtl_restore_interrupts (interrupt_state);  

  return 0;
}
#endif
#endif


#ifdef ARTIS_CONF_METAREASONING
/* ========================================================================= */
int
rtl_start_application (void)
/* ========================================================================= */
{
  rt_task_t *task;
  long interrupt_state;

#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  int nReactDegree;
#endif

#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  rt_behav_t * rt_auxBeh;
  rt_behav_task_t *beh_task;
#endif

  int i;


  rtl_no_interrupts (interrupt_state);

  /***/
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  nReactDegree = * (int *) KDM_Get_Value( pMetaClass->reactivityDegree );
#endif
/***/


  // Set the linux task identifier
  rtl_linux_task.identifier = 1;
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    rtl_linux_task.identifier = MAX(rtl_linux_task.identifier,
                                    task->identifier);
  }
  rtl_linux_task.identifier += 1;

#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  nCurrentBehavior = * (int *) KDM_Get_Value( pMetaClass->nCurrentBehavior );

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(RTL) (sched_artis) start_application: Before setting the current behavior tasks, the current behavior id is %d %p\n", 
             nCurrentBehavior, rt_behav);
#endif

  /// rt_auxBeh must point to the current behavior
  for (i = 1, rt_auxBeh = rt_behav; i <= nCurrentBehavior; i++)
    rt_auxBeh = rt_auxBeh->pNext;

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(RTL) (sched_artis) start_application: Accessing to the proper behavior, with %d tasks\n", 
             rt_auxBeh->nTasks);
#endif

  /// Set the current Behavior tasks
  for (beh_task = rt_auxBeh->task_list, task = beh_task->task, 
         rt_tasks = task, task->nBehav = nCurrentBehavior, 
         nBehTasks = rt_auxBeh->nTasks, i = 0;
       i < nBehTasks - 1;
       i++, beh_task = beh_task->pNext, task->next = beh_task->task, 
         task = task->next, task->nBehav = nCurrentBehavior );

  task->next = &rtl_linux_task;

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(RTL) (sched_artis) start_application: After setting the current behavior tasks\n");
#endif

#endif ///#ifdef ARTIS_CONF_METAREASONING_BEHAVS

  // Get initial time
  ini_time = systime_gettime ();
  now = ini_time;

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
  trace_start();
  // The number of tasks is the id. of Linux minus 1
  // This makes enough space in kiwi to display all potential
  // task ids present in the application, even though some of
  // them may not be in use.
  trace_event(NTASKS, rtl_linux_task.identifier-1, 0, 0);
#endif
#endif

  // All tasks will be active
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
#ifdef ARTIS_CONF_RS_DEBUG_1
    rtl_printf("(RTL) (sched_artis) start_application: now: %lld, offset: %lld\n",
		    now, task->offset);
#endif
    task->resume_time       = now + task->offset;
    ///task->resume_time       = (systime_t) (((double) now) + task->offset);
    task->absolute_deadline = task->resume_time + task->deadline;
    task->state             = RT_TASK_DELAYED;
  }

  // Calculate the initial slack time for each task 
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    rt_calculate_slack (task, &(task->slack));


/***
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
    if (!nReactDegree)
	    task->slack = (systime_t) 0;
    else task->slack =
            (systime_t) (((double) task->slack / (double) 100
                    * nReactDegree));
#endif
***/
    


#ifdef ARTIS_CONF_RS_DEBUG_1
    rtl_printf("(RSERVER) [start_app] task %d: slack=%d\n",
             (int) task->basepriority, (int) task->slack);
#endif
  }


  /***/
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  nOldReactDegree = nReactDegree;
#endif
/***/


#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  /// To indicate that there is no active transition between behavs.
  nLastBehavior = -1;  

  KDM_Set_Value ( pMetaClass->nLastBehavior, nLastBehavior );
  nOldBehTasks = 0; 
#endif

  // Start running the application: update + schedule 
  last_kernel_time = ini_time;
  last_update      = ini_time;
  app_started      = TRUE;
  rtl_restore_interrupts (interrupt_state);

  rtl_update ();
  rtl_schedule ();
  return 0;
}
#else
/* ========================================================================= */
int
rtl_start_application (void)
/* ========================================================================= */
{
  rt_task_t *task;
  long interrupt_state;

  rtl_no_interrupts (interrupt_state);

  // Set the linux task identifier
  rtl_linux_task.identifier = 1;
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    rtl_linux_task.identifier = MAX(rtl_linux_task.identifier,
                                    task->identifier);
  }
  rtl_linux_task.identifier += 1;

  // Get initial time
  ini_time = systime_gettime ();
  now = ini_time;

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
  trace_start();
  // The number of tasks is the id. of Linux minus 1
  // This makes enough space in kiwi to display all potential
  // task ids present in the application, even though some of
  // them may not be in use.
  trace_event(NTASKS, rtl_linux_task.identifier-1, 0, 0);
#endif
#endif

  // All tasks will be active
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    task->resume_time       = now + task->offset;
    task->absolute_deadline = task->resume_time + task->deadline;
    task->state             = RT_TASK_DELAYED;
  }

  // Calculate the initial slack time for each task 
  for (task = rt_tasks; task != &rtl_linux_task; task = task->next) {
    rt_calculate_slack (task, &(task->slack));
#ifdef ARTIS_CONF_RS_DEBUG_1
    rtl_printf("(RSERVER) [start_app] task %d: slack=%d\n",
             (int) task->basepriority, (int) task->slack);
#endif
  }

  // Start running the application: update + schedule 
  last_kernel_time = ini_time;
  last_update      = ini_time;
  app_started      = TRUE;
  rtl_restore_interrupts (interrupt_state);

  rtl_update ();
  rtl_schedule ();
  return 0;
}
#endif /// #ifdef ARTIS_CONF_METAREASONING



/* ========================================================================= */
void 
rtl_end_application (void)
/* ========================================================================= */
{
#ifdef ARTIS_CONF_RS_DEBUG_1
    rtl_printf("(RSERVER) [rtl_end_application]\n");
    _debug_print_tasks();
#endif
    rt_end_scheduler();
}


#ifdef ARTIS_CONF_METAREASONING
/* ========================================================================= */
static int
rt_task_wait (void)
/* ========================================================================= */
{
  rt_task_t *task;
  systime_t  gain_time;
  systime_t task_slack;


  /***/
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  int nReactDegree;
#endif
/***/


  rt_task_t *aux_task, *prev;

#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  rt_behav_t *rt_auxBeh;
  rt_behav_task_t *beh_task;
  int nBehav;
#endif

  int i;


  rtl_update ();


#ifdef ARTIS_CONF_METAREASONING_BEHAVS
#ifdef ARTIS_CONF_RS_DEBUG_2
  if (nLastBehavior != -1)
    rtl_printf("(RTL) rt_task_wait:: After rtl_update in a beh. transition\n");
#endif
#endif

  /*
     This function is called each time a task ends its initial or its
     final part.

     Common activity: account for the task's gain time (=execution time
     left that becomes slack time for lower priority tasks)
   */

  /***/
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
  nReactDegree = * (int *) KDM_Get_Value( pMetaClass->reactivityDegree );
#endif
/***/

  if (rtl_running->state == RT_TASK_INITIAL) {
    gain_time = SUBTRACT (rtl_running->execution_left,
			  rtl_running->final_wcet);

/***
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
    if (!nReactDegree)
	    gain_time = (systime_t) 0;
    else gain_time = (systime_t) ((double) gain_time / (double) 100 * nReactDegree);
#endif
***/

    if (gain_time) {
      for (task = rtl_running; task != &rtl_linux_task; task = task->next) {
	task->slack += gain_time;
#ifdef ARTIS_CONF_RS_DEBUG_1
        rtl_printf("(RSERVER) [rt_task_wait] SLACK: %lld Gain: %lld React. Degree: %d\n", 
                           task->slack, gain_time, nReactDegree);
#endif
      }
    }

    /* Change the task status */
    rtl_running->execution_left = rtl_running->final_wcet;
    rtl_running->state          = rtl_running->num_optionals > 0 ?
                                  RT_TASK_OPTIONAL : RT_TASK_FINAL;
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
    trace_event (INITIAL_END, rtl_running->identifier, 0, 0);
#endif
#endif

    /* 
       Send the 'CONSIDER_PERIODIC_AGENT' event to the 2nd.level sched. 
       in order to activate the task's optional levels.
     */
#ifdef ARTIS_CONF_METAREASONING_BEHAVS
    if (rtl_running->nBehav != nLastBehavior)
#endif
      /// If it is not in a behavior transition, or the task belongs to the new behavor
      if (rtl_running->num_optionals > 0) {
        rt_slack_available (rtl_running->priority, &task_slack);
        Event_Notify( CONSIDER_PERIODIC_AGENT, (void *) NULL, MAX_IMPORTANCE_EV - 2, TIME_VALUE,
                      (long long) task_slack, (KDM_Value) -1, 
                      (long long) rtl_running->identifier, (KDM_Value) -1 );
///                      (long long) rtl_running->optional_func_id, (KDM_Value) -1 );
      }
  }

  else { /* RT_TASK_FINAL */

    /* Gain time */
///    gain_time = SUBTRACT (rtl_running->execution_left,
///			xs  rtl_running->final_wcet);
    gain_time = rtl_running->execution_left;
	
    if (gain_time)
      for (task = rtl_running; task != &rtl_linux_task; task = task->next) {
	task->slack += gain_time;
#ifdef ARTIS_CONF_RS_DEBUG_1
        rtl_printf("(RSERVER) [rt_task_wait] Gain: %lld SLACK: %lld React. Degree: %d\n", 
                   gain_time, task->slack, nReactDegree);
#endif
      }

    /* Change the task status */
    rtl_running->state              = RT_TASK_DELAYED;
    rtl_running->execution_left     = 0;
    nFinalsExecuted++;

#ifdef ARTIS_CONF_METAREASONING_BEHAVS
    if (rtl_running->nBehav != nLastBehavior)
#endif
      /** 
       * If it is not in a behavior transition, or the task belongs to the
       *       new behavior
       */
      rtl_running->absolute_deadline += rtl_running->period;

#ifdef ARTIS_CONF_METAREASONING_BEHAVS
    else { ///< Transition of behaviors -> 'delete' the task
#ifdef ARTIS_CONF_RS_DEBUG_2
      rtl_printf("(RTL) rtl_task_wait:: Trans. of behav. --> 'Delete' task %d, Last Behav. %d, New Behav: %d, nOldBehTasks:%d, Next Task: %d, Next Task Beh: %d\n",
                 rtl_running->identifier, nLastBehavior, nCurrentBehavior, nOldBehTasks, rtl_running->next->identifier, rtl_running->next->nBehav);
#endif

      if (rt_tasks == rtl_running)
        rt_tasks = rtl_running->next;
      else {
        for (aux_task = rt_tasks; aux_task->next != rtl_running; 
             aux_task = aux_task->next);

        aux_task->next = rtl_running->next;
      }

      nOldBehTasks--;     

      /// This is the place where the thread/task rtl_running should be reset
 
      /// Synchronous MCProtocol: 
      ///   If there is not any more previous tasks, insert the new beh. ones
      if (!nOldBehTasks) {

#ifdef ARTIS_CONF_RS_DEBUG_2
        rtl_printf("(RTL) rt_task_wait:: There is no previous task left\n");
#endif

        nBehav = * (int *) KDM_Get_Value( pMetaClass->nCurrentBehavior );

        /// rt_auxBeh must point to the new behavior
        for (i = 0, rt_auxBeh = rt_behav; i < nBehav; i++)
          rt_auxBeh = rt_auxBeh->pNext;

        /// Insert the new beh. tasks
        for (beh_task = rt_auxBeh->task_list, 
               prev = task = rt_tasks, i = 0;
             i < nBehTasks;
             beh_task = beh_task->pNext, i++) {  

#ifdef ARTIS_CONF_RS_DEBUG_2
          rtl_printf("(RTL)rt_task_wait:: Trying to insert new beh.task %d\n",
                     beh_task->task->identifier);
#endif
          
          /// Look for the proper place for beh_task
          for ( ; 
                ((task != &rtl_linux_task) && 
                 (task->priority < beh_task->task->priority));
                prev = task, task = task->next );
          
          if (task != beh_task->task) {
            beh_task->task->next = task;

#ifdef ARTIS_CONF_RS_DEBUG_2            
            rtl_printf("(RTL)rt_task_wait:: beh.task %d, prev %d, task %d\n",
                     beh_task->task->identifier, prev->identifier, 
                       task->identifier);
#endif

            if (task == rt_tasks) {
              rt_tasks = beh_task->task;
            }
            else {
              prev->next = beh_task->task;
            }
          
            prev = beh_task->task;

            beh_task->task->nBehav = nBehav;
            beh_task->task->priority = beh_task->task->basepriority;
            beh_task->task->resume_time = 
              systime_gettime () + beh_task->task->offset;
            beh_task->task->absolute_deadline = 
              beh_task->task->resume_time + beh_task->task->deadline;
            beh_task->task->state = RT_TASK_DELAYED;
            rt_calculate_slack (beh_task->task, &(beh_task->task->slack));

/***	    
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
	    if (!nReactDegree)
		    beh_task->task->slack = (systime_t) 0;
	    else beh_task->task->slack =
                   (systime_t) (((double) beh_task->task->slack / (double) 100
                        * nReactDegree ));
#endif
***/

          }
        }

/***	
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
        nOldReactDegree = nReactDegree;
#endif
***/
      
#ifdef ARTIS_CONF_RS_DEBUG_2
        for (task = rt_tasks; task != &rtl_linux_task; task = task->next)
          rtl_printf("(RTL) rt_task_wait:: Task: %d \n");
#endif

        /// End of Beh. Transition
        nLastBehavior = -1;
        KDM_Set_Value( pMetaClass->nLastBehavior, nLastBehavior );
        
        MR_Action_Set_Attention_Degree( fOldAttentionDegree );
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
        trace_event (MCR_END, 0, 0, 0);
#endif
#endif

      }
    }
#endif ///#ifdef ARTIS_CONF_METAREASONING_BEHAVS

    /// Send the 'CONSIDER_PERIODIC_AGENT' event to the 2nd.level sched. 
    ///   in order to DE-activate the task's optional levels.
    if (rtl_running->num_optionals > 0) {
      Event_Notify( CONSIDER_PERIODIC_AGENT, (void *) NULL, MAX_IMPORTANCE_EV - 2, TIME_VALUE,
                    (long long) -1, (KDM_Value) -1, 
                    (long long) rtl_running->identifier, (KDM_Value) -1 );

    }

    /// Recalculate the task's slack time
    rt_calculate_slack (rtl_running, &(rtl_running->slack));

/***    
#ifdef ARTIS_CONF_METAREASONING_REACT_DEGREE
    nReactDegree = * (int *) KDM_Get_Value( pMetaClass->reactivityDegree );
    if (nReactDegree < 100) {
	    if (!nReactDegree)
		    rtl_running->slack = (systime_t) 0;
	    else rtl_running->slack =
        		(systime_t) (((double) (rtl_running->slack / (double) 100
                            * nReactDegree )));
#ifdef ARTIS_CONF_RS_DEBUG_0
            rtl_printf("(RSERVER) [rt_task_wait] SLACK: %lld React. Degree: %d\n", 
                           rtl_running->slack, nReactDegree);
#endif
    }
#endif
***/


#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(RSERVER) [task_wait] Final part completed. Slack=%lld\n",
             rtl_running->slack);
#endif

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
    trace_event (ACTIV_END, rtl_running->identifier, 0, 0);
#endif
#endif
  }
  

  rtl_schedule ();

  return 0;
}
#else
/* ========================================================================= */
static int
rt_task_wait (void)
/* ========================================================================= */
{
  rt_task_t *task;
  systime_t  gain_time;
  systime_t task_slack;

  rtl_update ();

  /*
     This function is called each time a task ends its initial or its
     final part.

     Common activity: account for the task's gain time (=execution time
     left that becomes slack time for lower priority tasks)
   */

  if (rtl_running->state == RT_TASK_INITIAL) {
    gain_time = SUBTRACT (rtl_running->execution_left,
			  rtl_running->final_wcet);
    if (gain_time) {
      for (task = rtl_running; task != &rtl_linux_task; task = task->next) {
	task->slack += gain_time;
      }
    }

    /* Change the task status */
    rtl_running->execution_left = rtl_running->final_wcet;
    rtl_running->state          = rtl_running->num_optionals > 0 ?
                                  RT_TASK_OPTIONAL : RT_TASK_FINAL;
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
    trace_event (INITIAL_END, rtl_running->identifier, 0, 0);
#endif
#endif

    /* 
       Send the 'CONSIDER_PERIODIC_AGENT' event to the 2nd.level sched. 
       in order to activate the task's optional levels.
     */
    if (rtl_running->num_optionals > 0) {
      rt_slack_available (rtl_running->priority, &task_slack);
      Event_Notify( CONSIDER_PERIODIC_AGENT, (void *) NULL, MAX_IMPORTANCE_EV - 2, TIME_VALUE,
                    (long long) task_slack, (KDM_Value) -1, 
                    (long long) rtl_running->identifier, (KDM_Value) -1 );
    }
  }

  else { /* RT_TASK_FINAL */

    /* Gain time */
    gain_time = SUBTRACT (rtl_running->execution_left,
			  rtl_running->final_wcet);
    if (gain_time)
      for (task = rtl_running; task != &rtl_linux_task; task = task->next) {
	task->slack += gain_time;
      }

    /* Change the task status */
    rtl_running->state              = RT_TASK_DELAYED;
    rtl_running->execution_left     = 0;
    rtl_running->absolute_deadline += rtl_running->period;

    /* 
       Send the 'CONSIDER_PERIODIC_AGENT' event to the 2nd.level sched. 
       in order to DE-activate the task's optional levels.
     */
    if (rtl_running->num_optionals > 0) {
      Event_Notify( CONSIDER_PERIODIC_AGENT, (void *) NULL, MAX_IMPORTANCE_EV - 2, TIME_VALUE,
                    (long long) -1, (KDM_Value) -1, 
                    (long long) rtl_running->identifier, (KDM_Value) -1 );

    }

    /* Recalculate the task's slack time */
    rt_calculate_slack (rtl_running, &(rtl_running->slack));
#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RSERVER) [task_wait] Final part completed. Slack=%d\n",
             rtl_running->slack);
#endif

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
    trace_event (ACTIV_END, rtl_running->identifier, 0, 0);
#endif
#endif
  }
  
  rtl_schedule ();
  return 0;
}
#endif /// #ifdef ARTIS_CONF_METAREASONING


/* ========================================================================= */
int
rt_task_delete (rt_task_t * task)
/* ========================================================================= */
{
  rt_task_t *t;
  int found = 0;
  long interrupt_state;

  rtl_no_interrupts (interrupt_state);

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c:rt_task_delete) Deleting task %d\n", task->identifier);
#endif

  if (task->magic != RT_TASK_MAGIC) {
    rtl_printf ("(RSERVER) [rt_task_delete]: bad magic\n");
    rtl_restore_interrupts (interrupt_state);
    return -EINVAL;
  }

#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  if ( task->nBehav == nCurrentBehavior ) {
#endif
#endif
	if (task != rt_tasks) {
    		for (t = rt_tasks; t; t = t->next) {
      			if (t->next == task) {
				t->next = task->next;
				found = 1;
				break;
      			}
    		}
    		if (!found) {
      			rtl_printf ("(rt_task_delete: task not found\n");
      			rtl_restore_interrupts (interrupt_state);
      			return -EINVAL;
    		}
  	}
  	else {
    		rt_tasks = task->next;
  	}

#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_BEHAVS
}
#endif
#endif

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c:rt_task_delete) Still Deleting task %d\n", task->identifier);
#endif

  task->magic = 0;
  rtl_context_destroy (task->context);

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c:rt_task_delete) Before ending\n");
#endif

  rtl_restore_interrupts (interrupt_state);

  return 0;
}


/* ========================================================================= */
static int
rt_slack_available (int prio_level, systime_t * available_slack)
/* ========================================================================= */
{
  rt_task_t *task;

  *available_slack = (systime_t) MAXLONG;

  for (task = rt_tasks;
       task != &rtl_linux_task && task->basepriority < prio_level;
       task = task->next)
    /* void */;

  for (; task != &rtl_linux_task; task = task->next) {
#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf("(RSERVER) [slack_available] task %d:",
             (int) task->basepriority);
#endif
    if (*available_slack > task->slack) {
      *available_slack = MAX((systime_t) (task->slack - task->blocking_time),
                             (systime_t) 0);
    }
  }

  if (*available_slack == (systime_t) MAXLONG) {
    *available_slack = (systime_t) 0;
  } else {
    *available_slack -= OVERHEAD_FACTOR;
  }

  
#ifdef ARTIS_CONF_RS_DEBUG_2
  rtl_printf(" Slack available for prio %d =%d\n",
             prio_level, *available_slack);
#endif
 
  return 0;
}


/* ========================================================================= */
static int
rt_calculate_slack (rt_task_t * the_task, systime_t * result)
/* ========================================================================= */
{
  rt_task_t *task;
  systime_t the_deadline, aux, Si = 0, Ij = 0, SumIj = 0,
    cj = 0, Cj = 0, Tj = 0, fj = 0, xj = 0;

  the_deadline = the_task->absolute_deadline;

  /* Calculation of the effective deadline. Corrected from original
     R. Davis' version, by  Dr. Ana Garcia-Fornes.                  */

  
  for (task = rt_tasks;
       task->priority <= the_task->priority; task = task->next) {

    Cj = task->initial_wcet + task->final_wcet;
    Tj = task->period;
    xj = task->resume_time;
    fj = (systime_t) ((float) (the_task->absolute_deadline - xj) / (float) Tj);
    if (fj > 0) {
      aux = xj + (fj * Tj) + Cj;
      if (aux >= the_deadline)
	the_deadline = xj + (fj * Tj);
    }
  }
  

  /* Now calculate the slack time from now until the_deadline */
  for (task = rt_tasks;
       task->priority <= the_task->priority; task = task->next) {

    Cj = task->initial_wcet + task->final_wcet;
    cj = task->execution_left;
    Tj = task->period;
    xj = task->resume_time;

    fj = (systime_t) ((double) (the_deadline - xj) / (double) Tj);
    aux = the_deadline - xj - (fj * Tj);
    aux = MAX (0, aux);
    aux = MIN (aux, Cj);

    Ij = cj + (fj * Cj) + aux;

    SumIj += Ij;
  }

  //This is a relative result.
  Si = MAX ((the_deadline - systime_gettime ()) - SumIj, 0);

  *result = Si;
  return 0;
}


#ifdef ARTIS_CONF_METAREASONING
/* ========================================================================= */
void
rt_end_scheduler (void)
/* ========================================================================= */
{
  rt_task_t *t;
  rt_task_t *next_t;
  long interrupt_state;

#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  rt_behav_t *rt_auxBeh;
  rt_behav_task_t *beh_task, *next_beh_task;
  int i;
#endif

  rtl_no_interrupts (interrupt_state);

  /* 
     Tell the 2nd scheduler to exit the application. This is a event
     message plus a sched message. The first informs it to exit, the
     second lets it run.
   */

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c::rt_end_scheduler) Before Event_Notify: It is the end\n");  
#endif

  Event_Notify( EXIT_APP, (void *) NULL, MAX_IMPORTANCE_EV - 1, TIME_VALUE,
                (long long) -1, (KDM_Value) -1, 
                (long long) -1, (KDM_Value) -1 );

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c:rt_end_scheduler) After Event_Notify: It is the end\n");
#endif
  
  sched_mssg.slack_interval_time = (systime_t) 10000000000;
  sched_mssg.magic = SCHED_MSSG_MAGIC;
  rtf_flush (SCHED_RTFIFO);
  rtf_put (SCHED_RTFIFO, (char *) &sched_mssg, sizeof (sched_mssg_t));


  MR_Action_Set_Attention_Degree( 1.0 );

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c:rt_end_scheduler) Before eliminating tasks\n");
#endif

  

#ifdef ARTIS_CONF_METAREASONING_BEHAVS
  for ( rt_auxBeh = rt_behav; 
	rt_auxBeh != (rt_behav_t *) NULL; 
	rt_auxBeh = rt_auxBeh->pNext ) {
#ifdef ARTIS_CONF_RS_DEBUG_1
        rtl_printf("(rserver.c:rt_end_scheduler) Before eliminating beh.\n");
#endif
	for ( i = 0, beh_task = rt_auxBeh->task_list;
		i < rt_auxBeh->nTasks;
		i++, beh_task = next_beh_task ) {
#ifdef ARTIS_CONF_RS_DEBUG_1
                rtl_printf("(rserver.c:rt_end_scheduler) Before eliminating task %d\n",
			beh_task->task->identifier);
#endif
		next_beh_task = beh_task->pNext;
		if (beh_task->task != &rtl_linux_task) {
			rt_task_delete (beh_task->task);
		}
	}
}

rt_behav = (rt_behav_t *) NULL;
#endif

/***
  for (t = rt_tasks; t; t = next_t) {
    next_t = t->next;
    if (t != &rtl_linux_task) {
      rt_task_delete (t);
    }
  }
***/

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c:rt_end_scheduler) After eliminating tasks\n");
#endif

  rtl_free_timer ();
  rtl_restore_interrupts (interrupt_state);
}
#else
/* ========================================================================= */
void
rt_end_scheduler (void)
/* ========================================================================= */
{
  rt_task_t *t;
  rt_task_t *next_t;
  long interrupt_state;

  rtl_no_interrupts (interrupt_state);

  /* 
     Tell the 2nd scheduler to exit the application. This is a event
     message plus a sched message. The first informs it to exit, the
     second lets it run.
   */

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c::rt_end_scheduler) Before Event_Notify: It is the end\n");  
#endif

  Event_Notify( EXIT_APP, (void *) NULL, MAX_IMPORTANCE_EV - 1, TIME_VALUE,
                (long long) -1, (KDM_Value) -1, 
                (long long) -1, (KDM_Value) -1 );

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c:rt_end_scheduler) After Event_Notify: It is the end\n");
#endif
  
  sched_mssg.slack_interval_time = (systime_t) 10000000;
  sched_mssg.magic = SCHED_MSSG_MAGIC;
  rtf_flush (SCHED_RTFIFO);
  rtf_put (SCHED_RTFIFO, (char *) &sched_mssg, sizeof (sched_mssg_t));

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c:rt_end_scheduler) Before eliminating tasks\n");
#endif

  for (t = rt_tasks; t; t = next_t) {
    next_t = t->next;
    if (t != &rtl_linux_task) {
      rt_task_delete (t);
    }
  }

#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf("(rserver.c:rt_end_scheduler) After eliminating tasks\n");
#endif

  rtl_free_timer ();
  rtl_restore_interrupts (interrupt_state);
}
#endif


#ifdef ARTIS_CONF_METAREASONING
/* ========================================================================= */
int
rserver_init_module (void)
/* ========================================================================= */
{
  long interrupt_state;

#ifdef ARTIS_CONF_METAREASONING_WAKEUP
  char wakeup_rtf_name[30];
#endif

  rtl_no_interrupts (interrupt_state);
  rtf_create (SCHED_RTFIFO, 4000);

#ifdef ARTIS_CONF_METAREASONING_WAKEUP
  /// Creation of the rtf used to be waked up by the RS
  rtf_create ( ADVANCE_FINAL_PART_FIFO, 4000 );
  rtf_create_handler( ADVANCE_FINAL_PART_FIFO, fifo_handler ); 
  sprintf( wakeup_rtf_name, "/dev/rtf%d", ADVANCE_FINAL_PART_FIFO );
  if ((fdWakeUpRTFIFO = open(wakeup_rtf_name, O_RDONLY|O_NONBLOCK)) < 0) 
    rtl_printf("(RTL) ERROR-> wake up rtf was not created\n");
#endif

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
  trace_stop ();
#endif
#endif

  // Timer-related initialization:
  rtl_no_timer ();
  rtl_request_timer (rtl_sched_timer_interrupt);
  rtl_set_oneshot_mode ();

  // Create the "Linux task"
  rtl_context_create_linux (&(rtl_linux_task.context));
  rtl_context_use_fp (rtl_linux_task.context, 1);
  rtl_linux_task.identifier        = RTL_LINUX_ID;
  rtl_linux_task.next              = (rt_task_t *) 0;
  rtl_linux_task.num_optionals     = 0;
  rtl_linux_task.priority          = RTL_LOWEST_PRIORITY;
  rtl_linux_task.basepriority      = RTL_LOWEST_PRIORITY;
  rtl_linux_task.state             = RT_TASK_LINUX;
  rtl_linux_task.resume_time       = RT_TIME_END;
  rtl_linux_task.execution_left    = RT_TIME_END;
  rtl_linux_task.absolute_deadline = RT_TIME_END;
  rtl_linux_task.initial_wcet      = RT_TIME_END;
  rtl_linux_task.final_wcet        = RT_TIME_END;

  // Insert the task in the list of ready tasks (rt_tasks)
  rt_tasks    = &rtl_linux_task;
  rtl_running = &rtl_linux_task;

#ifdef ARTIS_CONF_RS_DEBUG_1
  _debug_print_tasks();
  rtl_printf ("(RSERVER) [rserver_init_module] All done!\n");
#endif
  //rtl_no_timer ();
  rtl_restore_interrupts (interrupt_state);

  return 0;
}
#else 
/* ========================================================================= */
int
rserver_init_module (void)
/* ========================================================================= */
{
  long interrupt_state;

  rtl_no_interrupts (interrupt_state);
  rtf_create (SCHED_RTFIFO, 4000);

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_RS_DEBUG_0
  trace_stop ();
#endif
#endif

  // Timer-related initialization:
  rtl_no_timer ();
  rtl_request_timer (rtl_sched_timer_interrupt);
  rtl_set_oneshot_mode ();

  // Create the "Linux task"
  rtl_context_create_linux (&(rtl_linux_task.context));
  rtl_context_use_fp (rtl_linux_task.context, 1);
  rtl_linux_task.identifier        = RTL_LINUX_ID;
  rtl_linux_task.next              = (rt_task_t *) 0;
  rtl_linux_task.num_optionals     = 0;
  rtl_linux_task.priority          = RTL_LOWEST_PRIORITY;
  rtl_linux_task.basepriority      = RTL_LOWEST_PRIORITY;
  rtl_linux_task.state             = RT_TASK_LINUX;
  rtl_linux_task.resume_time       = RT_TIME_END;
  rtl_linux_task.execution_left    = RT_TIME_END;
  rtl_linux_task.absolute_deadline = RT_TIME_END;
  rtl_linux_task.initial_wcet      = RT_TIME_END;
  rtl_linux_task.final_wcet        = RT_TIME_END;

  // Insert the task in the list of ready tasks (rt_tasks)
  rt_tasks    = &rtl_linux_task;
  rtl_running = &rtl_linux_task;

#ifdef ARTIS_CONF_RS_DEBUG_1
  _debug_print_tasks();
  rtl_printf ("(RSERVER) [rserver_init_module] All done!\n");
#endif
  //rtl_no_timer ();
  rtl_restore_interrupts (interrupt_state);

  return 0;
}
#endif /// #ifdef ARTIS_CONF_METAREASONING

/* ========================================================================= */
void
rserver_cleanup_module (void)
/* ========================================================================= */
{

  rtl_free_timer (); // Check if this here causes some havoc

  rtf_destroy (SCHED_RTFIFO);
  
#ifdef ARTIS_CONF_METAREASONING
#ifdef ARTIS_CONF_METAREASONING_WAKEUP
  rtf_destroy ( ADVANCE_FINAL_PART_FIFO );
#endif
#endif

  // everything is done in rt_end_scheduler;
#ifdef ARTIS_CONF_RS_DEBUG_1
  rtl_printf ("(RSERVER) [rserver_cleanup_module] All done!\n");
#endif
}


/*
 * Overrides for Emacs so that we almost follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
