/*

   FILE:
	rtl_core.h						

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-06

  CHANGELOG:


  DESCRIPTION:

    The RTL_CORE module provides a set of bsic support services for
    RT-Linux schedulers, in particular to the ARTIS' first-level
    scheduler. So this module mainly supports the RSERVER module.

    It hides a HAL (Hardware Abstraction Layer) and provides two basic
    interfaces:

      (1) Context management facilities which are independent of the
          scheduling policy (context creation and switching, etc.)

      (2) Basic access to the hardware timer. Since the time
          facilities required by the RSERVER module are a lot simpler
          than the support provided by RT-Linux, this interface looses
          the dependency to the particular RT-Linux version.

    In general, all these functions return 0 if everything went OK and
    some negative number if something went wrong.

*/


#ifndef __RTL_CORE_H
#define __RTL_CORE_H

#include "../../time/include/time.h"
#include "rtl_core/rtl_hal.h"




/* 
   (1) Context-related definitions, variables and functions.

   This level offers a context abstraction to upper levels.
   Each rt-task must have one context. Context are created
   and destroyed by calling the functions below.

   The Linux context *must* be created by calling
   rt_create_linux_context at the scheduler's initilization stage.
   This is mandatory for any scheduler.

*/

// Visible types

// NOTE: "rt_context_t" is defined in rtl_core/rtl_hal.h



// Visible functions

/* 
   (a) Context manipulation:
 */

 int rtl_context_create (void (*fn)(int), int data, 
			       int stack_size, rt_context_t **task_context);
 int rtl_context_destroy (rt_context_t *context);
 int rtl_context_use_fp (rt_context_t *context, int flag);
 int rtl_context_create_linux (rt_context_t **linux_context);

/* 
   (b) Low-level (core) layer of system calls / interrupt handlers:

   In upper levels, each system call or interrupt handler must:
      1) Call rt_update_system
      2) Serve the call/handle the interrupt
      3) Call rt_dispatch or rt_resume
*/

void rtl_update_system (void);
void rtl_dispatch (rt_context_t *new_context, systime_t next_interrupt);
void rtl_resume (void);
void rtl_program_timer (systime_t next_interrupt);


// Visible global constants
/* 
   ADJUSTUP is the value by which interrupts are not programmed
   too soon.

   It is used to release tasks and to program the next interrupt
   to a future time value which ensures that it is *in the future*

*/
#define ADJUSTUP  (systime_t) 2000



// Visible global variables
/* 
   If the three-step mechanism of servicing syscalls and interrupts
   described above is met, this layer automatically updates three
   global variables which can be then used from the scheduler:
      now              = the time at the beginning of the current call.
      last_update      = the "now" time at the last call.
      last_kernel_time = the time at the end of the last call.
*/
extern systime_t now;
extern systime_t last_update;
extern systime_t last_kernel_time;




/* 
   (2) Timer-related definitions, variables and functions.

   This level offers simplified timer API to upper levels, including
   basic functions to configure the hardware timer, install a timer
   handler and set a timer to interrupt at a particular time.

*/

// Visible definitions and macros

#define RT_TIME_END    HRTIME_INFINITY



// Visible functions

int rtl_set_timer (systime_t time);

int rtl_no_timer (void);

int rtl_request_timer(clock_irq_handler_t handler);

int rtl_free_timer(void);

int rtl_set_oneshot_mode (void);

systime_t rtl_set_periodic_mode (systime_t period);


#endif // __RTL_CORE_H
