/*  

   FILE:

	rtl_core.c						       
	
   AUTHOR: 
        Copyright (C) 1999-2001
	GTI-IA, UPV

        This is free software; you can redistribute it and/or modify
	it under terms of the GNU General Public License as published
	by the Free Soft- ware Foundation; either version 2, or (at
	your option) any later ver- sion.  It is distributed in the
	hope that it will be useful, but WITH- OUT ANY WARRANTY;
	without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
	License for more details.  You should have received a copy of
	the GNU General Public License; see file COPYING.  If not,
	write to the Free Software Foundation, 59 Temple Place - Suite
	330, Boston, MA 02111-1307, USA.


   DESCRIPTION:

        Implementation of the system core layer for the rtl scheduler
	(this layer include all functions independent from both the
	concrete task model and scheduling policy used). In particular,
	it provides the abstraction named "task context".

   CREATION DATE:
        26-Aug-1999

   REVISED:

	14-Nov-2003  Incorporated to ARTIS v2 source tree

        12-Oct-2003  Revised compatibility with RTL 3.2-pre1

        8-Apr-2001   Added compatibility to RT-Linux V3.


*/

#include <asm/errno.h>
#include <linux/slab.h>

#include "../../common/include/artis_conf.h"

#include "../../time/include/time.h"
#include "../include/rtl_core/rtl_hal.h"
#include "../include/rtl_core.h"

#include "../../trace/include/trace.h"

// Uncomment this for printing debug information:
//#define _RTL_CORE_DEBUG



/* 
   Public time-related variables 
   (visible for all the ARTIS modules)
*/
systime_t last_kernel_time = (systime_t) 0;
systime_t now              = (systime_t) 0;
systime_t last_update      = (systime_t) 0;


/* 
   Public context-related variable 
   (only visible within the rtl_core module) 
*/
rt_context_t rtl_linux_context;	


/* Private context-related variables */
static rt_context_t *rtl_current_context;
static rt_context_t *rtl_context_fpu_owner;








extern int rtl_startup (void (*fn) (int), int data);



int 
rtl_context_create (void (*fn) (int), int data, int stack_size,
		    rt_context_t ** task_context) 
{
  int *st;
  rt_context_t *new_context;

#ifdef _RTL_CORE_DEBUG
  rtl_printf("    (rtl_context_create) func %p, data %d, stacksize %d\n",
	     fn, data, stack_size);
#endif

  new_context = (rt_context_t *) 
                kmalloc (sizeof (rt_context_t), GFP_KERNEL);

  if (!new_context) {
    return -ENOMEM;
  }

  st = (int *) kmalloc (stack_size, GFP_KERNEL);
  if (!st) {
    kfree (new_context);
    return -ENOMEM;
  }

  *task_context = new_context;
  rtl_save_interrupts (new_context->interrupt_state);
  new_context->stack_bottom = st;
  new_context->stack = st + stack_size / sizeof (int);
  new_context->uses_fp = 0;
  rtl_init_stack (new_context, fn, data, rtl_startup);

  return 0;
}






int 
rtl_context_create_linux (rt_context_t ** linux_context) 
{
  /*
     In upper levels, a context is a pointer to a rt_context_t
     structure, properly initialized by calling rtl_context_create.
     In the case of the "Linux task", the context structure is a
     static variable in this file, called "rtl_linux_context", a
     pointer to which returns this function.
  */
  rtl_linux_context.uses_fp = 1;
  rtl_save_interrupts (rtl_linux_context.interrupt_state);
  *linux_context = &rtl_linux_context;
  rtl_context_fpu_owner = &rtl_linux_context;
  rtl_current_context = &rtl_linux_context;
  return 0;
}






int 
rtl_context_use_fp (rt_context_t * context, int flag) 
{
  context->uses_fp = flag;
  if (flag) {
    rtl_context_init_fpu (context, rtl_context_fpu_owner);
  }
  else {
    if (rtl_context_fpu_owner == context) {
      rtl_context_fpu_owner = 0;
    }
  }

  return 0;
}






int 
rtl_context_destroy (rt_context_t * context)
{
#ifdef _RTL_CORE_DEBUG
  rtl_printf("      (rtl_context_destroy) stack %#x, context: %#x\n",
	     context->stack_bottom, context);
#endif
  if (rtl_context_fpu_owner == context) {
    rtl_context_fpu_owner = 0;
  }
  if (context->stack_bottom) {
    kfree(context->stack_bottom);
    kfree(context);
  } else {
    return -EINVAL;
  }

  return 0;
}






void 
rtl_update_system (void)
{
  rtl_no_interrupts (rtl_current_context->interrupt_state);
  last_update = now;
  now = systime_gettime ();
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_CORE_DEBUG
  trace_event(OVERHD_START, 0, now, 0);
#endif
#endif
}





inline void 
rtl_program_timer (systime_t next_interrupt) 
{
  systime_t     time_now;

  time_now = systime_gettime();
  if (next_interrupt < time_now + ADJUSTUP) {
    next_interrupt = time_now + ADJUSTUP;
  }

  if (next_interrupt == RT_TIME_END) {
    rtl_set_timer ((HRTICKS_PER_SEC / HZ) / 2);
  } else {
    rtl_set_timer (next_interrupt - time_now);
  }
  
}





void 
rtl_dispatch (rt_context_t * new_context, systime_t next_interrupt)
{

  rtl_program_timer (next_interrupt);
  last_kernel_time = systime_gettime ();

#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_CORE_DEBUG
 trace_event(OVERHD_STOP, 0, last_kernel_time, 0);
#endif
#endif

  if (new_context == rtl_current_context) {
#ifdef ARTIS_CONF_CORE_DEBUG
    rtl_printf("(RTL_CORE) [rtl_dispatch] Same context. RETURNING\n");
#endif
    rtl_restore_interrupts (rtl_current_context->interrupt_state);
    return;
  }
  else {
    if (new_context == &rtl_linux_context) {
      rtl_make_rt_system_idle ();
    }
    else {
      rtl_make_rt_system_active ();
    }
#ifdef ARTIS_CONF_CORE_DEBUG
    rtl_printf("(RTL_CORE) [rtl_dispatch] JUST before switching\n");
#endif
    rtl_switch_to (&rtl_current_context, new_context);

    /* 
       Delay switching the FPU context until it is really needed
       this trick works for uniprocessors and SMP with no RT-task 
       migration (?) 
     */

    if (rtl_current_context->uses_fp
	&& rtl_context_fpu_owner != rtl_current_context) {
      if (rtl_context_fpu_owner) {
	rtl_fpu_save (rtl_context_fpu_owner);
      }
      rtl_fpu_restore (rtl_current_context);
      rtl_context_fpu_owner = rtl_current_context;
    }
    rtl_restore_interrupts (rtl_current_context->interrupt_state);
  }
}








void 
rtl_resume (void)
{
  last_kernel_time = systime_gettime ();
#ifdef ARTIS_TRACE_EVENTS
#ifdef ARTIS_CONF_CORE_DEBUG
 trace_event(OVERHD_STOP, 0, last_kernel_time, 0);
#endif
#endif
  rtl_restore_interrupts (rtl_current_context->interrupt_state);
  return;
}
