/*

  STORAGE.H

  Copyright (C) 1998-2004
  Ana Garcia Fornes
  Andres Terrasa Barrena
  Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write
  to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston,
  MA 02111-1307, USA.                                                      
*/


/*
  DESCRIPTION

  A storage is a memory region in which you can allocate three types
  of entities: objects, pools and items.

  An object is a chunk of memory of a given size. It can not be
  deallocated.

  A pool is a set of items. The maximum number of items in a pool must
  be specified. All the items in a pool are of the same size, and this
  size is specified when the pool is created. Pools can no be
  deallocated. Items can be allocated and deallocated within a pool.

  Storages has been designed to be shared between different virtual
  address spaces. In this way, all the pointers defined in this
  interface are addresses relative to the beginning of the storage,
  and they must be converted to virtual addresses (function
  Storage_Get_Address) in order to use them in your program.

  Objects and pools are assigned an entity number when they are
  created. This entity number is useful to identify the object or
  pool from a virtual address space different than the one which
  created the object or the pool.

  ERROR REPORTING

  NULL or STORAGE_FAILURE when the function fails 
  Not NULL or STORAGE_SUCCESS in other case */


#ifndef _STORAGE
#define _STORAGE

#define ARTIS // Comment this line if you are not using ARTIS


// Visible macros
 
#define STORAGE_SUCCESS 1
#define STORAGE_FAILURE 0


// Visible types

typedef void          *Storage_T;
typedef unsigned long  Storage_Entity_Size_T;
typedef void          *Storage_Pool_Handler_T;  



// Visible functions



//////////////////////////////////////////
// Artis Storage Management
//////////////////////////////////////////

#ifdef ARTIS

int
Storage_Create (char           *Name,
		unsigned long   Storage_Size,
		unsigned long   Max_Entities,
		Storage_T      *Stoid
		);


int 
Storage_Destroy (char      *Name,
		 Storage_T  Stoid) ;

#endif //ARTIS



#ifndef ARTIS

int
_Storage_Create (void           *Initial_Address,
		 unsigned long   Storage_Size,
		 unsigned long   Max_Entities,
		 Storage_T      *Stoid
		 );

#endif // no ARTIS




//////////////////////////////////////////
// Objects Management
//////////////////////////////////////////

void *
Storage_Create_Object (Storage_T             Stoid,
		       Storage_Entity_Size_T Size_Of_Object,
		       unsigned long         Entity_Id
		       );

void *
Storage_Get_Object (Storage_T     Stoid,
		    unsigned long Entity_Id);




//////////////////////////////////////////
// Pools Management
//////////////////////////////////////////

Storage_Pool_Handler_T
Storage_Create_Pool (Storage_T             Stoid,
		     Storage_Entity_Size_T Item_Size, 
		     unsigned int          How_Many_Items,
		     unsigned long         Entity_Id
		     );

Storage_Pool_Handler_T
Storage_Get_Pool (Storage_T     Stoid,
		  unsigned long Entity_Id
		  );

int
Storage_Delete_Pool (Storage_T              Stoid,
		     Storage_Pool_Handler_T In_Pool
		     );


// Dangerous pool functions

int
Storage_Destroy_Pool (Storage_T              Stoid,
		      Storage_Pool_Handler_T The_Pool
		      );




//////////////////////////////////////////
// Items Management
//////////////////////////////////////////

void *
Storage_Create_Item (Storage_T              Stoid,
		     Storage_Pool_Handler_T In_Pool
		     );


int
Storage_Destroy_Item (Storage_T               Stoid,
		      Storage_Pool_Handler_T  In_Pool,
		      void                    *One_Item
		      );

// Dangerous item functions

void *
Storage_Get_Item (Storage_T              Stoid,
		  Storage_Pool_Handler_T In_Pool, 
		  unsigned int           iItem_Position
		  );




//////////////////////////////////////////
// Miscellanea
//////////////////////////////////////////

int Storage_Check(void *address); // AGUS Visible?

inline void * 
Storage_Get_Address (Storage_T  Stoid,
		     void      *internal_addr
		     );

inline void * 
Storage_Get_Relative (Storage_T  Stoid,
		     void      *external_addr
		     );



#ifdef MODULE
int storage_init_module (void);
void storage_cleanup_module(void);
#endif
 

#endif /* _STORAGE */


