/*!
  \file kdmparser.c
  \date Time-stamp: <30-03-2005 08:35>
  \author M. Rebollo
  \version 0.1

  \brief kdm specific parsing functions
*/


  
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include "globals.h"
#include "kdmparser.h"
#include "symtab.h"

#define HEADERFILE "userclass.h"
#define CODEFILE   "userclass.c"
#define UCLASS1  "ucav2.1"
#define UCLASS2  "ucav2.2"
#define UCLASS3  "ucav2.3"
#define UCLASS4  "ucav2.4"
#define UCLASS5  "ucav2.5"
#define UCLASS6  "ucav2.6"
#define UCLASS7  "ucav2.7"

#define _build_char( ch ) 'ch'
#define _build_string( str ) #str

/*!
  \var char skdmtype[7][12]
  \brief KDM internal data types
*/
char skdmtype[7][12]={ "KDM_ERROR", "KDM_VOID", "KDM_INT", "KDM_FLOAT", 
			  "KDM_CHAR", "KDM_ARRAY", "KDM_HANDLER" };

/*!
  \var char sevtype[7][12]
  \brief Event types
*/
char sevtype[2][12]={ "NO_EVENT", "MODIFICATION" };

//Global variables definition
FILE *fheader, *fcode;                  //!< output header and code files
FILE *fdvar, *fdclass, *fdobj, *fdslot; //!< temporal output files
symbol_table *ToS;                      //!< table of symbols for the XML compiler
static int entities = 0;                //!< total number of user entities


/*!
  \fn int is_atomic( st_item* theItem )
  \brief 1 if \em theItem is an atomic data type. 0 otherwise
  \param theItem variable whose type is going to be checked
*/
inline static int
//----------------------------------------------
is_atomic( st_item* theItem )
//----------------------------------------------
{
  return ( theItem->type == tvoid || theItem->type == tint || 
           theItem->type == tfloat || theItem->type == tchar );
} //is_atomic()


/*!
  \fn int is_string( st_item* theItem )
  \brief 1 if \em theItem is a string. 0 otherwise
  \param theItem variable whose type is going to be checked
*/
inline static int
//----------------------------------------------
is_string( st_item* theItem )
//----------------------------------------------
{
  if( theItem->type == tarray && ((at_item*)theItem->ref)->type == tchar )
    return 1;
  else return 0;
} //is_string()


/*!
  \fn int is_array( st_item* theItem )
  \brief 1 if \em theItem is an array. 0 otherwise
  \param theItem variable whose type is going to be checked
*/
inline static int
//----------------------------------------------
is_array( st_item* theItem )
//----------------------------------------------
{
  return ( theItem->type == tarray );
} //is_array()


static void
//----------------------------------------------
insert_into_table( const char *itemname, 
		   int itemcat, 
		   int itemtype,
		   int itemsize)
//----------------------------------------------
{
  //st_item theItem;                    
  
  //make_empty_item( &theItem );                                  
  st_add_symbol( itemname, itemcat, itemtype, ToS );                
} //insert_into_table()


/*!
  \fn void copy_from_template(const char *source_filename, FILE *target )
  \brief Copy he file passed as parameter to the final code file
  \param source_filename temporal file to extract 
  \param target user final code file generated from XML class definition.
*/
static void 
//-----------------------------------------------
copy_from_template( const char *source_filename, 
		    FILE *target )
//-----------------------------------------------
{
  FILE *source;
  char ch;

  if( (source = fopen( source_filename, "r" )) != NULL ){
    while( (fscanf( source, "%c", &ch )) != EOF )
      fprintf( target, "%c", ch );
    fclose( source);
  }
  else{
    fprintf( stderr, "%s file does not exist\n", source_filename );
    exit( -1 );
  } //else (file can't be opened)
} //copy_from_template()



/*!
  \fn void merge_code_files
  \brief Merge all temporal files into the userclass.c final file
*/
static void
//-----------------------------------------------
merge_code_files( void )
//-----------------------------------------------
{
  int i;
  char filename[10];

  for( i = 1; i <= 7; i++ ){
    sprintf( filename, "ucav2.%d", i );
    copy_from_template( filename, fcode );
  } //for each template file
} //merge_code_files()



/*! 
  \fn void hfileheading( const char *xmlfile )
  \brief begins the header user file
  \param xmlfile the xml file to parse
  \exception userclass.h header file does not exist
*/
static void 
//-----------------------------------------------
hfileheading( const char *xmlfile )
//-----------------------------------------------
{
  time_t today;

  assert( fheader != NULL );

  //gets the current data and generates header for the file
  today = time( NULL );
  fprintf( fheader, "/*\n" );
  fprintf( fheader, "  file: userclass.h\n" );
  fprintf( fheader, "  generated from %s\n", xmlfile );
  fprintf( fheader, "  date: %s\n", ctime( &today ) );

  fprintf( fheader, "  PLEASE, DO NOT MODIFIY THIS FILE\n" );
  fprintf( fheader, "*/\n\n" );

  fprintf( fheader, "#ifndef _USERCLASS_H_\n" );
  fprintf( fheader, "#define _USERCLASS_H_\n\n" );

  fprintf( fheader, "#include <artis/kdm/include/kdm.h>\n\n" );


} //hfileheading()

static void
//-----------------------------------------------
hfileending( void )
//-----------------------------------------------
{

  // adds the declaration of public functions and ends the file
  fprintf( fheader, "\nextern int user_kdm_load( void );\n" );
  fprintf( fheader, "extern int user_kdm_unload( void );\n" );
  fprintf( fheader, "\n#endif //_USERCLASS_H_\n" );

} //hfileending()


static void 
//-----------------------------------------------
cfileheading( const char *xmlfile )
//-----------------------------------------------
{
  time_t today;
  FILE *templfd;
  char ch;
  

  assert( fcode != NULL );

  today = time( NULL );
  fprintf( fcode, "/*\n" );
  fprintf( fcode, "  file: userclass.c\n" );
  fprintf( fcode, "  generated from %s\n", xmlfile );
  fprintf( fcode, "  date: %s\n", ctime( &today ) );

  fprintf( fcode, "  PLEASE, DO NOT MODIFIY THIS FILE\n" );
  fprintf( fcode, "*/\n\n" );

} //hfileheading()


static void
//-----------------------------------------------
cfileending( void )
//-----------------------------------------------
{
  int i, classnum = 0;

} //cfileending()


void 
//-----------------------------------------------
kdm_startclass( const char *classname )
//-----------------------------------------------
{
  assert( fheader != NULL );

  st_add_symbol( classname, tidentifier, tclass, ToS );
  fprintf( fheader, "typedef struct{\n" );
  
} //kdm_startclass()


void 
//-----------------------------------------------
kdm_endclass( const char *classname )
//-----------------------------------------------
{
  assert( fheader != NULL );

  fprintf( fheader, "} %s;\n\n", classname );

} //kdm_endclass()


void 
//-----------------------------------------------
kdm_addslot( const char *classname,
	     const char *slotname, 
	     const char *slottype )
//-----------------------------------------------
{
  int resp;

  assert( fheader != NULL );

  if( strcmp( slottype, "int" ) == 0 )
    st_add_symbol( slotname, tidentifier, tint, ToS );
  else if( strcmp( slottype, "char" ) == 0 )
    st_add_symbol( slotname, tidentifier, tchar, ToS );
  else if(  strcmp( slottype, "float" ) == 0 )
    st_add_symbol( slotname, tidentifier, tfloat, ToS );
    
  fprintf( fheader, "  static_slot %s;\n", slotname );

} //kdm_addslot()


void 
//-----------------------------------------------
kdm_addslotvalue( const char *instname,
		  const char *slotname,
		  const char *slotvalue,
		  const char *eventtype,
		  const char *eventimp )
//-----------------------------------------------
{
  int slottype;
  st_item *theItem;
  int i;

  theItem = st_search( slotname, ToS );

  assert( theItem != NULL );

  //sets the datatype for the slot
  slottype = (int)(theItem->type);
  fprintf( fdslot, "  %s->%s.type = %s;\n ", 
	   instname, slotname, skdmtype[slottype+1] );

  //sets the defalut value for the slot
  if( is_atomic( theItem ) ){
    if( strlen( slotvalue ) > 0 ){
      if( theItem->type == tchar )
	fprintf( fdslot, "  %s->%s.value.c = '%s';\n ", 
		 instname, slotname, slotvalue );
      else if( theItem->type == tint )
	fprintf( fdslot, "  %s->%s.value.i = %s;\n ", 
		 instname, slotname, slotvalue );
      else if( theItem->type == tfloat)
	fprintf( fdslot, "  %s->%s.value.f = %s;\n ", 
		 instname, slotname, slotvalue );
      else 
	fprintf( fdslot, "  %s->%s.value = %s;\n ", 
		 instname, slotname, slotvalue );
    } //if it has default value
  } //if it is an atomic data type

  else{
    assert( theItem->type == tarray );
    assert( theItem->ref != NULL );
   
    //creates an aditional object for the array

    entities++;  //increments the number of storage's entities

    fprintf( fdobj, "  _create_array( %s->%s.value.pt, %s, %d );\n",
	     instname, slotname,
	     sdatatype[((at_item*)theItem->ref)->type+1], 
	     ((at_item*)theItem->ref)->size );
    
    if( strlen( slotvalue ) > 0 ){
      if( is_string( theItem ) ){
	fprintf( fdslot, "  strcpy( Storage_Get_Address(tbn.stdata, %s->%s.value.pt), \"%s\" );\n ", 
		 instname, slotname, slotvalue );
      } //if it is an string
      else if( is_array( theItem ) ){
	fprintf( fdslot, "  %s->%s.value.pt[0] = %s;\n ", 
		 instname, slotname, strtok( slotvalue, " " ) );
	for( i = 1; i < ((at_item*)theItem->ref)->size; i++ ){
	  fprintf( fdslot, "  %s->%s.value.pt[%d] = %s;\n ", 
		   instname, slotname, i, strtok( NULL, " " ) );
	} //for each default value for the array
      } //if it is an array
    } //if it has default value
  } //else (it is a complex type -array-)
  
  //sets the initial set of events
  fprintf( fdslot, "  %s->%s.event = %s;\n", 
	   instname, slotname, eventtype );
  fprintf( fdslot, "  %s->%s.imp = %s;\n\n ", 
	   instname, slotname, eventimp );  
  
} //kdm_addslotvalue()


void 
//-----------------------------------------------
kdm_addarray( const char *classname,
	      const char *slotname, 
	      const char *slottype,
	      const char *slotsubtype,
	      const char *slotsize )
//-----------------------------------------------
{
  int resp;

  assert( fheader != NULL );

  if ( strcmp( slottype, "string" ) == 0 )
    st_add_array( slotname, tidentifier, tarray, tchar, atoi( slotsize ), ToS );
  else if( strcmp( slottype, "array" ) == 0 ){
    if( strcmp( slotsubtype, "int" ) == 0 )
      st_add_array( slotname, tidentifier, tarray, tint, atoi( slotsize ), ToS );
    else if( strcmp( slotsubtype, "char" ) == 0 )
      st_add_array( slotname, tidentifier, tarray, tchar, atoi( slotsize ), ToS );
    else if(  strcmp( slotsubtype, "float" ) == 0 )
      st_add_array( slotname, tidentifier, tarray, tfloat, atoi( slotsize ), ToS );
  } //if generic array variable
    
  fprintf( fheader, "  static_slot %s;\n", slotname );

} //kdm_addarray()


void 
//-----------------------------------------------
kdm_addinstance( const char *classname, 
		 const char *instname )
//-----------------------------------------------
{
  assert( fheader != NULL );

  st_add_symbol( instname, tidentifier, tvoid, ToS );

  fprintf( fheader, "extern %s *%s;\n", classname, instname );

  fprintf( fdvar, "%s *%s;\n", classname, instname );
  fprintf( fdclass, "  _create_object( %s, %s );\n", classname, instname );
  entities++;  //increments the number of storage's entities
  

} //kdm_addinstance()


int
//-----------------------------------------------
kdm_beginparsing( const char *xmlfile )
//-----------------------------------------------
{

  // create header file
  if( (fheader = fopen( HEADERFILE, "w" )) == NULL ){
    fprintf( stderr, "Error creating class file\n" );
    exit( -1 );
  }
  hfileheading( xmlfile );

  // create files for the user objects
  if( (fdvar = fopen( UCLASS2, "w" )) == NULL ||
      (fdclass = fopen( UCLASS4, "w" )) == NULL ||
      (fdobj = fopen( UCLASS5, "w" )) == NULL ||
      (fdslot = fopen( UCLASS6, "w" )) == NULL 
      ){
    fprintf( stderr, "Error creating class files\n" );
    exit( -1 );
  }
  // create the hash table for the identifiers
  ToS = st_newtable();
} //kdm_beginparsing()


void
//-----------------------------------------------
kdm_endparsing( const char *xmlfile )
//-----------------------------------------------
{
  //Sets the final number of user entities in the storage
  fprintf( fdvar, "\n#define NUM_USER_ENTITIES	%d\n\n", entities );
  fprintf( fdobj, "\n#ifndef MODULE\n" );

  //finishes and closes header files
  hfileending();
  fclose( fheader );

  //closes temporal code files
  fclose( fdvar );
  fclose( fdclass);
  fclose( fdobj );
  fclose( fdslot );

  // create code file by merging generated and static files
  if( (fcode = fopen( CODEFILE, "w" )) == NULL ){
    fprintf( stderr, "Error creating class files\n" );
    exit( -1 );
  }
  cfileheading( xmlfile );
  merge_code_files();
  cfileending();
  fclose( fcode );
 
  // deletes the hash table and frees the memory
  if( verbose_mode ) st_printtable( ToS );
  st_deltable( ToS );

} //kdm_endparsing()
