/*!
  \file parser.c
  \date Time-stamp: <30-03-2005 09:10>
  \author M. Rebollo (mrebollo@dsic.upv.es)
  \version 0.3

  \brief Translate xml file with class definition into C structures
*/


#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <expat.h>
#include "globals.h"
#include "kdmparser.h"

/*! 
  \def MARK( f, msg )
  \brief prints the message 'msg', marking the function 'f'
*/ 
#define MARK(f, msg) if( verbose_mode ) printf( "(%s) %s(%d)\n", f, msg, userdata->depth )
//#define MARK(string) ;//empty


/*!
  \struct XML_State
  \brief  structure to pass information between handlers without using globals
*/
typedef struct userdata{
  XML_Parser parser;        //!< pointer to the current handler
  int wrong_version;        //!< true if incorrect version of XML detected
  int depth;                //!< current label depth
  bool is_a_class;          //!< true if a class is been analised
  bool is_an_slot;          //!< true if an slot is been analised
  char cname[MAXBUFF];      //!< name of the current class
  char sname[MAXBUFF];      //!< name of the current slot
  char stype[MAXBUFF];      //!< type of the current slot
  char slength[MAXBUFF];    //!< number of elements for array-typed slots
  char elemtype[MAXBUFF];   //!< type of subelement for arrays or structs
  char iname[MAXBUFF];      //!< name of the current instance
  char svalue[MAXBUFF];     //!< current value of the slot
  char sevent[MAXBUFF];     //!< asociated initial event for the slot
  char eventimp[2];         //!< importance associated to the event
}XML_State;

/*!
  \enum deeplevel
  \brief deepness in the XML tree depending on the type of the element
*/
enum deeplevel{ TBNLVL, CLASSLVL, SLOTLVL, INITLVL, EVENTLVL, CONSTLVL };

void analyse( const char *filename );



/////////////////////////////////////////////////////////////
// Utility functions
/////////////////////////////////////////////////////////////


/*!
  \fn 
  \param
  \return
  \brief
*/
static bool
//------------------------------------------------------------
is_a_class( const char  **attr )
//------------------------------------------------------------
{
  int i;
  
  for( i = 0; attr[i]; i += 2 ) 
    // An static_slot has a required attribute called kdm:data
    if( strcmp( attr[i], "name" ) == 0 && 
	strcmp( attr[i+1], "default" ) == 0 )
      return false;
    else if( strcmp( attr[i], "kdm:data" ) == 0 )
      return false;
  
  return true;

} //is_a_class()


/*!
  \fn 
  \param
  \return
  \brief
*/
static bool
//------------------------------------------------------------
is_an_slot( const char  **attr )
//------------------------------------------------------------
{
  int i;
  
  for( i = 0; attr[i]; i += 2 ) 
    // An static_slot has a required attribute called kdm:data
    if( strcmp( attr[i], "kdm:data" ) == 0 )
      return true;
  
  return false;

} //is_an_slot()

/*!
  \fn 
  \param
  \return
  \brief
*/
static void
//------------------------------------------------------------
show_data( void* userData )
//------------------------------------------------------------
{
  XML_State* userdata = userData;
  printf_verbose( "userdata.version = %d\n", userdata->wrong_version );
  printf_verbose( "userdata.depth = %d\n", userdata->depth );
  printf_verbose( "userdata.is_a_class = %d\n", userdata->is_a_class );
  printf_verbose( "userdata.cname = %s\n", userdata->cname );
  printf_verbose( "userdata.sname = %s\n", userdata->sname );
  printf_verbose( "userdata.stype = %s\n", userdata->stype );
  printf_verbose( "userdata.elemtype = %s\n", userdata->elemtype );
  printf_verbose( "userdata.iname = %s\n", userdata->iname );
  printf_verbose( "userdata.svalue = %s\n", userdata->svalue );
} //show_data()


/*!
  \fn static void create_temporal_files( void )
  \brief Prepares the parser for its execution.
Copy from the templates directory the temporal files with fixed information.
*/
static void
//------------------------------------------------------------
create_temporal_files( void )
//------------------------------------------------------------
{
  printf_verbose( "Creating temporal files..." );
  system( "cp /usr/local/artis/tools/templates/ucav* ." );
  printf_verbose( "done\n" );
} //create_temporal_files()


/*!
  \fn static void remove_temporal_files( void )
  \brief Restores the system to its previous state. 
Deletes temporal files used for class generation
*/
static void
//------------------------------------------------------------
remove_temporal_files( void )
//------------------------------------------------------------
{
  printf_verbose( "Revoming temporal files..." );
  system( "rm -f ucav*" );
  printf_verbose( "done\n" );
} //remove_temporal_files()




/////////////////////////////////////////////////////////////
// Parsing handlers
/////////////////////////////////////////////////////////////


/*!
  \fn static void xmldecl_handler()
  \brief Handles the XML version checking process
*/
static void
//------------------------------------------------------------
xmldecl_handler(void            *userData,
                const XML_Char  *version,
                const XML_Char  *encoding,
                int              standalone)
//------------------------------------------------------------
{
  static const XML_Char Version_1_0[] = {'1', '.', '0', 0};

  int i, len;  
  XML_State* userdata = userData;

  len = strlen( Version_1_0 );
  for (i = 0; i < len; i++) {
    if (version[i] != Version_1_0[i]) {
      userdata->wrong_version = 1;
      //clear all handlers
      XML_SetCharacterDataHandler(userdata->parser, NULL);
      XML_SetElementHandler( userdata->parser, NULL, NULL );
      XML_SetCharacterDataHandler( userdata->parser, NULL );
      XML_SetCommentHandler( userdata->parser, NULL );
      XML_SetProcessingInstructionHandler( userdata->parser, NULL );
      fprintf( stderr, "Wrong version of XML document\n" ); 
      exit( -1 );
    } //if wrong version detected
  }//for i
} //xmldecl_handler()


/*!
  \fn void xmlstartelem_handler()
  \brief Called when a new label is opened in the XML file. 
  It copies to the userdata structure the information about the current element
*/
static void
//------------------------------------------------------------
xmlstartelem_handler( void        *userData,
		      const char  *el, 
		      const char  **attr )
//------------------------------------------------------------
{
  int i;
  XML_State* userdata = userData;
  int schema_file = 0;

  MARK( "xmlstartelem_handler", el );

  // The first level only contains a <tbn> element
  if( strcmp( el, "tbn" ) == 0 ){
    (userdata->depth) = TBNLVL;
    for (i = 0; attr[i]; i += 2) {
      if( strcmp( attr[i], "xsi:noNameSpaceSchemaLocation" ) == 0 )
	schema_file = i + 1;
    } //for i
  } //if tbn

  
  // Into the <tbn> element only classes can be defined
  else if( userdata->depth == TBNLVL ){
    //load class and instance information
    strcpy( userdata->cname, el );
    for( i = 0; attr[i]; i += 2 )
      if( strcmp( attr[i], "id" ) == 0 )
	strcpy( userdata->iname, attr[i+1] );
    (userdata->depth)++;
  } //else (reading class information)
  
  
  // Each class has the definition of its own slots
  else if( userdata->depth == CLASSLVL ){
    strcpy( userdata->sname, el );
    printf_verbose( "(%s, %s, %s)\n", userdata->cname, userdata->iname, userdata->sname );
    (userdata->depth)++;
    printf_verbose( "level %d\n", userdata->depth );
  } //else (reading slot information)

 
  // For each slot, a default value and event information can be initialised
  else if( userdata->depth == SLOTLVL ){
    //analyses a default value
    if( strcmp( el, "default" ) == 0){
      printf_verbose( "default value: " );
    } //if the element is a default value

    else if( strcmp( el, "event" ) == 0){
      printf_verbose( "Event definition for %s: ", userdata->sname );
    } //if the element is an event

    else{
      fprintf( stderr, "** Unexpected element %s **\n", el );
      exit( -1 );
    } //else (unrecognised element)
    (userdata->depth)++;
  } //else (reading initial values for the slot


  // The event is defined using its type and its importance
  else if( userdata->depth == INITLVL ){
    //analyses the information of the events
    if( strcmp( el, "type" ) == 0 ){
      printf_verbose( "- event type: ", userdata->sname );
    } //if reading event type
    
    else if( strcmp( el, "importance" ) == 0 ){
      printf_verbose( "- event importance: ");
    } //if reading event importance

    else{
      fprintf( stderr, "** Unexpected element %s **\n", el );
      exit( -1 );
    } //else (unrecognised element)
    (userdata->depth)++;
  } //else

  // If an schema is detected, we must analise it before proceed 
  // with the rest of the file, because the schema contains the
  // information about the classes and the slots.
  //
  // The XML file only contains the instances and the initial values
  if( schema_file > 0 ){
    printf_verbose( "**\nSchema detected...\nAnalysing schema structure...\n**\n" );
    analyse( attr[schema_file] );
    printf_verbose( "**\nEnd of schema.\nContinuing with code file...\n**\n" );
  } //if schema detected
  
} //xmlstartelem_handler()

/*!
  \fn void xmlendelem_handler()
  \brief Called when and ending label is read in he XML file. 
  It creates the instances and slots in the user's class files.
*/
static void
//------------------------------------------------------------
xmlendelem_handler( void        *userData, 
		    const char  *el )
//------------------------------------------------------------
{
  XML_State* userdata = userData;

  MARK( "xmlendelem_handler", el );

  if( (userdata->depth) == CLASSLVL ){
    kdm_addinstance( userdata->cname, userdata->iname );
    (userdata->depth)--;
  } //if class found
  

  else if( (userdata->depth) == SLOTLVL ){
    printf_verbose( "Level %d\n", userdata->depth );
    printf_verbose( "I'm going to insert the slot %s.%s = %s...",
	    userdata->iname, userdata->sname, userdata->svalue );
    kdm_addslotvalue( userdata->iname, userdata->sname, userdata->svalue, userdata->sevent, userdata->eventimp );
    printf_verbose( "done\n" );
    //deletes de value after using it to allow empty default values
    strcpy( userdata->svalue, "\0" );
    (userdata->depth)--;
  } //if slot found

  
  else if( userdata->depth == INITLVL){
    printf_verbose( "... slot data readed\n" );
    (userdata->depth)--;    
  } // if

  
  else if( userdata->depth == EVENTLVL){
    printf_verbose( "... event readed\n" );
     (userdata->depth)--;
  } // if
} //xmlendelem_handler() 


/*!
  \fn void xmltext_handler( void *userData,const char *el,int len )
  \brief This handler is called when free text is read is the XML file.
*/
static void
//------------------------------------------------------------
xmltext_handler( void       *userData, 
		 const char *el, 
		 int        len )
//------------------------------------------------------------
{
  int i, j;
  XML_State* userdata = userData;
  char *target;

  for( i = 0; i < len; i++ ){
    // delete white spaces
    if( el[i] == ' ' || el[i] == '\t' || el[i] == '\n' )
      ;//empty
    // print free text
    else{
      
      MARK( "xmltext_handler", el );

      
 
      //choose the label to assign the value
      switch( userdata->depth ){
      case INITLVL: 
	target = userdata->svalue; 
	break;
      case EVENTLVL:
	if( isalpha( el[i] ) )
	  target = userdata->sevent; 
	else
	  target = userdata->eventimp;
	break;
      } //switch deepness of analysis
     
      for( j = i; j < len; j++ ){
	printf_verbose( "%c", el[j] );
	target[j-i]=el[j];
      } //for j
      target[j] = '\0';
      printf_verbose( "readed value: %s\n", target );

      return;
    } //else
  } //for i
} //xmltext_handler()



/*!
  \fn void xsdstartelem_handler(void *userData, const char  *el, const char  **attr )
  \param userData structure to pass information between calls 
  \param el name of the strating elemen (open label)
  \param attr list of attributes and values defined inside the element. The odd items are the names of the attributes and the even items are their corresponding values
  \brief Called when a new opening label is read in the XSD file.
  This function (a) creates a new class; or (b) adds to the userclass
  structure the information about current slot.
*/
static void
//------------------------------------------------------------
xsdstartelem_handler( void        *userData,
		      const char  *el, 
		      const char  **attr )
//------------------------------------------------------------
{
  int i;
  XML_State* userdata = userData;

  MARK( "xsdstartelem_handler", el );

  if( strcmp( el, "xs:element" ) != 0 )
    return;
 
  // checks if the element is a class
  if( is_a_class( attr ) ){
    printf_verbose( "is a class = %d\n", userdata->is_a_class );
    userdata->is_a_class = true;
    
    //takes the class' name
    for( i = 0; attr[i]; i += 2 ) 
      if( strcmp( attr[i], "name" ) == 0 ){
	strcpy( userdata->cname, attr[i+1] );
	break;
      } //if name of the class detected

    kdm_startclass( userdata->cname );

    printf_verbose( "new class %s\n", userdata->cname );
  } //if the element is a class

  // checks if the element is an static slot
 else if( is_an_slot( attr ) ){
    printf_verbose( "static slot " );
    userdata->is_an_slot = true;

   //takes de characteristics of the slot and fills the
    // userdata structure with this information
    for( i = 0; attr[i]; i += 2 ) {
      printf_verbose( "%s=\"%s\" ", attr[i], attr[i+1] );
      if( strcmp( attr[i], "name" ) == 0 )
	strcpy( userdata->sname, attr[i+1] );
      else if( strcmp( attr[i], "kdm:data" ) == 0 )
	strcpy( userdata->stype, attr[i+1] );
      else if( strcmp( attr[i], "kdm:size" ) == 0 )
	strcpy( userdata->slength, attr[i+1] );
      else if( strcmp( attr[i], "kdm:item" ) == 0 )
	strcpy( userdata->elemtype, attr[i+1] );
    } //for each attribute (i)
    printf_verbose( "\n" );
  } //if the element is an slot
  
  (userdata->depth)++;
  
} //xsdstartelem_handler()


/*!
  \fn void xsdendelem_handler()
  \brief Called when a ending label is read in the XSD file.
  It inserts a new slot in the user's class files.
  
*/
static void
//------------------------------------------------------------
xsdendelem_handler( void        *userData, 
		    const char  *el )
//------------------------------------------------------------
{
  XML_State* userdata = userData;

  MARK( "xsdendelem_handler", el );

  if( strcmp( el, "xs:element" ) != 0 )
    return;

  (userdata->depth)--;
  
  if( userdata->depth == SLOTLVL ) return;

  show_data( userdata );
 
  if( userdata->is_an_slot ){
    //inserts a new slot
    if( strcmp( userdata->stype, "string" ) == 0 ||
	strcmp( userdata->stype, "array" ) == 0 )
      kdm_addarray( userdata->cname, userdata->sname, userdata->stype,
		    userdata->elemtype, userdata->slength );
    else
      kdm_addslot( userdata->cname, userdata->sname, userdata->stype );

    userdata->is_an_slot = !userdata->is_an_slot;
  }
  else{
    //inserts a new class
    kdm_endclass( userdata->cname );

    (userdata->is_a_class) = !(userdata->is_a_class);
  }
  

} //xsdendelem_handler() 




/*!
  \fn void analyse( const char *filename )
  \brief This funtion is responsible for analising files.
  It is in charge of register the handles for each file type.
*/
void 
//-----------------------------------------------------------
analyse( const char *filename )
//-----------------------------------------------------------
{
  FILE *fin;               //xml source file
  XML_Parser parser;       //parser for the elements
  XML_State userdata;      //global user data
  char buff[MAXBUFF];      //string with element to parse
  int len, done, line;

  // open xml file 
  if( (fin = fopen( filename, "r" )) == NULL ){
    fprintf( stderr, "Error opening file %s\n", filename );
    exit( -1 );
  } // if error opening file

  // create parser and analyse document
  parser = XML_ParserCreate( NULL );
  if( !parser ){
    fprintf(stderr, "Couldn't allocate memory for parser\n");
    exit( -1 );
  } // if not enough memory
  
  // define handlers for analysing xml code
  XML_SetUserData( parser, &userdata );
  XML_SetXmlDeclHandler( parser, xmldecl_handler );
  if( strstr( filename, ".xml" ) != NULL )
    XML_SetElementHandler( parser, xmlstartelem_handler, xmlendelem_handler );
  else if( strstr( filename, ".xsd" ) != NULL )
    XML_SetElementHandler( parser, xsdstartelem_handler, xsdendelem_handler );
  else{
    fprintf( stderr, "Error: invalid filetype\n" );
    XML_ParserFree( parser );
    exit( -1 );
  } //else (not an xml or xsd file)
  XML_SetCharacterDataHandler( parser, xmltext_handler );

  // analising xml file
  userdata.parser = parser;
  userdata.wrong_version = 0;
  userdata.depth = 0;
  userdata.is_a_class = 0;
  for( line = 1, done = 0; fgets( buff, MAXBUFF, fin ) != NULL; line++ ){
#ifdef DEBUG    
    printf_verbose( "line %d(%d): %s\n", line, strlen(buff), buff );
#endif    
    if( XML_Parse( parser, buff, strlen( buff ), done) == XML_STATUS_ERROR) {
	fprintf( stderr, "Parse error at line %d:\n%s\n",
              XML_GetCurrentLineNumber( parser ),
              XML_ErrorString( XML_GetErrorCode( parser )));
	exit( -1 );
    } //if error in xml file 
  } //for !EOF
  
  // free memory and close files
  XML_ParserFree( parser );
  fclose( fin );

} //analyse()


/////////////////////////////////////////////////////////////
// Main program
/////////////////////////////////////////////////////////////



int 
//------------------------------------------------------------
main( int  argc, 
      char *argv[] )
//------------------------------------------------------------
{
  int syntax_error = 0;
  char* filename;

  atexit( remove_temporal_files );
  verbose_mode = 0;
  
  printf( "KDM parser v 0.3 (Build 20050321)" );

  // check sintax
  if( argc < 2 )
    // name of file needed
    syntax_error = 1;
  else if( argc == 2 )
    // the argument must be the name of he file
    if( argv[1][0] == '-' )
     syntax_error = 1;
    else
      filename = argv[1];
  else if( argc == 3 )
    // only '-v' option is allowed
    if( strcmp( argv[1],"-v" ) != 0 )
      syntax_error = 1;
    else{
      verbose_mode = 1;
      filename = argv[2];     
    }
  if( syntax_error ){
    fprintf( stderr, "Usage: parser [-v] <file.xml>\n" );
    return -1;
  } //if bad sintax

  // analyse xml file
  create_temporal_files();
  kdm_beginparsing( filename );
  analyse( filename );
  kdm_endparsing(filename );
  remove_temporal_files();

  printf( "userclass.h\n" );
  printf( "userclass.c\n" );

  printf_verbose( "** XML file processed **\n" );

  return 0;

} //main()

