/*!
  \file symtab.h
  \date Time-stamp: <21-03-2005 12:34>
  \author M. Rebollo (mrebollo@dsic.upv.es)
  \version 0.1

  \brief Implements a symbol table for a dictionary 
           about XML and ARTIS expressions
*/


#ifndef _SYMTAB_H_
#define _SYMTAB_H_

/*!
  \def   NKEYS
  \brief Number of trays in the main array
*/
#define NKEYS   211

/*!
  \enum  datatype
  \brief Data types for user objects
*/
enum datatype { terror=-1, tvoid, tint, tfloat, tchar, tarray, tclass };



/*!
  \var   sdatatype
  \brief Data types recognised by the KDM
*/

extern char sdatatype[7][10];  //!< types recognised by the KDM
/*!
  \enum  category
  \brief It represents the lexical category of the object
*/
enum category { tidentifier, tfunction, tparam };


///////////////////////////////////////////////////////
//Definition of the items
//////////////////////////////////////////////////////

/*!
  \struct st_item
  \brief basic element for the symbol's table

  Each item correspond to an user object. These objects can be
  variables, function names or function parameters. The object
  correspond to a basic data type (integer, real an so on) or to
  an structurated one (pointer, array, struct -class-). For structured
  datatypes, an auxiliar structure is needed.
*/ 
struct st_item{
  char name[50];           //!<Label for the item
  enum category categ;     //!<Object structure code
  enum datatype type;      //!<Datatype of the item
  void *ref;               //!<Reference to other subtables
  struct st_item *next;    //!<Pointer to the next element with the same key
};

/*!
  \struct pt_item
  \brief basic element for the pointer's table

  Each item represents a pointer to an user object. The object can be 
  another pointer, so this representation allows to nest several 
  pointer definitions.
*/
struct pt_item{
  enum datatype type;      //!<Type of the pointed element
  void *ref;               //!<Reference to another subtable
};


/*!
  \struct at_item
  \brief basic element for the array's table

  Each item represents an array defined by an user. This structure
  stores the type of the elements of the array and its size. Fuheermore, 
  the "ref" slot is used to join with complex items (arrays of arrays,
  structures or pointers)
*/
struct at_item{
  enum datatype type;          //!<Type of the array's elements
  int size;                    //!<Number of elements
  void *ref;                   //!<Reference to other subtables
};


/*!
  \struct rt_item
  \brief basic element for the record's table
*/
struct rt_item{
  char name[50];               //!<Name for the slot
  enum datatype type;          //!<Type of the record's slot
  void *ref;                   //!<Reference to another subtable
  struct rt_item *next;        //!<Pointer for the next slot in the structure
};




//Type definition for items of symbol table variables 
typedef struct st_item st_item;
typedef struct at_item at_item;

//Type definition for symbol table variables
typedef struct st_item *symbol_table;

//A variable with type symbol_table must be created as follows
//symbol_table symtab[NKEYS]; or
//symbol_table *symtab;


//////////////////////////////////////////////////////
// access functions
//////////////////////////////////////////////////////

//----------------------------------------------------
/*! \fn symbol_table* st_newtable() 
    \return Pointer to the new table of symbols
    \brief  Creates a new empty symbol's table
*/
symbol_table* st_newtable();
//---------------------------------------------------

//----------------------------------------------------
/*! \fn void st_deltable( symbol_table *st )
    \param symbol_table *st table of symbols
    \exception table hasn't been created
    \brief Deletes the content of the symbol's table 
    and frees the memory
*/
void st_deltable( symbol_table *st );
//----------------------------------------------------

//----------------------------------------------------
/*! \fn st_printtable( symbol_table *st )
    \param symbol_table *st table of symbols
    \exception table hasn't been created
    \brief Shows the content of the symbol's table on the screen
*/
void st_printtable( symbol_table *st );
//---------------------------------------------------

//----------------------------------------------------
/*! \fn st_search( const char *elem, symbol_table *st )
    \param  char *elem       string to be found
    \param  symbol_table *st table of symbols
    \return Pointer to the element or NULL if not found
    \exception table hasn't been created
    \brief  Search the item "elem" in the table.
*/
st_item* st_search( const char *elem, symbol_table *st );
//----------------------------------------------------

//----------------------------------------------------
/*! \fn void st_add_symbol( const char *name, int cat, 
                            int type, symbol_table *st )
    \param const char *name
    \param in cat
    \param it type
    \param symbol_table *st table of symbols
    \exception table hasn't been created
    \brief adds a new entry in the symbol's table
*/
void st_add_symbol( const char *name, 
		    int cat, 
		    int type, 
		    symbol_table *st );
//----------------------------------------------------


//----------------------------------------------------
/*! \fn void st_add_symbol( st_item *elem, symbol_table *st )
    \param  char *elem       string to be found
    \param  symbol_table *st table of symbols
    \exception table hasn't been created
    \brief adds a new entry in the symbol's table
*/
void st_add_array( const char *name,
		   int cat,
		   int type,
		   int subtype,
		   int size,
		   symbol_table *st );
//----------------------------------------------------


//----------------------------------------------------
/*! \fn st_del( const char *elem, symbol_table *st )
    \param char *elem
    \param symbol_table *st
    \exception table hasn't been created
    \brief deletes the item "elem" from the table
*/
void st_del( const char *elem, symbol_table *st );
//----------------------------------------------------


#endif //_SYMTAB_H_
