/*
  file: parser.c
  date: feb 05
  author: G. Aranda

  purpose: translate xml file with metarule definitions into C code

  based on kdm parser by M. Rebollo
  based on entities parser by J. Fabregat

*/

#include <stdio.h>
#include <string.h>
#include <expat.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include <ctype.h>
#include "parser.h"

#define MAXBUFF 255



/************************************************
 
METARULES-RELATED DATA STRUCTURES AND DEFINITIONS

*************************************************/

/// Hypothetical maximum number of conditions
#define MAXCONDITIONS 255

/// Hypothetical maximum number of actions
#define MAXACTIONS 255

/// Hypothetical maximum number of meta-rules
#define MAXRULES 255

/// Hypothetical maximum number of attention focuses
#define MAXAFS 255

/// Structure defining a meta-rule
typedef struct mrule_ {
	/// Meta-rule's id
	char id[MAXBUFF];
	//char name[MAXBUFF];  // IGNORED
	//int number;  // IGNORED
	/// Metarule's type. See ARTIS documentation
	int type;
	/// Slot related to the event that triggers the meta-rule.
	/// Takes the "class.instance.slot" syntax.
	/// Empty if the event is of "CON_PERIOD" type
	char slot[MAXBUFF];
	/// Array of indexes to the conditions global array
	int conds[MAXCONDITIONS];
	/// Array of indexes to the actions global array
	int acts[MAXACTIONS];
} mrule;

/// Structure defining an attention focus
typedef struct atfocus_ {
	/// Focus' id
	char id[MAXBUFF];
	/// Focus's name
	char name[MAXBUFF];
	/// Array formed with the ids of the meta-rules conforming this focus
	char *rules[MAXRULES];
	/// Boolean active state.
	/// 0 = inactive , 1 = active
	int active;
} atfocus;

/// Userdata structure for the parser
typedef struct userdata{
  XML_Parser parser;
  int wrong_version;
  int depth;
  char cname[MAXBUFF];
  char sname[MAXBUFF];
  char stype[MAXBUFF];
  char slength[MAXBUFF];
  char elemtype[MAXBUFF];
  char iname[MAXBUFF];
  char svalue[MAXBUFF];
}XML_State;

/// Formal structure containing a decomposed condition
typedef struct condition_formal_ {
	/// The first operand.
	/// If it's a slot, it takes the "class.instance.slot" form
	char operand1[MAXBUFF];
	/// Type of the first operand
	int operand1_t;
	/// Operator code. See ARTIS documentation
	int operator;
	/// The second operand
	/// If it's a slot, it takes the "class.instance.slot" form
	char operand2[MAXBUFF];
	/// Type of the second operand
	int operand2_t;
} condition_formal;

/// Formal structure containing a decomposed action
typedef struct action_formal_ {
	/// Operator code. See ARTIS documentation
	int operator;
	/// First argument's type (mandatory)
	int arg1_t;
	/// Second argument's type (optional, depending on the action code)
	int arg2_t;
	/// Third argument's type (only for the special "code 8" case)
	int arg3_t;
	/// First argument (mandatory)
	char arg1[MAXBUFF];
	/// Second argument (optional, depending on the action code)
	char arg2[MAXBUFF];
	/// Third argument (only for the special "code 8" case)
	char arg3[MAXBUFF];
} action_formal;

/// Pointer to the current attention focus during the parsing
atfocus *current_af;

/// Current first level tag during the parsing
char section[MAXBUFF];

/// Current second level tag during the parsing
char current_tag[MAXBUFF];

/// Current atom class.
/// 0 = condition , 1 = action , -1 = neither
int atom_class;

/// List of attention focuses
atfocus *afs[MAXAFS];

/// List of meta-rules
mrule *mrules[MAXRULES];

/// Index with a dual purpose: the first time, it represents
/// the first free element in the active focuses list.
/// The second time, it acts as the iterator of such list
int free_af;

/// Index with a dual purpose: the first time, it represents
/// the first free element in the meta-rules list.
/// The second time, it acts as the iterator of such list
int free_mr;

/// List of conditions in "raw text" format
char *conditions[MAXBUFF];

/// List of actions in "raw text" format
char *actions[MAXBUFF];

/// Index with a dual purpose: the first time, it represents
/// the first free element in the conditions lists.
/// The second time, it acts as the iterator of such lists
int free_co;


/// Index with a dual purpose: the first time, it represents
/// the first free element in the actions lists.
/// The second time, it acts as the iterator of such lists
int free_ac;

/// List of conditions stored in the structured formal way
condition_formal *cfs[MAXCONDITIONS];

/// List of actions stored in the structured formal way
action_formal *actfs[MAXACTIONS];

/// Structure for date/time calculations
struct timeval tv;

/// Structure for date/time calculations
struct timezone tz;

/// Buffer for storing current date and time
char date_buff[26];

//XML_State *ud;

// Function headers
void add_condition(int mr, int co);
void add_action(int mr, int ac);
void init_mrule(int mr);
void init_af(int af);
void add_mrule(int af, char *id);
void generate_condition_formals(void);
void generate_action_formals(void);
void generateCode(void);
int parametersCount(int *neventtypes, int *nexprused, int *nactused, int *nmetarules, int *nattfocus, int *nactivefocus);
void analyse( const char *filename );


enum deeplevel{ TBNLVL, CLASSLVL, SLOTLVL, ARRAYLVL, CONSTLVL };


static void
//------------------------------------------------------------
xmldecl_handler(void            *userData,
                const XML_Char  *version,
                const XML_Char  *encoding,
                int              standalone)
//------------------------------------------------------------
{
  static const XML_Char Version_1_0[] = {'1', '.', '0', 0};

  int i, len;  
  XML_State *userdata = userData;

  len = strlen( Version_1_0 );
  for (i = 0; i < len; i++) {
    if (version[i] != Version_1_0[i]) {
      userdata->wrong_version = 1;
      //clear all handlers
      XML_SetCharacterDataHandler(userdata->parser, NULL);
      XML_SetElementHandler( userdata->parser, NULL, NULL );
      XML_SetCharacterDataHandler( userdata->parser, NULL );
      XML_SetCommentHandler( userdata->parser, NULL );
      XML_SetProcessingInstructionHandler( userdata->parser, NULL );
      fprintf( stderr, "Wrong version of XML document\n" ); 
      exit( -1 );
    } //if wrong version detected
  }//for i
} //xmldecl_handler()


/// Handler for starting elements
static void
//------------------------------------------------------------
xmlstartelem_handler( void        *userData,
		      const char  *el, 
		      const char  **attr )
//------------------------------------------------------------
{
  const char *next = NULL;
  int i = 0;

  XML_State *userdata = userData;

 	/* Aqui codigo a ejecutar al leer la <etiqueta> */
	// En '*el' tienes la cadena 'etiqueta' y en '**attrib' un vector de elementos rollo argc

  // DEBUG
  /*printf("##LEIDA LA ETIQUETA: %s\n", el);

  if (attr != NULL) {
	  printf("###CON LOS ATRIBUTOS:\n");
	  next = attr[i];
	  while (next != NULL) {
		printf("#### %d ## %s\n", i, next);
		i++;
		next = attr[i];
	  };
		  
  };*/
  // END DEBUG

  // "switch"
  if ( strcmp(el, "control") == 0 ) return;
  if ( strcmp(el, "atfocus") == 0 ) {
	  afs[free_af] = (atfocus *)malloc(sizeof(atfocus));
	  init_af(free_af);
	  strcpy(afs[free_af]->id, attr[1]);
	  //printf("\nADDED ID %s TO ATTENTION FOCUS !!!\n\n", afs[free_af]->id);
	  strcpy(section, "atfocus");
	  return;
  }; 
  if ( strcmp(el, "name") == 0 ) {
	  strcpy( current_tag, "name" );
	  return;
  };
  if ( strcmp(el, "mrule") == 0 ) {
	  mrules[free_mr] = (mrule *)malloc(sizeof(mrule));
	  init_mrule(free_mr);
	  strcpy(mrules[free_mr]->id, attr[1]);
	  //printf("\nADDED ID %s TO METARULE !!!\n\n", mrules[free_mr]->id);
	  strcpy(section, "mrule");
	  return;
  }; 
  if ( strcmp(el, "event") == 0 ) {
	  strcpy( current_tag, "event" );
	  if ( strcmp(attr[1], "MODIFICATION") == 0 ) mrules[free_mr]->type = 1;
	  else if ( strcmp(attr[1], "DELETION") == 0 ) mrules[free_mr]->type = 2;
	  else if ( strcmp(attr[1], "NEW_PRED") == 0 ) mrules[free_mr]->type = 3;
	  else if ( strcmp(attr[1], "MATCHED_PRED") == 0 ) mrules[free_mr]->type = 4;
	  else if ( strcmp(attr[1], "CON_PERIOD") == 0 ) mrules[free_mr]->type = 5;
	  else if ( strcmp(attr[1], "UNMATCHED_PRED") == 0 ) mrules[free_mr]->type = 6;
  };
  if ( strcmp(el, "condition") == 0 ) {
	  strcpy( current_tag, "condition" );
	  return;
  };
  if ( strcmp(el, "action") == 0 ) {
	  strcpy( current_tag, "action" );
	  return;
  };
  if ( strcmp(el, "atom") == 0 ) {
	  if ( strcmp(current_tag, "condition" ) == 0  ) atom_class = 0;
	  else if ( strcmp(current_tag, "action" ) == 0  ) atom_class = 1;
	  return;
  };
  if ( strcmp(el, "element") == 0 ) {
	  add_mrule(free_af, (char *)attr[3]); 
  	  return;
  };
  if ( strcmp(el, "active") == 0 ) {
	  strcpy( current_tag, "active" );
  };
  
} //xmlstartelem_handler()


/// Handler for ending elements
static void
//------------------------------------------------------------
xmlendelem_handler( void        *userData, 
		    const char  *el )
//------------------------------------------------------------
{
  XML_State *userdata = userData;

 /* Aqui codigo a ejecutar al leer la </etiqueta> */
	// En '*el' tienes la cadena 'etiqueta'

  //printf("##ACABADA LA ETIQUETA: %s\n", el);

  if ( strcmp(el, "atfocus") == 0 ) {
  	free_af = free_af+1;  // check overflow
  	//printf("\nFREE AF INCREMENTED: %d\n\n", free_af);
  } else
  if ( strcmp(el, "mrule") == 0 ) {
  	free_mr = free_mr+1;  // check overflow
  	//printf("\nFREE MR INCREMENTED: %d\n\n", free_mr);
  } else
  if ( strcmp(el, "atom") == 0 ) {
	if ( strcmp(current_tag, "condition" ) == 0 ) {
		add_condition(free_mr, free_co);
		free_co++;  // Check overflow
		atom_class = -1;
	} else 
	if ( strcmp(current_tag, "action" ) == 0 ) {
		add_action(free_mr, free_ac);
		free_ac++;  // Check overflow
		atom_class = -1;
	};
	return;  // Because we are not efectively changing the tag
  } else
  if ( strcmp(el, "condition") == 0 ) atom_class = -1;
  else if ( strcmp(el, "action") == 0 ) atom_class = -1;
  
  strcpy( current_tag, "\0");

  return;
  
} //xmlendelem_handler() 


/// Handler for text contained in an element
static void
//------------------------------------------------------------
xmltext_handler( void       *userData, 
		 const char *el, 
		 int        len )
//------------------------------------------------------------
{
  char *aux;
  aux = (char *)malloc((len+1)*sizeof(char));
	
  XML_State *userdata = userData;

 	/* Aqui codigo a ejecutar al leer el texto entre etiquetas */
 	// En '*el' tienes el puntero a la cadena que queda por parsear
	// En len tienes el numero de caracteres entre la etiqueta de apertura y la de cierre

  if ( strcmp(section, "atfocus") == 0) {
  	if ( strcmp(current_tag, "name") == 0 ) {
		strncpy(aux, el, len);
		strcat(aux, "\0");
		strcpy(afs[free_af]->name, aux);
		//printf("\nADDED NAME %s TO ATTENTION FOCUS %d!!!\n\n", afs[free_af]->name, free_af);
		return;
  	};
  };
  if ( strcmp(current_tag, "event") == 0 ) {
	  if ( mrules[free_mr]->type != 5 ) {
	  	strncpy(aux, el, len);
	  	strcat(aux, "\0");
	  	strcpy(mrules[free_mr]->slot, aux);
	  	//printf("\nADDED SLOT %s TO METARULE %d !!!\n\n", mrules[free_mr]->slot, free_mr);
	  	return;
	  };
  };
  if ( strcmp(current_tag, "condition") == 0 ) {
	  if (atom_class == 0) {
	  	strncpy(aux, el, len);
	  	strcat(aux, "\0");
	  	//printf("CONDITION: len=%d\n%s\n", len, aux);
		conditions[free_co] = (char *) malloc((len+1)*sizeof(char));
		strcpy(conditions[free_co], aux);
		return;
	  };
  };
  if ( strcmp(current_tag, "action") == 0 ) {
	  if (atom_class == 1) {
	  	strncpy(aux, el, len);
	  	strcat(aux, "\0");
	  	//printf("ACTION: len=%d\n%s\n", len, aux);
		actions[free_ac] = (char *) malloc((len+1)*sizeof(char));
		strcpy(actions[free_ac], aux);
		return;
	  };
  };
  if ( strcmp(current_tag, "active") == 0 ) {
	  strncpy(aux, el, len);
	  strcat(aux, "\0");
	  if ( strcmp(aux, "yes") == 0 ) {
		afs[free_af]->active = 1;
	  };
	  return;
  };
 
} //xmltext_handler()

/// Prints the symbols' tables
void
//------------------------------------------------------------
printSymbols(void)
//------------------------------------------------------------
{

  int i = 0;
  int j = 0;
 
  printf("\n\nSYMBOLS\n=======\n");
  printf("ATTENTION FOCUSES\n---\n");
  for ( i = 0 ; i < MAXAFS ; i++ ) {
      if ( afs[i] != NULL ) {
	      printf("[%d] Id: %s\n", i, afs[i]->id);
	      printf("    Name: %s\n", afs[i]->name);
	      printf("    Rules:\n");
	      j = 0;
	      while ( afs[i]->rules[j] != NULL  ) {
	      	printf("      %s\n", afs[i]->rules[j]);
		j++;
	      };
      };
  };
  printf("\nMETARULES\n---------\n");
  for ( i = 0 ; i < MAXRULES ; i++ ) {
      if ( mrules[i] != NULL ) {
	      printf("[%d] Id: %s\n", i, mrules[i]->id);
	      printf("    Type: %d\n", mrules[i]->type);
	      if (mrules[i]->type != 5)
	      	printf("    Slot: %s\n", mrules[i]->slot);
	      printf("     Conditions:\n");
	      j = 0;
	      while (mrules[i]->conds[j] != -1) {
	      	printf("       %s\n", conditions[mrules[i]->conds[j]]);
		j++;
	      };
	      printf("     Actions:\n");
	      j = 0;
	      while (mrules[i]->acts[j] != -1) {
	      	printf("       %s\n", actions[mrules[i]->acts[j]]);
		j++;
	      };;
      };
  };

  printf("\nCONDITIONS\n----------\n");
  for ( i = 0 ; i < MAXCONDITIONS; i++ ) {
	if ( conditions[i] != NULL ) {
		printf("[%d] %s\n", i, conditions[i]);
	};
  };

  printf("\nCONDITION FORMALS\n-----------------\n");
  for ( i = 0 ; i < MAXCONDITIONS; i++ ) {
 	if ( cfs[i] != NULL ) {
		printf("[%d] Operand1: %s\n",i, cfs[i]->operand1);
		printf("    Operator Number: %d\n", cfs[i]->operator);	
		printf("    Operand2: %s\n", cfs[i]->operand2);	
	};
  }; 
  
  printf("\nACTIONS\n-------\n");
  for ( i = 0 ; i < MAXACTIONS; i++ ) {
	if ( actions[i] != NULL ) {
		printf("[%d] %s\n", i, actions[i]);
	};
  };

  printf("\nACTION FORMALS\n-----------------\n");
  for ( i = 0 ; i < MAXCONDITIONS; i++ ) {
 	if ( actfs[i] != NULL ) {
		printf("[%d] Operator Number: %d\n", i, actfs[i]->operator);
		printf("    Arg1: %s\n",actfs[i]->arg1);
		printf("    Arg1 Type: %d\n", actfs[i]->arg1_t);
		if ( strcmp(actfs[i]->arg2, "-1") != 0 ) {
			printf("    Arg2: %s\n", actfs[i]->arg2);
			printf("    Arg2 Type: %d\n", actfs[i]->arg2_t);
		};
		if ( strcmp(actfs[i]->arg3, "-1") != 0 ) {
			printf("    Arg3: %s\n", actfs[i]->arg3);
			printf("    Arg3 Type: %d\n", actfs[i]->arg3_t);
		};
	};
  }; 

  printf("\n");

  return;

};

/// Initialize a meta-rule structure before using it
/// @param mr Index of the meta-rule structure to initialize
void
//-----------------------------------------------------------
init_mrule( int mr )
//-----------------------------------------------------------
{
	int i = 0;

	for ( i = 0 ; i < MAXCONDITIONS ; i++ ) {
		mrules[mr]->conds[i] = -1;
	};
	for ( i = 0 ; i < MAXACTIONS ; i++ ) {
		mrules[mr]->acts[i] = -1;
	};

	return; 
};

/// Initialize an attention focus structure before using it
/// @param af Index of the attention focus structure to initialize
void
//-----------------------------------------------------------
init_af( int af )
//-----------------------------------------------------------
{
	int i = 0;

	for ( i = 0 ; i < MAXRULES ; i++) {
		afs[af]->rules[i] = NULL;	
	};

	afs[af]->active = 0;

	return;
};

/// Add a condition to a meta-rule
/// @param mr Index of the meta-rule to add a condition to
/// @param co Index of the condition to be added
void
//-----------------------------------------------------------
add_condition( int mr, int co )
//-----------------------------------------------------------
{
	int i = 0;

	while (mrules[mr]->conds[i] != -1) i++;
	mrules[mr]->conds[i] = co;
	return;	
};

/// Add an action to a meta-rule
/// @param mr Index of the meta-rule to add an action to
/// @param ac Index of the action to be added
void
//-----------------------------------------------------------
add_action( int mr, int ac )
//-----------------------------------------------------------
{
	int i = 0;

	while (mrules[mr]->acts[i] != -1) i++;
	mrules[mr]->acts[i] = ac;
	return;	
};


/// Add a meta-rule to an attention focus
/// @param af Index of the attention focus
/// @param id Textual id of the meta-rule
void
//-----------------------------------------------------------
add_mrule(int af, char *id)
//-----------------------------------------------------------
{
	int i = 0;

	while (afs[af]->rules[i] != NULL) i++;
		afs[af]->rules[i] = (char *) malloc(strlen(id)*sizeof(char));
		strcpy(afs[af]->rules[i], id);
	return;
};

/// Deduce the type of an operand from its syntax
/// It currently identifies the next types:
/// - Integer (code 3)
/// - Float (code 4)
/// - Slot (code 1)
/// @param s String containing the operand
/// @return Code of the type
int
//-----------------------------------------------------------
get_operand_type(char *s)
//-----------------------------------------------------------
{
	int i = 0;
	int check = 0;

	// Checking for integer
	for ( i = 0 ; i < strlen(s) ; i++ ) {
		if ( isdigit(s[i]) == 0 ) {check = 1; break;}
	};
	if (check == 0) return 3;

	// Check for float
	check = 0;
	for ( i = 0 ; i < strlen(s) ; i++ ) {
		if ( (isdigit(s[i]) || ispunct(s[i]) ) == 0 ) {check = 1; break;}
	};
	if (check == 0) return 4;

	// Check for slot
	check = 0;
	for ( i = 0 ; i < strlen(s) ; i++ ) {
		if ( (isalnum(s[i]) || ispunct(s[i]) ) == 0 ) {check = 1; break;}
	};
	if (check == 0) return 1;

	//TODO: Check for the rest of types
};

/// Generates the formal structures of the conditions
/// It parses the conditions in the "raw text" format and, for each
/// and everyone of them, generates a formal structure of the decomposed
/// condition
void
//-----------------------------------------------------------
generate_condition_formals(void)
//-----------------------------------------------------------
{
	int i = 0;
	int j = 0;
	int k = 0;
	char aux[MAXBUFF];
	char **ap;
	char *argv[3];
	char *inputstring;
	char *s;
	int ot = -1;
	
	while (conditions[i] != NULL) {
		inputstring = (char *) malloc(strlen(conditions[i])*sizeof(char));
		strcpy(inputstring, conditions[i]);
		
		// Copied from "strsep" man page
		for (ap = argv; (*ap = strsep(&inputstring, " \t")) != NULL;)
                   if (**ap != '\0')
                           if (++ap >= &argv[3])
                                   break;

		//printf("GDF: %s %s %s\n", argv[0], argv[1], argv[2]);
	
		cfs[i] = (condition_formal *) malloc(sizeof(condition_formal));
		strcpy(cfs[i]->operand1, argv[0]);
		strcpy(cfs[i]->operand2, argv[2]);
		cfs[i]->operator = -1;
		
		// "switch" operator
		if ( strcmp(argv[1], "") == 0 ) cfs[i]->operator = 0;
		else if ( strcmp(argv[1], "<>") == 0 ) cfs[i]->operator = 1;
		else if ( strcmp(argv[1], "<") == 0 ) cfs[i]->operator = 2;
		else if ( strcmp(argv[1], ">") == 0 ) cfs[i]->operator = 3;
		else if ( strcmp(argv[1], ">=") == 0 ) cfs[i]->operator = 4;
		else if ( strcmp(argv[1], "<=") == 0 ) cfs[i]->operator = 5;
		else if ( strcmp(argv[1], "=") == 0 ) cfs[i]->operator = 6;
		else if ( strcmp(argv[1], "+") == 0 ) cfs[i]->operator = 7;
		else if ( strcmp(argv[1], "-") == 0 ) cfs[i]->operator = 8;
		else if ( strcmp(argv[1], "*") == 0 ) cfs[i]->operator = 9;
		else if ( strcmp(argv[1], "/") == 0 ) cfs[i]->operator = 10;
		i++;
	}; 
	
	// Now we get the type of each operand and
	// turn all those "class.instance.slot" into "instance->slot"
	i = 0;
	while (cfs[i] != NULL) {
		// Operand 1
		s = cfs[i]->operand1;
		ot = get_operand_type(s);
		cfs[i]->operand1_t = ot;
		if ( cfs[i]->operand1_t == 1  ) {
			j = 0;
			while ( s[j] != '.') j++;
			j++;
			k=0;
			while ( s[j] != '.') {
				aux[k] = s[j];
				k++; j++;
			};
			j++;
			aux[k] = '-';
			k++;
			aux[k] = '>';
			k++;
			while ( s[j] != '\0') {
				aux[k] = s[j];
				k++; j++;
			};
			aux[k] = '\0';
			strcpy(cfs[i]->operand1, aux);
		};

		// Operand 2
		s = cfs[i]->operand2;
		ot = get_operand_type(s);
		cfs[i]->operand2_t = ot;
		if ( cfs[i]->operand2_t == 1  ) {
			j = 0;
			while ( s[j] != '.') j++;
			j++;
			k=0;
			while ( s[j] != '.') {
				aux[k] = s[j];
				k++; j++;
			};
			j++;
			aux[k] = '-';
			k++;
			aux[k] = '>';
			k++;
			while ( s[j] != '\0') {
				aux[k] = s[j];
				k++; j++;
			};
			aux[k] = '\0';
			strcpy(cfs[i]->operand2, aux);
		};
	i++;
	};

	return;
};



/// Generates the formal structures of the actions
/// It parses the actions in the "raw text" format and, for each
/// and everyone of them, generates a formal structure of the decomposed
/// action
void
//-----------------------------------------------------------
generate_action_formals(void)
//-----------------------------------------------------------
{
	int i = 0;
	int j = 0;
	int k = 0;
	int l = 0;
	char aux[MAXBUFF];
	char **ap;
	char *argv[4];  // Arguments extracted from raw action
	char *ef_args[4];  // Effective arguments used (post-processed)
	char *inputstring;
	char *s;
	int ot = -1;
	
	while (actions[i] != NULL) {
		inputstring = (char *) malloc(strlen(actions[i])*sizeof(char));
		strcpy(inputstring, actions[i]);

		// Initially, all the "arguments" are NULL
		argv[0] = NULL;
		argv[1] = NULL;
		argv[2] = NULL;
		argv[3] = NULL;
		ef_args[0] = NULL;
		ef_args[1] = NULL;
		ef_args[2] = NULL;
		ef_args[3] = NULL;
		
		// Copied from "strsep" man page
		for (ap = argv; (*ap = strsep(&inputstring, " \t")) != NULL;)
                   if (**ap != '\0')
                           if (++ap >= &argv[5])
                                   break;

		//printf("GAF: %s %s %s %s\n", argv[0], argv[1], argv[2], argv[3]);

		// The next loop filters undesired control "characters" that may have
		// been inserted into the "argv" array due to the previous loop operations
		for ( l = 0 ; l < 4 ; l++ ) {
			if ( argv[l] != NULL ) {
				ef_args[l] = (char *) malloc( (strlen(argv[l])+1) * sizeof(char) );
				strncpy(ef_args[l], argv[l], strlen(argv[l]));
				strcat(ef_args[l], "\0");
				for ( j=0 ; j<strlen(ef_args[l]); j++ ) {
					if ( ef_args[l][j] == '\31' ) ef_args[l][j] = '\0';
					if ( ef_args[l][j] == '\21' ) ef_args[l][j] = '\0';
				};
			};
		};

		// For later uses
		j = 0;
		
		// And the proper arguments are linked back to "argv"
		argv[0] = ef_args[0];
		argv[1] = ef_args[1];
		argv[2] = ef_args[2];
		argv[3] = ef_args[3];
		
		// If an argument is not present, argX is set to "\0"
		actfs[i] = (action_formal *) malloc(sizeof(action_formal));
		if ( argv[1] == NULL ) {
			strcpy( actfs[i]->arg1, "-1" );
		} else {
			sprintf(actfs[i]->arg1, "%s\0", argv[1]);
		};
		if ( argv[2] != NULL ) strcpy(actfs[i]->arg2, argv[2]); else strcpy( actfs[i]->arg2, "-1" );
		if ( argv[3] != NULL ) strcpy(actfs[i]->arg3, argv[3]); else strcpy( actfs[i]->arg3, "-1" );

		// "switch" operator
		if ( strcmp(argv[0], "KDM_Set_Value") == 0 ) actfs[i]->operator = 0;
		if ( strcmp(argv[0], "Assign") == 0 ) actfs[i]->operator = 0;
		if ( strcmp(argv[0], "FunctionCall") == 0 ) actfs[i]->operator = 1;
		if ( strcmp(argv[0], "TriggerSporadicLevel") == 0 ) actfs[i]->operator = 2;
		if ( strcmp(argv[0], "SetReactivityDegree") == 0 ) actfs[i]->operator = 3;
		if ( strcmp(argv[0], "SetPolicy") == 0 ) actfs[i]->operator = 4;
		if ( strcmp(argv[0], "SetFocus") == 0 ) actfs[i]->operator = 5;
		if ( strcmp(argv[0], "SetAttentionDegree") == 0 ) actfs[i]->operator = 6;
		// Number 7 replicated due to lexicon conflicts (Behavior VS Behaviour)
		if ( strcmp(argv[0], "SetBehaviour") == 0 ) actfs[i]->operator = 7;
		if ( strcmp(argv[0], "SetBehavior") == 0 ) actfs[i]->operator = 7; 
		if ( strcmp(argv[0], "KDM_Add_Event") == 0 ) actfs[i]->operator = 8;
		if ( strcmp(argv[0], "KDM_Release_Event") == 0 ) actfs[i]->operator = 9;
		if ( strcmp(argv[0], "Trigger_Optional") == 0 ) actfs[i]->operator = 10;
		
		//printf("GAF: %d %s %s %s\n", actfs[i]->operator, actfs[i]->arg1, actfs[i]->arg2, actfs[i]->arg3);
		
		i++;
	};

	
	// Now we get the type of each operand and
	// turn all those "class.instance.slot" into "instance->slot"
	i = 0;
	while (actfs[i] != NULL) {
		// Argument 1
		s = actfs[i]->arg1;
		ot = get_operand_type(s);
		actfs[i]->arg1_t = ot;
		if ( actfs[i]->arg1_t == 1  ) {
			j = 0;
			while ( s[j] != '.') j++;
			j++;
			k=0;
			while ( s[j] != '.') {
				aux[k] = s[j];
				k++; j++;
			};
			j++;
			aux[k] = '-';
			k++;
			aux[k] = '>';
			k++;
			while ( s[j] != '\0') {
				aux[k] = s[j];
				k++; j++;
			};
			aux[k] = '\0';
			strcpy(actfs[i]->arg1, aux);
		};

		// Argument 2
		if ( strcmp(actfs[i]->arg2, "-1") != 0 ) {
			s = actfs[i]->arg2;
			ot = get_operand_type(s);
			actfs[i]->arg2_t = ot;
			if ( actfs[i]->arg2_t == 1  ) {
				j = 0;
				while ( s[j] != '.') j++;
				j++;
				k=0;
				while ( s[j] != '.') {
					aux[k] = s[j];
					k++; j++;
				};
				j++;
				aux[k] = '-';
				k++;
				aux[k] = '>';
				k++;
				while ( s[j] != '\0') {
					aux[k] = s[j];
					k++; j++;
				};
				aux[k] = '\0';
				strcpy(actfs[i]->arg2, aux);
			};
		};

		// Argument 3
		if ( strcmp(actfs[i]->arg3, "-1") != 0 ) {
			s = actfs[i]->arg3;
			ot = get_operand_type(s);
			actfs[i]->arg3_t = ot;
			if ( actfs[i]->arg3_t == 1  ) {
				j = 0;
				while ( s[j] != '.') j++;
				j++;
				k=0;
				while ( s[j] != '.') {
					aux[k] = s[j];
					k++; j++;
				};
				j++;
				aux[k] = '-';
				k++;
				aux[k] = '>';
				k++;
				while ( s[j] != '\0') {
					aux[k] = s[j];
					k++; j++;
				};
				aux[k] = '\0';
				strcpy(actfs[i]->arg3, aux);
			};
		};

	i++;
	};
	
	return;
};

/// Converts from "class.instance.slot" notation into "instance->slot" notation
void slot_to_pointer(char *pSlot, char*orig) {
	char *s;
	int j = 0 , k = 0;
	char aux[MAXBUFF];
	
	s = orig;
	j = 0;
	while ( s[j] != '.') j++;
	j++;
	k=0;
	while ( s[j] != '.') {
		aux[k] = s[j];
		k++; j++;
	};
	j++;
	aux[k] = '-';
	k++;
	aux[k] = '>';
	k++;
	while ( s[j] != '\0') {
		aux[k] = s[j];
		k++; j++;
	};
	aux[k] = '\0';
	strcpy(pSlot, aux);

	return;
};

int get_rule_by_id(char *id) {
	int i = 0;

	for ( i = 0 ; i < MAXRULES ; i++ ) {
		if (mrules[i] != NULL) {
			if ( strcmp(id, mrules[i]->id) == 0) {
				return i;
			};
		};
	};
};

/// Calculates several global parameters for the store calculations
/// It currently succesfully calculates:
/// - The number of event types
/// - The number of expressions (in this case, conditions) created
/// - The number of actions created
/// - The number of meta-rules created
/// - The number of attention focuses created
/// - The number of ACTIVE attention focuses
/// @param neventtypes Number of event types
/// @param nexprused Number of expressions (conditions)
/// @param nactused Number of actions used
/// @param nmetarules Number of meta-rules
/// @param nattfocus Number of attention focuses
/// @param nactivefocus Number of ACTIVE attention focuses
int
//-----------------------------------------------------------
parametersCount(int *neventtypes, int *nexprused, int *nactused, int *nmetarules, int *nattfocus, int *nactivefocus)
//-----------------------------------------------------------
{
	int i, j = 0;

	// NEVENTTYPES
	// The next loop counts how many DIFFERENT event types have
	// been defined
	*neventtypes = 0;
	
	for ( i = 0 ; i < free_mr ; i++ ) {
		for ( j = 0 ; j < i ; j++  ) {
			if ( mrules[i] != NULL ) {
				if ( mrules[i]->type == mrules[j]->type ) break;
			} else break;
		};
		if ( i == j ) (*neventtypes)++;
	};

	// NEXPRUSED
	// Number of expressions used by the designer
	(*nexprused) = free_co;

	// NACTUSED
	(*nactused) = free_ac;

	// NMETARULES
	(*nmetarules) = free_mr;

	// NATTFOCUS
	(*nattfocus) = free_af;

	// NACTIVEFOCUS
	(*nactivefocus) = 0;
	for ( i = 0 ; i < free_af ; i++ ) {
		if ( afs[i] != NULL ) {
			if ( afs[i]->active == 1 ) {
				(*nactivefocus)++;
			};
		};
	};

	return 0;
};


/// Get the numeric equivalent for a an attention focus given its high level alphanumeric id
/// Most actions regarding attention focuses won't use their ids at all, but instead
/// will use an internal numeric value representing each attention focus.
/// This function returns such value given a certain id
/// @param id The attention focus' id
/// @return The attention focus' numeric value or -1 in a catastrophic scenario
int 
//-----------------------------------------------------------
get_focus_from_id(char *id)
//-----------------------------------------------------------
{
	int i;

	for ( i = 0 ; i < MAXAFS ; i++) {
		if ( afs[i] != NULL )
			if ( strcmp(id, afs[i]->id) == 0 )
				return i;
	};

	// We REALLY REALLY should never get this far
	fprintf(stderr, "MAY DAY! MAY DAY! THERE IS NO SUCH ID IN ANY ATTENTION FOCUS! WE NEED SOME BACKUP!!");
	return -1;
};
	
/// Generate the ARTIS C code files.
/// The files currently generated are:
/// - attention_focus_definition.c
/// - attention_focus_definition.h
void
//-----------------------------------------------------------
generateCode(void)
//-----------------------------------------------------------
{
	FILE *fout;
	char etype[MAXBUFF];
	char pSlot[MAXBUFF];
	int i = 0;
	int rule_number = 0;
	int neventtypes, nexprused, nactused, nmetarules, nattfocus, nactivefocus; 
	// Set Focus
	int sf = -1;
	
  	if( (fout = fopen( "attention_focus_definition.c", "w" )) == NULL ){
    		fprintf( stderr, "Error opening file for writing: attention_focus_definition.c\n");
    	exit( -1 );
  	} // if error opening file

	fprintf(fout, "%s\n\n", header1);
	fprintf(fout, "%s\n\n", header2);
	fprintf(fout, "%s\n %s\n", header3, date_buff);
	fprintf(fout, "%s\n\n", header4);
	fprintf(fout, "%s\n", header5);
	fprintf(fout, "%s\n", header6);
	fprintf(fout, "%s\n", header7);
	fprintf(fout, "%s\n", header8);
	fprintf(fout, "%s\n", header9);
	fprintf(fout, "%s\n", header10);
	fprintf(fout, "%s\n", header11);
	fprintf(fout, "%s\n", header12);
	fprintf(fout, "%s\n", header13);
	fprintf(fout, "%s\n", header14);
	fprintf(fout, "%s\n", header15);
	fprintf(fout, "%s\n", header16);
	fprintf(fout, "%s\n", header17);
	fprintf(fout, "%s\n", header18);
	fprintf(fout, "%s\n", header19);
	fprintf(fout, "%s\n", header20);
	fprintf(fout, "%s\n", header21);
	fprintf(fout, "\n");
	fprintf(fout, "%s\n", include1);
	fprintf(fout, "%s\n", include2);
	fprintf(fout, "%s\n", include3);
	fprintf(fout, "%s\n", include4);
	fprintf(fout, "%s\n", include5);
	fprintf(fout, "%s\n", include6);
	
	// We fill the parameters needed later
	parametersCount(&neventtypes, &nexprused, &nactused, &nmetarules, &nattfocus, &nactivefocus); 
	
	fprintf(fout, "#define MAXCONDITIONS %d\n#define MAXACTIONS %d\n#define MAXRULES %d\n#define MAXAFS %d\n", nexprused, nactused, nmetarules, nattfocus);
	fprintf(fout, "\n");
	// AttentionFocusInit
	//printf("\nAFI\n");
	fprintf(fout, "%s\n", afi1);
	fprintf(fout, "%s\n", afi2);
	fprintf(fout, "%s\n", afi3);
	fprintf(fout, "%s\n", afi4);
	fprintf(fout, "%s\n", afi5);
	fprintf(fout, "%s\n", afi6);
	fprintf(fout, "%s\n", afi7);
	fprintf(fout, "%s\n", afi8);
	
	// The last parameter is "nattfocus*2" because every attention focus may potentially be an active
	// one, thus we need to make room for that possibility
	fprintf(fout, "\n\tif ((error = Meta_Rule_AF_Store_Creation( %d, %d, %d, %d, %d, %d )) < 0 )\n\t\t\treturn -1;\n\telse {\n", neventtypes, nexprused, nactused, nmetarules, nattfocus, nattfocus*2 );
	
	fprintf(fout, "%s\n", afi9);
	
	// THE Else
	//printf("\nELSE\n");


	// ATTENTION FOCUSES CREATION
	fprintf(fout, "\n");
	fprintf(fout, "\t\t// ATTENTION FOCUSES CREATION\n");
	free_af = 0;
	while ( (free_af < MAXAFS) && (afs[free_af] != NULL) ) {
		fprintf(fout, "\t\twhile ((nAttFocus[%d] = Meta_Rule_Get_Empty_Att_Focus_Node()) < 0 );\n" , free_af );
		free_af++;
	};
	free_af = 0;
	
	// Conditions
	//printf("\nCONDITIONS\n");
	fprintf(fout, "\n");
	free_co = 0;
	while ( cfs[free_co] != NULL ) {
		fprintf(fout, "\t\t// %s\n", conditions[free_co]);
		fprintf(fout, "\t\twhile ((nCond[%d] = Meta_Rule_Create_Expr( %d, %d, %d, &(%s), %s) ) < 0 );\n" , free_co, cfs[free_co]->operator, cfs[free_co]->operand1_t, cfs[free_co]->operand2_t, cfs[free_co]->operand1, cfs[free_co]->operand2 );
		free_co++;
	};

	// Actions
	//printf("\nACTIONS\n");
	fprintf(fout, "\n");
	fprintf(fout, "\t\t// ACTIONS\n");
	free_ac = 0;
	while ( actfs[free_ac] != NULL ) {
		// Print out a comment with the action expressed in formal language
		fprintf(fout, "\t\t// %s\n", actions[free_ac]);
		switch (actfs[free_ac]->operator) {
			case 1:
			case 2:
			case 3:
			case 4:
			case 6:
			case 7:
			case 10:
				fprintf(fout,  "\t\twhile ((nAct[%d] = Meta_Rule_Create_Act( %d, %s, -1 ) ) < 0 );\n" , free_ac, actfs[free_ac]->operator, actfs[free_ac]->arg1 );
				break;
			case 0:
				// The second argument is considered an INTEGER
				fprintf(fout,  "\t\twhile ((nAct[%d] = Meta_Rule_Create_Act( %d, &(%s), %s ) ) < 0 );\n" , free_ac, actfs[free_ac]->operator, actfs[free_ac]->arg1, actfs[free_ac]->arg2 );
				break;
			case 5:
				sf = get_focus_from_id(actfs[free_ac]->arg1);
				fprintf(fout,  "\t\twhile ((nAct[%d] = Meta_Rule_Create_Act( %d, nAttFocus[%d], %s ) ) < 0 );\n" , free_ac, actfs[free_ac]->operator, sf, actfs[free_ac]->arg2 );
				break;
			case 9:
				fprintf(fout,  "\t\twhile ((nAct[%d] = Meta_Rule_Create_Act( %d, &(%s), %s ) ) < 0 );\n" , free_ac, actfs[free_ac]->operator, actfs[free_ac]->arg1, actfs[free_ac]->arg2 );
				break;
			case 8:  // Dr.Evil's #8 Dirty Hack
				fprintf(fout,  "\t\twhile ((nAct[%d] = Meta_Rule_Create_Act( %d, &(%s), %s ) ) < 0 );\n" , free_ac, actfs[free_ac]->operator, actfs[free_ac]->arg1, actfs[free_ac]->arg2 );
				fprintf(fout,  "\t\twhile ((nAct8[%d] = Meta_Rule_Create_Act( %s, -1, -1 ) ) < 0 );\n" , free_ac, actfs[free_ac]->arg3 );
		};

		free_ac++;
         };

	// META-RULES
	//printf("\nMETARULES\n");
	fprintf(fout, "\n");
	fprintf(fout, "\t\t// METARULES\n");
	free_mr = 0;
	while ( mrules[free_mr] != NULL ) {
		if (mrules[free_mr]->type != 5) {
			slot_to_pointer(pSlot, mrules[free_mr]->slot);
			strcpy(mrules[free_mr]->slot, pSlot);
		};
		switch ( mrules[free_mr]->type ) {
			case 1:
				strcpy(etype, "MODIFICATION");
				if ( mrules[free_mr]->conds[0] != -1 )
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , &(%s), nCond[%d], nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->slot, mrules[free_mr]->conds[0], mrules[free_mr]->acts[0] );
				else
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , &(%s), -1, nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->slot, mrules[free_mr]->acts[0] );
					
				break;
			case 2:
				strcpy(etype, "DELETION");
				if ( mrules[free_mr]->conds[0] != -1 )
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , &(%s), nCond[%d], nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->slot, mrules[free_mr]->conds[0], mrules[free_mr]->acts[0] );
				else
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , &(%s), -1, nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->slot, mrules[free_mr]->acts[0] );
				break;
			case 3:
				strcpy(etype, "NEW_PREDICTION");
				if ( mrules[free_mr]->conds[0] != -1 )
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , &(%s), nCond[%d], nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->slot, mrules[free_mr]->conds[0], mrules[free_mr]->acts[0] );
				else
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , &(%s), -1, nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->slot, mrules[free_mr]->acts[0] );	
				break;
			case 4:
				strcpy(etype, "MATCHED_PREDICTION");
				if ( mrules[free_mr]->conds[0] != -1 )
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , &(%s), nCond[%d], nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->slot, mrules[free_mr]->conds[0], mrules[free_mr]->acts[0] );
				else
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , &(%s), -1, nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->slot, mrules[free_mr]->acts[0] );	
				break;
			case 6:
				strcpy(etype, "UNMATCHED_PREDICTION");
				if ( mrules[free_mr]->conds[0] != -1 )
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , &(%s), nCond[%d], nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->slot, mrules[free_mr]->conds[0], mrules[free_mr]->acts[0] );
				else 
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , &(%s), -1, nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->slot, mrules[free_mr]->acts[0] );
				break;
			case 5:
				strcpy(etype, "CONSIDER_PERIODIC_AGENT");
				if ( mrules[free_mr]->conds[0] != -1 )
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , NULL, nCond[%d], nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->conds[0], mrules[free_mr]->acts[0] );
				else
					fprintf(fout, "\t\twhile (( nMetaRule[%d] = Meta_Rule_Create ( %s , NULL, -1, nAct[%d] )) < 0 );\n" , free_mr, etype, mrules[free_mr]->acts[0] );
				break;
		};  // switch type

		// Maybe there are more conditions ...
		for ( i = 1 ; i < MAXCONDITIONS ; i++ ) {
			if ( mrules[free_mr]->conds[i] != -1  ) {
				fprintf(fout, "\t\twhile (( Meta_Rule_Insert_Cond ( nCond[%d] , nMetaRule[%d] )) <0 );\n" , mrules[free_mr]->conds[i], free_mr);
			};
		};
		
		// Maybe there are more actions ...
		// The #8 Dirty Hack Again !!!
		if ( mrules[free_mr]->acts[0] != -1 )
			if ( actfs[mrules[free_mr]->acts[0]] )
				if ( actfs[mrules[free_mr]->acts[0]]->operator == 8 ) {  // It's the evil hack again !!!
					fprintf(fout, "\t\twhile (( Meta_Rule_Insert_Act ( nAct8[%d] , nMetaRule[%d] )) <0 );\n" , mrules[free_mr]->acts[0], free_mr);
				};
			
		for ( i = 1 ; i < MAXACTIONS ; i++ ) {
			if ( mrules[free_mr]->acts[i] != -1  ) {
				fprintf(fout, "\t\twhile (( Meta_Rule_Insert_Act ( nAct[%d] , nMetaRule[%d] )) <0 );\n" , mrules[free_mr]->acts[i], free_mr);
				if ( actfs[mrules[free_mr]->acts[i]]->operator == 8  ) {
					fprintf(fout, "\t\twhile (( Meta_Rule_Insert_Act ( nAct8[%d] , nMetaRule[%d] )) <0 );\n" , mrules[free_mr]->acts[i], free_mr);	
				};
			};
		};
		
		free_mr++;
	};
	


	// ATTENTION FOCUSES' COMPLETION
	//printf("\nAFS\n");
	fprintf(fout, "\n");
	fprintf(fout, "\t\t// ATTENTION FOCUSES' COMPLETION\n");
	free_af = 0;
	while ( (free_af < MAXAFS) && (afs[free_af] != NULL) ) {
		for ( i = 0 ; i < MAXRULES ; i++ ) {
			if (afs[free_af]->rules[i] != NULL) {  // If there's a rule
				rule_number = get_rule_by_id(afs[free_af]->rules[i]);
				fprintf(fout, "\t\twhile ( Meta_Rule_Insert_In_Att_Focus( nMetaRule[%d], nAttFocus[%d] ) < 0 );\n" , rule_number, free_af);
			};
		};

		// If an Attention Focus shall be an active one, this is the time to activate it
		if ( afs[free_af]->active == 1 )
			fprintf(fout, "\t\twhile ( Meta_Rule_Activate_Attention_Focus( nAttFocus[%d], 1 ) < 0 );\n", free_af);

		free_af++;
	};	

	fprintf(fout, "\n\t\treturn 0;\n");
	fprintf(fout, "\t};\n};");
	
	//printf("\nTHE END OF THE WORLD AS WE KNOW IT\n");

	fclose(fout);

	fprintf(stdout, "Generated \"attention_focus_definition.c\"\n");

	if( (fout = fopen( "attention_focus_definition.h", "w" )) == NULL ){
    		fprintf( stderr, "Error opening file for writing: attention_focus_definition.c\n");
    	exit( -1 );
  	} // if error opening file

	fprintf(fout, "extern int AttentionFocusInit(void);\n");

	fclose(fout);

	fprintf(stdout, "Generated \"attention_focus_definition.h\"\n");

	return;
};


/// Main parsing and analysing function
void 
//-----------------------------------------------------------
analyse( const char *filename )
//-----------------------------------------------------------
{
  FILE *fin;               //xml source file
  XML_Parser parser;       //parser for the elements
  XML_State userdata;      //global user data
  char buff[MAXBUFF];      //string with element to parse
  int len, done, line;

  int i;
  
  // Fill data structures and userdata
  //ud = &userdata;
  for ( i = 0 ; i < MAXAFS ; i++ ) afs[i] = NULL;
  for ( i = 0 ; i < MAXRULES ; i++ ) mrules[i] = NULL;
  for ( i = 0 ; i < MAXCONDITIONS ; i++ ) conditions[i] = NULL; 
  for ( i = 0 ; i < MAXACTIONS ; i++ ) actions[i] = NULL; 
  free_af = 0;
  free_mr = 0;
  free_co = 0;
  free_ac = 0;
  strcpy( section, "\0");
  strcpy( current_tag, "\0");
  current_af = NULL;
  atom_class = -1;
  
  // open xml file 
  if( (fin = fopen( filename, "r" )) == NULL ){
    fprintf( stderr, "Error opening file %s\n", filename );
    exit( -1 );
  } // if error opening file

  // create parser and analyse document
  parser = XML_ParserCreate( NULL );
  if( !parser ){
    fprintf(stderr, "Couldn't allocate memory for parser\n");
    exit( -1 );
  } // if not enough memory
  
  // define handlers for analysing xml code
  XML_SetUserData( parser, &userdata );
  XML_SetXmlDeclHandler( parser, xmldecl_handler );
  if( strstr( filename, ".xml" ) != NULL )
    XML_SetElementHandler( parser, xmlstartelem_handler, xmlendelem_handler );
  else{
    fprintf( stderr, "Error: invalid filetype\n" );
    XML_ParserFree( parser );
    exit( -1 );
  } //else (not an xml or xsd file)
  XML_SetCharacterDataHandler( parser, xmltext_handler );

  // analising xml file
  userdata.parser = parser;
  userdata.wrong_version = 0;
  userdata.depth = 0;
  //userdata.is_a_class = 0;
  for( line = 1, done = 0; fgets( buff, MAXBUFF, fin ) != NULL; line++ ){
#ifdef DEBUG    
    printf( "line %d(%d): %s\n", line, strlen(buff), buff );
#endif    
    if( XML_Parse( parser, buff, strlen( buff ), done) == XML_STATUS_ERROR) {
	fprintf( stderr, "Parse error at line %d:\n%s\n",
              XML_GetCurrentLineNumber( parser ),
              XML_ErrorString( XML_GetErrorCode( parser )));
	exit( -1 );
    } //if error in xml file 
  
  } //for !EOF


  // Code genration invocations
  generate_condition_formals();
  generate_action_formals();
  //printSymbols();
  gettimeofday(&tv, &tz);
  ctime_r((time_t *)&(tv.tv_sec), date_buff);
  //printf("%s\n", date_buff);
  generateCode();

  // free memory and close files
  XML_ParserFree( parser );
  fclose( fin );

} //analyse()



/// Madatory overall main function
int 
//------------------------------------------------------------
main( int  argc, 
      char *argv[] )
//------------------------------------------------------------
{
  // check sintax
  if( argc < 2 ){
    fprintf( stderr, "Usage: parser <file.xml>\n" );
    return -1;
  }
  //if bad sintax

  // analyse xml file
  analyse( argv[1] );
  //analyse( "../mrules_all.xml" );
  
  printf( "** %s file processed **\n" , argv[1] );
  return 0;

  

} //main()

