/*
   FILE:
	trace2kiwi.c

   DESCRIPTION: 
        Program that reads one or more files (up to MAX_INT_FILES)
	containing a trace and writes them in kiwi format


   Copyright (C) 2003
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Agustin Espinosa Minguet



  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "../../../common/include/artis_event_definition.h"

#define MAX_INT_FILES    2


#define KIWI_RELEASE   0	//|LLEGA activacin
#define KIWI_ENDTASK   1	//|ACABA finalizacin
#define KIWI_DEADLINE  2	//|PLAZO plazo de una tarea
#define KIWI_STARTCPU  4	//|C-EJE comienza el uso de la CPU
#define KIWI_STOPCPU   3	//|T-EJE termina el uso de la CPU
#define KIWI_READY     5	//|C-PRE comienza el estado preparado
#define KIWI_NOTREADY  6	//|T-PRE termina el estado preparado
#define KIWI_STARTCRIT 7	//|ENTSC comienza una seccin critica
#define KIWI_ENDCRIT   8	//|SALSC termina una seccin critica
#define KIWI_EVENTUP   9	//|EV-AR evento (dibujado arriba)
#define KIWI_EVENTDOWN 10	//    |EV-AB evento (dibujado abajo)
#define KIWI_SUSPEND   11	//|SUSPE tarea suspendida
#define KIWI_VERTICAL  12       // Draws a vertical black line
#define KIWI_TEXTOVER  13	// Text over the square in kiwi
#define KIWI_TEXTUNDER  14	// Text under the square in kiwi


#define DS_COLOR  "LightSkyBlue"
#define EM_COLOR  "SteelBlue"
#define SLS_COLOR "LightSteelBlue"

// Uncomment this for debug output
//#define DEBUG

#define OUTPUT_TO_KIWI(file,t,event,task) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld  %d %d\n", \
				    (long) sec, (long) nsec, (int) event, (int) task); \
                             };

#define OUTPUT_TO_KIWIINT(file,t,event,task,arg1) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld %d %d %d\n", \
			            (long) sec, (long) nsec, (int) event, (int) task, (int) arg1); \
                             };

#define OUTPUT_TO_KIWISTR(file,t,event,task,arg1) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld %d %d \"%s\"\n", \
			            (long) sec, (long) nsec, (int) event, (int) task, arg1); \
                             };

#define COLOR_OUTPUT_TO_KIWISTR(file,t,event,task,arg1, arg2) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld %d %d \"%s\" \"%s\"\n", \
			            (long) sec, (long) nsec, (int) event, (int) task, arg1, arg2); \
                             };

#define COLOR_OUTPUT_TO_KIWITEXTUNDER(file,t,task,arg1, arg2) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld TEXTUNDER %d \"%s\" \"%s\"\n", \
			            (long) sec, (long) nsec, (int) task, arg1, arg2); \
                             };

#define COLOR_OUTPUT_TO_KIWITEXTOVER(file,t,task,arg1, arg2) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld TEXTOVER %d \"%s\" \"%s\"\n", \
			            (long) sec, (long) nsec, (int) task, arg1, arg2); \
                             };

#define VLINE_COLOR_OUTPUT_TO_KIWISTR(file, t, arg1, arg2) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld VLINE \"%s\" \"%s\"\n", \
			            (long) sec, (long) nsec, arg1, arg2); \
                             };

#define KERNEL_TASK    0
#define NONE          -1


static FILE *kiwi_file;
static FILE *interm_file;




int
main (int argc, char *argv[])
{
  char kiwi_filename[32];
  char interm_filenames[MAX_INT_FILES][32];
  int  file_i;
  int  n_files;
  int  ntasks;
  long long initial_timestamp;
  long long timestamp;
  long long time;
  long long in_time, out_time;
  int  task;
  int  event;
  int  prev_task = NONE;
  double  mhz;

  int  ds_task = 0;
  int  linux_task = 0;

  int bSLS_Active = 0;
  int bEM_Active  = 0;
  int bCriticalSection = 0;
  int  i = 0;

  char sEvent[8];

  /* Test correct number of arguments */
  if (argc < 5) {
    fprintf (stderr, "Use: trace2kiwi mhz outfile n_files file1 [...]\n");
    fprintf (stderr, "     mhz         : Processor speed\n");
    fprintf (stderr,
	     "     outfile     : name of the (kiwi) output file (no extension needed)\n");
    fprintf (stderr,
	     "     nfiles      : number of entry (intermediate) files\n");
    fprintf (stderr, "     file1 [...] :  intermediate file name(s)\n");
    fprintf (stderr, "\n");
    exit (1);
  }

  /* Get the MHZ of the processor */
  mhz = atof (argv[1]);
  if (mhz < 0.0) {
    fprintf (stderr, "wtrace: Error: processr_mhz %f no valid\n", mhz);
    exit (1);
  }


  /* Create the output (kiwi) file */
  strcpy (kiwi_filename, argv[2]);
  strcat (kiwi_filename, ".ktr");
  if ((kiwi_file = fopen (kiwi_filename, "w")) == NULL) {
    fprintf (stderr, "trace2kiwi: Error in creating kiwi file %s\n", argv[2]);
    exit (1);
  }

  /* Get the number of intermediate files */
  n_files = atoi (argv[3]);
  if (n_files < 0 || n_files > MAX_INT_FILES) {
    fprintf (stderr, 
	     "trace2kiwi: Incorrect number of intermediate files %s (should be [1..%d])\n", 
	     argv[3], MAX_INT_FILES);
    exit (1);
  }

  /* The first event *in each file* is the amount of application rt-tasks */

  for (file_i=0; file_i < n_files; file_i++) {
      
    strcpy (interm_filenames[file_i], argv[4+file_i]);

#ifdef DEBUG
    fprintf(stderr, "\n\nFILE %d = %s\n", file_i, interm_filenames[file_i]);
#endif

    if ((interm_file = fopen (interm_filenames[file_i], "r")) == NULL) {
      fprintf (stderr, "trace2kiwi: couldn't open intermediate file %s\n", 
	       interm_filenames[file_i]);
      exit(1);
    }


    if (file_i == 0) {
      fscanf(interm_file, "%lld %d %d\n", &initial_timestamp, &event, &ntasks);

      if (event == NTASKS && ntasks > 0) {
	
	linux_task = ntasks + 1;
	ds_task    = ntasks + 2;

	fprintf (kiwi_file, "-1 LINE_NAME 0 Kernel\n");
	fprintf (kiwi_file, "-1 LINE_NAME %d Linux\n", linux_task);	// -1 is for the sed postprocess    
	fprintf (kiwi_file, "-1 LINE_NAME %d DS\n", ds_task);	// -1 is for the sed postprocess    

      }
    }

    i = 0;

    while (! feof (interm_file)) {
      
      fscanf(interm_file, "%lld %d %d %lld %lld\n", &timestamp, &event, &task, 
	     &in_time, &out_time);
      
      time = (long long) (((double) ((timestamp - initial_timestamp) * 1000)) 
			  / 
			  (double) mhz);

      if (time < 0) {
	continue;
      }

#ifdef DEBUG    
      fprintf(stderr, "%ld.%09ld %d %d\n", (long) (time / 1000000000), 
	      (long) (time % 1000000000), event, task);
#endif

      switch (event) {
      
      case OVERHD_START:
	//	if (prev_task != NONE) {
	//  OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STOPCPU, prev_task);
	//}
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, KERNEL_TASK);
	break;
	
      case OVERHD_STOP:
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STOPCPU, KERNEL_TASK);
	break;
	
      case CONTEXT_SW:
	if (prev_task != NONE) {
	  OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STOPCPU, prev_task);
	}
	if (bCriticalSection) {
	  OUTPUT_TO_KIWISTR (kiwi_file, time, KIWI_STOPCPU, prev_task, "red");
	}
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, task);
	prev_task = task;
	break;
	
      case ACTIV_START:
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_RELEASE, task);
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_READY, task);
	OUTPUT_TO_KIWI (kiwi_file, (time + in_time),
			KIWI_DEADLINE, task);
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, (time + out_time),
				 KIWI_EVENTUP, KERNEL_TASK, "R", "red");
	break;
	
      case ACTIV_END:
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTDOWN, task, "final", "brown4");
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_NOTREADY, task);
	break;
	
      case INITIAL_END:
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTDOWN, task, "ini", "brown4");
	break;
	
      case ADVANCED_INT:
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTUP, KERNEL_TASK, "A", "red");
	break;

      case INTERRUPT:
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTUP, KERNEL_TASK, "int", "blue");
	break;

      case MCR_BEGIN:
	VLINE_COLOR_OUTPUT_TO_KIWISTR(kiwi_file, time, "MCRB", "");
	break;

      case MCR_END:
	VLINE_COLOR_OUTPUT_TO_KIWISTR(kiwi_file, time, "MCRE", "");
	break;

      case MUTEX_LOCK:
	bCriticalSection = 1;
	sprintf( sEvent, "  %d", (int) in_time );
	OUTPUT_TO_KIWI(kiwi_file, time, KIWI_STOPCPU, task); 
	OUTPUT_TO_KIWI(kiwi_file, time, KIWI_STARTCPU, task); 
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_STARTCRIT, task, sEvent, "black");
	break;

      case MUTEX_UNLOCK:

	OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, task, "red"); 
	OUTPUT_TO_KIWI(kiwi_file, time, KIWI_STARTCPU, task); 
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time, KIWI_ENDCRIT, task, "", "black");
	bCriticalSection = 0;
	break;
	
      case SEND_SCHED_FIFO:
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTDOWN, KERNEL_TASK, "FIFO", "green");
	sprintf( sEvent, "EoS %d", (int) out_time );  // End of Slack <Reactivity Degree>
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, (time + in_time),
				  KIWI_EVENTDOWN, linux_task, sEvent, "red");
	break;

      case INI_START:
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTUP, task, "i_sta", "blue");
	break;
	
      case INI_END:
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTUP, task, "i_end", "blue");
	break;
	
      case FIN_START:
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTUP, task, "f_sta", "blue");
	break;
	
      case FIN_END:
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTUP, task, "f_end", "blue");
	break;

      case OPT_START:
#ifdef DEBUG    
	fprintf (stderr, "trace2kiwi: DS %d START\n", ds_task); 
#endif
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STOPCPU, linux_task);
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, ds_task);
	break;

      case OPT_END:
#ifdef DEBUG    
	fprintf (stderr, "trace2kiwi: DS %d STOP\n", ds_task); 
#endif
	OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, ds_task, DS_COLOR);
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, linux_task);
	break;
	
      case CALL_CONTROL:  //Not used anymore
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTDOWN, ds_task, "call_ds", "blue");
	break;

      case EXECUTE_UNTIL: //Not used anymore
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTDOWN, ds_task, "exec", "green");
	break;

      case INSERT_EVENT:
	if ( task == 1 ) {
	  task = linux_task;
	}
	
	sprintf( sEvent, "Ev %d", (int) in_time );
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTDOWN, task, sEvent, "green");
	break;

      case EXTRACT_EVENT:
	if ( task == 1 ) {
	  task = ds_task;
	}
	
	sprintf( sEvent, "Ev %d", (int) in_time );
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTUP, task, sEvent, "red");
	break;

      case META_RULE_ACTION:
	if ( task == 1 ) {
	  task = linux_task;
	}
	
	sprintf( sEvent, "MRA %d", (int) in_time );
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTDOWN, task, sEvent, "purple");
	break;
	
      case LEVEL_START:
	task = task + ds_task;
#ifdef DEBUG    
	fprintf (stderr, "trace2kiwi: LEVEL %d START\n", task); 
#endif
	OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, ds_task, SLS_COLOR);
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, task);
	break;
	
      case LEVEL_STOP:
	task = task + ds_task;
#ifdef DEBUG    
	fprintf (stderr, "trace2kiwi: LEVEL %d STOP\n", task); 
#endif
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STOPCPU,  task);
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, ds_task); //Restart SLS
	break;

      case LEVEL_SUSPEND:
	task = task + ds_task;
#ifdef DEBUG    
	fprintf (stderr, "trace2kiwi: LEVEL %d SUSPEND\n", task); 
#endif
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STOPCPU,  task);
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_SUSPEND,  task);
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, ds_task);  // Restart SLS
	break;

      case SLS_POLICY:
	if ( task == 1 ) {
	  task = ds_task;
	}
	sprintf( sEvent, "SLS %d", (int) in_time );
	COLOR_OUTPUT_TO_KIWITEXTUNDER(kiwi_file, time, task, sEvent, "DarkOliveGreen");
	break;

      case AA_QUALITY:
	if ( task == 1 ) {
	  task = ds_task;
	}
	sprintf( sEvent, "Qty %d", (int) in_time );
	COLOR_OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_TEXTUNDER, task, sEvent, "DarkOrchid");
      	break;

      case EVENT_RATE:
	if ( task == 1 ) {
	  task = ds_task;
	}
	sprintf( sEvent, "ERate %d", (int) in_time );
	COLOR_OUTPUT_TO_KIWITEXTOVER(kiwi_file, time, task, sEvent, "DarkOliveGreen");
	break;

      case EM_START:
	bEM_Active = 1;
        if (bSLS_Active) {
	  OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, ds_task, SLS_COLOR); 
	}
	else {
	  OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, ds_task, DS_COLOR); 
	}
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, ds_task);
	COLOR_OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_TEXTOVER, ds_task, " EM ", EM_COLOR);
      	break;

      case EM_STOP:  
	bEM_Active = 0;
	OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, ds_task, EM_COLOR); 
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, ds_task);
      	break;

      case SLS_START:
	bSLS_Active = 1;
	bEM_Active  = 0;
	OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, ds_task, EM_COLOR); 
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, ds_task);
	COLOR_OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_TEXTOVER, ds_task, " SLS ", SLS_COLOR);
      	break;

      case SLS_STOP:
	OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, ds_task, SLS_COLOR); 
	OUTPUT_TO_KIWI (kiwi_file, time, KIWI_STARTCPU, ds_task);
	bSLS_Active = 0;
      	break;

	//      case ADVANCE_FINAL:
	//	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
	//				 KIWI_EVENTUP, task, "ADV", "LightCoral");
	//	break;
	
      case WAKEUP_FIFO:
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTUP, task, "WAKE_FIFO", "blue");
	break;
	
      case SLACK_AVAILABLE:
	sprintf ( sEvent, "%d", (int) (in_time / 1000));
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTUP, KERNEL_TASK, sEvent, "orange");
	break;

      case OPT_MUTEX_LOCK:
	bCriticalSection = 1;

	if (task == 1) {
	  task = ds_task;
	}
	else {
	  task += ds_task;
	}

	if (task == ds_task) {
	  if (bSLS_Active) {
	    OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, ds_task, SLS_COLOR); 
	  }
	  else if (bEM_Active) {
	    OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, ds_task, EM_COLOR); 
	  }
	  else {
	    OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, ds_task, DS_COLOR); 
	  }
	}
	else {
	  OUTPUT_TO_KIWI(kiwi_file, time, KIWI_STOPCPU, task); 
	}
	
	sprintf( sEvent, "  %d", (int) in_time );
	OUTPUT_TO_KIWI(kiwi_file, time, KIWI_STARTCPU, task); 
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_STARTCRIT, task, sEvent, "black");
	
	break;

      case OPT_MUTEX_UNLOCK:
	if (task == 1)
	  task = ds_task;
	else
	  task = task + ds_task;

	OUTPUT_TO_KIWISTR(kiwi_file, time, KIWI_STOPCPU, task, "red"); 
	OUTPUT_TO_KIWI(kiwi_file, time, KIWI_STARTCPU, task); 
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time, KIWI_ENDCRIT, task, "", "black");

	bCriticalSection = 0;

	break;
	
      case DEBUG_EVENT:
	if ( task == 1 ) {
	  task = ds_task;
	}
	
	sprintf( sEvent, "DBG %d", (int) in_time );
	COLOR_OUTPUT_TO_KIWISTR (kiwi_file, time,
				 KIWI_EVENTDOWN, task, sEvent, "DarkGoldenrod3");
	break;

       }
    }
    fclose(interm_file);
  }
   return 0;
}
