/*

   FILE:
	trace.h						

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-06

  CHANGELOG:


  DESCRIPTION:

    The TRACE module provides a very simple set of tracing services to
    both RT-Linux and Linux sides of the ARTIS prototype.

    An "event" is here defined as an instance of a predefined struct
    type, which contains the following fields:

      - timestamp: the time at which the event was recorded, in TSC
      units.

      - type: integer identifier the event type produced. Event types
      are statically registered as integer constants in header file
      "../../common/include/artis_event_definition.h"

      - id: integer identifier of the "task" (RT-Linux task or Linux
      level) which produced the event.

      - data1: first data field (long long)

      - data2: second data field (long long)

    An event can have, at most, two long long data fields associated,
    which can also accomodate data of other types, by applying a  
    convenient casting.  For example:
          trace_event (NTASKS, 0,  (int) number_of_tasks, 0);

    The interface functions are equal to the Linux and RT-Linux sides
    of the application, exception being the initialization function of
    each subsystem.

    However, tracing works differently in Linux and RT-Linux:

     - In RT-Linux, events are written first to a shared memory area
       (mbuff) which, after the rt-application is over, is accessed by
       a Linux program called "mem2trace", which then reads all the
       events in memory and writes them into a file. "mem2trace"
       simply reads events from the mbuff until either it reaches the
       end of the mbuff area or it finds an "event" with a negative
       type (indicating that there are no more events).

     - In Linux, the application ("linux_side") directly generates
       trace events into a file. This file and the file generated
       by "mem2trace" are finally combined into a single trace file
       in "kiwi" format for graphical display. This is done by 
       another tool called "trace2kiwi".


    In general, all these functions return 0 if everything went OK and
    a negative number if something went wrong.

*/

#ifndef __TRACE_H
#define __TRACE_H


// Event types are defined here:
#include "../../common/include/artis_event_definition.h"



// Visible definitions and macros

/* 
   Note: the name and the size of the mbuff for events has to be public
   since it has to be known by the Linux tool "mem2trace".
*/

#define MBUFF_TRACE_EVENTS_NAME    "artis_trace_events_mbuff"
#define MBUFF_TRACE_EVENTS_SIZE    0x100000

#define VOID_EVENT_TYPE            -1   /* A void event in the array 
					   indicates that there are no
					   more events left */


// Visible types

typedef struct __event_t {
  long long  timestamp;
  int        type;       
  int        id;
  long long  data1;
  long long  data2;
} event_t;



// Visible functions:

#ifdef MODULE
int  trace_init_module (void);
void trace_cleanup_module (void);
#else 
int  trace_init (char *filename);
int  trace_cleanup(void);
#endif


int  trace_start (void);
int  trace_stop  (void);

int  trace_event (int type, int id, long long d1, long long d2); 


#endif // __TRACE_H
