/*

   FILE:
	trace.c						

   Copyright (C) 1998-2003
        GTI-IA, DSIC, UPV


  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

*/

/*

  AUTHOR(S): A. Terrasa

  CREATION DATE: 2003-10-06

  CHANGELOG:


  DESCRIPTION:

    The TRACE module provides a very simple set of tracing services to
    both RT-Linux and Linux sides of the ARTIS prototype.

    An "event" is here defined as an instance of a predefined struct
    type, which contains the following fields:

      - timestamp: the time at which the event was recorded, in TSC
      units.

      - type: integer identifier the event type produced. Event types
      are statically registered as integer constants, as specified
      in header file "../../common/include/artis_event_definition.h"

      - id: integer identifier of the "task" (RT-Linux task or Linux
      level) which produced the event.

      - data1: first data field (long long)

      - data2: second data field (long long)

    An event can have, at most, two long long data fields associated,
    which can also accomodate data of other types, by applying a  
    convenient casting.  For example:
          trace_event (NTASKS, 0,  (int) number_of_tasks, 0);

    The interface functions are equal to the Linux and RT-Linux sides
    of the application, exception being the initialization function of
    each subsystem.

    In general, all these functions return 0 if everything went OK and
    a negative number if something went wrong.

    This source file is compiled twice, one as a kernel module and
    another one as a regular Linux object file.

*/

#ifdef MODULE
#include <mbuff.h>
#include <asm/timex.h>
#else
#include <stdio.h>
#include <string.h>
#endif

#include "../../common/include/artis_conf.h"
#include "../include/trace.h"

///#define DEBUG




#ifdef MODULE

#include <rtl.h>

static volatile event_t *events = (event_t *) NULL;
static int ev_index = 0;
static int rt_trace_running = 0;
static event_t  *last_event = NULL;


int  
trace_init_module (void) {

  events =  (volatile event_t *) mbuff_attach (MBUFF_TRACE_EVENTS_NAME,
					       MBUFF_TRACE_EVENTS_SIZE);

  if (events == NULL) {
#ifdef DEBUG
    rtl_printf("(trace.c [RTL]) Cannot allocate mbuff memory for events\n");
#endif
    return -ENOMEM;
  }

  last_event = (event_t *) ((unsigned long) events 
			    + MBUFF_TRACE_EVENTS_SIZE 
			    - sizeof(event_t));
  rt_trace_running = 0;

  events[0].type  = VOID_EVENT_TYPE;

  return 0;
}





void
trace_cleanup_module (void) {

#ifdef DEBUG
  rtl_printf("(rt_trace.c::trace_cleanup_module) Ending\n");
#endif

  mbuff_detach (MBUFF_TRACE_EVENTS_NAME,  (void *) events);
}





int  
trace_start (void) {
  if (events != (event_t *) NULL) {
    rt_trace_running = 1;
  }
  return 0;
}




int  
trace_stop  (void) {
  rt_trace_running = 0;
  return 0;
}






int  
trace_event (int type, int id, long long d1, long long d2) {

  if (!rt_trace_running || type < 0) {
    return -1;
  }

  if ((event_t *) &events[ev_index] < last_event) {
    // Store the event:
    events[ev_index].type  = type;
    events[ev_index].id    = id;
    events[ev_index].data1 = d1;
    events[ev_index].data2 = d2;
    rdtscll (events[ev_index].timestamp);
    ev_index += 1;
    
    // Set the next event void:
    if ((event_t *) &events[ev_index] < last_event) {
        events[ev_index].type  = VOID_EVENT_TYPE;
    }
  }
    
  return 0;
}


#else // ! MODULE -------------------------------------------------


#define FILE_NAME_SIZE  256

static FILE    *trace_file;
static int     trace_running = 0;

static __inline__ unsigned long long rdtsc(void)
{
        unsigned long long ret;
        __asm__ __volatile__("rdtsc"
                            : "=A" (ret)
                            : /* no inputs */);
        return ret;
}





int  
trace_init (char *filename) {
  char trace_filename[FILE_NAME_SIZE];
  
  strncpy(trace_filename, filename, FILE_NAME_SIZE);
  strcat(trace_filename, ".tri");
  
  if ((trace_file = fopen(trace_filename, "w")) == NULL) {
#ifdef DEBUG
    fprintf(stderr, "(trace.c [L]) Cannot create trace file %s\n", 
	    trace_filename);
#endif
    return -1;
  }
#ifdef DEBUG
  else fprintf(stderr, "(trace.c [L] Trace file %s .tri created\n", 
	       filename);
#endif

  trace_running = 0;
  return 0;
}





int  
trace_cleanup (void) {
#ifdef DEBUG
  fprintf(stderr, "(trace.c [L]) Closing trace file\n");
#endif
  return (fclose (trace_file));
}






int  
trace_start (void) {
  trace_running = 1;
  return 0;
}




int  
trace_stop (void) {
  trace_running = 0;
  return 0;
}





int  
trace_event (int type, int id, long long d1, long long d2) {

#ifdef DEBUG
  fprintf(stderr, "(trace.c [L] Before tracing Event %d \n", type);
#endif

  if (trace_running) {

#ifdef DEBUG
    fprintf(stderr, "(trace.c [L] Event %d \n", type);
#endif

    fprintf(trace_file, "%lld %d %d %lld %lld\n",
	    rdtsc(),
	    type,
	    id,
	    d1,
	    d2);
    fflush(trace_file);
  }
  return 0;
}


#endif // #ifdef MODULE
